#ifndef C3DVIEWER_H
#define C3DVIEWER_H

#include <wx/glcanvas.h>

#include "CMesh.h"
#include "CProjectData.h"
#include "trackball.h"
#include "typenames.h"

// declaring prototypes
struct Vrtx;
typedef map<uInt,Vrtx> mapVrtx;
typedef mapVrtx::iterator mapVrtxIt;
typedef mapVrtx::const_iterator mapVrtxConstIt;
struct Prim;
typedef map<uInt,Prim> mapPrim;
typedef mapPrim::iterator mapPrimIt;
typedef mapPrim::const_iterator mapPrimConstIt;

//-----------------------------------------------------------------------/
// structure that stores vertex data
//-----------------------------------------------------------------------/
struct Vrtx
{
    uInt id;
    double coordinates[3];
    double color[3];
    bool inSelection;
};

//-----------------------------------------------------------------------/
// structure that stores primitive data
//-----------------------------------------------------------------------/
struct Prim
{
    uInt id;
    vecVecDbl coordinates;
    vecVecDbl color;
    double normal[3];
    bool inSelection;
};

class C3DViewer : public wxGLCanvas
{
public:
    C3DViewer(wxWindow* parent,
              wxWindowID id=wxID_ANY,
              const wxPoint& pos=wxDefaultPosition,
              const wxSize& size=wxDefaultSize,
              long style=0,
              const wxString& name=wxGLCanvasName,
              int* attribList=0,
              const wxPalette& palette=wxNullPalette) : wxGLCanvas(parent, id, attribList, pos, size, style, name, palette){
        clearData();
        m_context = new wxGLContext(this);
    }
    virtual ~C3DViewer();

    bool m_viewerInitialized; // flag if 3D Viewer is initialized
    uInt m_selectionMode;     // selection mode: 0-element, 1-node

    void initialize();
    void clearData();
    void updateMeshData(
            CMesh* pMesh,
            CProjectData* pProjectData);
    void updateResultColors(
            CMesh* pMesh,
            CProjectData* pProjectData,
            CResults* pResults);
    int updateColor(
            CMesh* pMesh,
            CProjectData* pProjectData,
            setUInt& elemSelection,
            setUInt& nodeSelection);
    void updateLighting(CProjectData* pProjectData);
    void refit(CMesh* pMesh);
    void refresh3DViewer(
            CMesh* pMesh,
            CProjectData* pProjectData,
            setUInt& elemSelection,
            setUInt& nodeSelection);
    void setViewToFront();
    void setViewToTop();
    void setViewToRight();
    void setViewToBack();
    void setViewToBottom();
    void setViewToLeft();
    void setViewToIso();
    void onLeftMouseButton(
            wxMouseEvent& event,
            setUInt& elemSelection,
            setUInt& nodeSelection);
    void onMouseMiddleUp(
            wxMouseEvent& event,
            CMesh* pMesh);
    void onMouseMiddleDown(wxMouseEvent& event);
    void onMouseWheel(wxMouseEvent& event);
    void onMouseMove(
            wxMouseEvent& event,
            CProjectData* pProjectData);

private:
    wxGLContext* m_context;

    // mesh related data
    mapVrtx m_vrtxMap;   // map of vertices
    mapPrim m_primMap;   // map of primitives
    double m_maxNode[3]; // maximum node bounds
    double m_minNode[3]; // minimum node bounds
    double m_maxBody[3]; // maximum body bounds (body+dummy)
    double m_minBody[3]; // minimum body bounds (body+dummy)
    // results related data
    vecVecVecDbl m_resultColors;   // colors at each node of primitive extracted from surface results
    // viewer operations data
    uInt     m_neighElement;       // current watched element for neighboring analysis
    bool     m_leftMouseDown;      // flag if left mouse button is held down
    bool     m_waitForRotCenter;   // flag if waiting for rotation center (MMB up)
    bool     m_boxSelectionMode;   // flag if in box selection mode
    float    m_mousePosition  [2]; // mouse position
    double   m_selectionCoords[4]; // selection box coordinates
    GLdouble m_corners        [6]; // 3D viewer window coordinates (including z coordinates - far and near planes)
    GLdouble m_rotCenter      [3]; // rotation center coordinates
    // model state in 3D Viewer data
    float    m_quaternions[4]; // model orientation
    double   m_scale      [2]; // model scaling
    double   m_translation[2]; // model translation

    void drawPrimitives(
            bool drawForSelection,
            GLdouble* color=NULL,
            bool drawForLines=false);
    void drawVertices(bool drawForSelection);
    void drawCoordinateSystem();
    void drawNodeIDs();
    void drawElemIDs();
    void drawSelectionBox();
    void drawLegend(CProjectData* pProjectData);
    void makeRasterFont();
    void getPerspective(wxSize& sz);
    void setPerspective();
    void modifyPerspectiveForPick(
            double& mouseX,
            double& mouseY,
            wxSize& sz);
    void modifyPerspectiveForSelection(
            bool& addition,
            double& mouseX,
            double& mouseY,
            wxSize& sz);
    void processPick(
            GLuint* buffer,
            const uInt& selectionSize,
            int& pickID);
    void processSelection(
            GLuint* buffer,
            uInt selSize,
            vecUInt& hitIDs);
    void printString(const string& textToPrint);
    int calcNormal(
            const vecVecDbl& coordinates,
            double normal[3]);
    void fitDataToUnit(
            vecDbl& fieldData,
            const double& max,
            const double& min);
    vecDbl RGBFilter(const double& value);
};

#endif
