#ifndef MESH_H
#define MESH_H

#include <deque>
#include <cstdlib>
#include <algorithm>

#include "params.h"
#include "typenames.h"

// declaring prototypes
struct Node;
// map of nodes with IDs
typedef map<uInt,Node> mapNode;
typedef mapNode::iterator mapNodeIt;
typedef mapNode::const_iterator mapNodeConstIt;
// vector of pointers to node structures
typedef vector<Node*> vecPNode;
// deque of pointers to node structures
typedef deque<Node*> deqPNode;
typedef deque<Node*>::iterator deqPNodeIt;
typedef vector<deqPNode> vecDeqPNode;
// set of pointers to node structures
typedef set<Node*> setPNode;
typedef setPNode::iterator setPNodeIt;
typedef setPNode::const_iterator setPNodeConstIt;
struct Elem;
// map of elements with IDs
typedef map<uInt,Elem> mapElem;
typedef mapElem::iterator mapElemIt;
typedef mapElem::const_iterator mapElemConstIt;
// vector of pointers to element structures
typedef vector<Elem*> vecPElem;
// set of pointers to element structures
typedef set<Elem*> setPElem;
typedef setPElem::iterator setPElemIt;
typedef setPElem::const_iterator setPElemConstIt;

//-----------------------------------------------------------------------/
// structure that holds node info
//-----------------------------------------------------------------------/
struct Node
{
    uInt id;                // node ID
    double coordinates[3];  // node coordinates (x,y,z)
    setPElem elements;      // elements that use this node

    bool operator< (const Node &Node2) const{return id< Node2.id;}
    bool operator==(const Node &Node2) const{return id==Node2.id;}
    bool operator!=(const Node &Node2) const{return id!=Node2.id;}
};

//-----------------------------------------------------------------------/
// structure that holds element info
//-----------------------------------------------------------------------/
struct Elem
{
    uInt id;            // element ID
    uInt type;          // type of element (1-body quad, 2-body tria, 10-wake quad, 11-wake tria, 20-dummy quad, 21-dummy tria)
    vecPNode nodes;     // constituing nodes (vector is used because order matters, but duplicates possible)
    setPElem neighbors; // neighboring elements (set is used because order doesn't matter and duplicates avoided)
    setPElem wakes;     // wakes that use this element (used in non-wake elements only)
    Elem* trailUpper;   // upper trailing element (used in wake elements only)
    Elem* trailLower;   // lower trailing element (used in wake elements only)

    Elem();

    bool operator< (const Elem &Elem2) const{return id< Elem2.id;}
    bool operator==(const Elem &Elem2) const{return id==Elem2.id;}
    bool operator!=(const Elem &Elem2) const{return id!=Elem2.id;}
};

//-----------------------------------------------------------------------/
// contains complete mesh information and methods for its manipulation
//-----------------------------------------------------------------------/
class CMesh
{
public:
    // define operator "=" for mesh copying
    CMesh& operator=(const CMesh &Mesh){
        // check for self-assignment
        if (this == &Mesh)
            return *this;

        // clear all mesh data first
        clearData();

        // copy nodes
        uInt i; // this index is used for more efficient map inserting  CHECK if any performance is gained!
        mapNodeConstIt nodeMapIt;
        for (nodeMapIt=Mesh.meshNodes.begin(), i=0; nodeMapIt!=Mesh.meshNodes.end(); nodeMapIt++, i++)
            forceInsertNodeWithIDCoords(nodeMapIt->second.id, nodeMapIt->second.coordinates, i);

        // copy elements
        mapElemConstIt elemMapIt;
        for (elemMapIt=Mesh.meshElems.begin(), i=0; elemMapIt!=Mesh.meshElems.end(); elemMapIt++, i++)
            forceInsertElemWithIDNodes(elemMapIt->second.id, elemMapIt->second.type, elemMapIt->second.nodes, i);

        // add neighbors
        for (elemMapIt=Mesh.meshElems.begin(); elemMapIt!=Mesh.meshElems.end(); elemMapIt++){
            // go through neighbors and add by ID
            for (setPElemConstIt it=elemMapIt->second.neighbors.begin(); it!=elemMapIt->second.neighbors.end(); it++)
                meshElems[elemMapIt->second.id].neighbors.insert(&meshElems[(*it)->id]);
        }

        // add wakes
        for (elemMapIt=Mesh.meshElems.begin(); elemMapIt!=Mesh.meshElems.end(); elemMapIt++){
            // go through neighbors and add by ID
            for (setPElemConstIt it=elemMapIt->second.wakes.begin(); it!=elemMapIt->second.wakes.end(); it++)
                meshElems[elemMapIt->second.id].wakes.insert(&meshElems[(*it)->id]);
        }

        // add trailing elements
        for (elemMapIt=Mesh.meshElems.begin(); elemMapIt!=Mesh.meshElems.end(); elemMapIt++){
            // check if pointer to trailing element is valid and add
            if (elemMapIt->second.trailUpper) meshElems[elemMapIt->second.id].trailUpper = &meshElems[(*elemMapIt->second.trailUpper).id];
            if (elemMapIt->second.trailLower) meshElems[elemMapIt->second.id].trailLower = &meshElems[(*elemMapIt->second.trailLower).id];
        }

        return *this;
    }

    CMesh();
    ~CMesh();
    uInt getNumBodyQuad();
    uInt getNumDummQuad();
    uInt getNumWakeQuad();
    uInt getNumBodyTria();
    uInt getNumDummTria();
    uInt getNumWakeTria();
    uInt getNumNode();
    uInt getMaxNodeID();
    uInt getMaxElemID();
    double* getMaxNode();
    double* getMinNode();
    double* getMaxBody();
    double* getMinBody();
    double* getMaxWake();
    double* getMinWake();
    bool checkIfNodeExists(const uInt& nodeID);
    bool checkIfElemExists(const uInt& elemID);
    int getNodeCoords(
            const uInt& nodeID,
            double coordinates[3]);
    void getNodeData(
            vecUInt& nodeIDs,
            vecVecDbl& coordinates);
    int getElemData(
            const uInt& elemID,
            uInt& elemType,
            vecUInt& nodeIDs,
            vecUInt& neighborTrailIDs);
    int getElemCoords(
            const uInt& elemID,
            vecVecDbl& coordinates);
    int getElemsData(
            vecUInt& elemIDs,
            vecUInt& elemTypes,
            vecVecUInt& nodeIDs);
    int getElemIDsTypes(
            vecUInt& elemIDs,
            vecUInt& elemTypes);
    int getElemSolverData(
            vecUInt& elemIDs,
            vecUInt& elemTypes,
            vecVecUInt& elemNodeIDs,
            vecVecUInt& neighborIDs,
            vecVecUInt& trailIDs);
    void getElemIDsTypesCoords(
            vecUInt& ids,
            vecUInt& types,
            vecVecVecDbl& coordinates);
    void getNeighbors(
            const uInt& monitoredElement,
            vecVecUInt& neighboring);
    void getIDWData(
            vecUInt& nodeIDs,
            vecVecDbl& coordinates,
            vecUInt& elemIDs,
            vecVecUInt& elemNodeIDs);
    int insertNodeWithCoords(const double coordinates[3]);
    int insertNodeWithIDCoords(
            const uInt& id,
            const double coordinates[3]);
    int insertElemWithNodeCoords(
            const vecVecDbl& coordinatesAll,
            const uInt& type);
    int insertElemWithNodeIDs(
            const uInt& type,
            const vecUInt& nodeIDs);
    int insertElemWithIDNodeIDs(
            const uInt& id,
            const vecUInt& nodeIDs);
    int insertElemWithIDTypeNodeIDs(
            const uInt& id,
            const uInt& type,
            const vecUInt& nodeIDs);
    int insertNeighborsToElem(
            const uInt& id,
            const vecUInt& neighbors);
    int insertTrailsToElem(
            const uInt& id,
            const vecUInt& trailIDs);
    int editNode(
            const uInt& nodeID,
            const uInt& newNodeID,
            const double newCoordinates[3]);
    int editElement(
            const uInt& elemID,
            const uInt& newElemID,
            const uInt& elemType,
            const vecUInt& nodeIDs,
            const vecUInt& neighborTrailIDs);
    void reverseNormal(const setUInt& elemSelection);
    void mergeNodes(
            const double& tolerance,
            const string& type,
            const setUInt& nodeSelection=setUInt());
    void mergeNodePair(
            Node* mastr,
            Node* slave,
            bool toMiddle=false);
    void scale(
            const double ratio[3],
            const double origin[3],
            const setUInt& nodeSelection=setUInt(),
            const setUInt& elemSelection=setUInt());
    void translate(
            const double distance[3],
            const setUInt& nodeSelection=setUInt(),
            const setUInt& elemSelection=setUInt());
    void rotate(
            const double angle,
            const double origin[3],
            const double tip[3],
            const setUInt& nodeSelection=setUInt(),
            const setUInt& elemSelection=setUInt());
    void rotateNode(
            Node* node,
            const double angle,
            const double origin[3],
            const double tip[3]);
    void mirror(
            const double origin[3],
            const double tip[3],
            const setUInt& elemSelection=setUInt());
    int convertElemType(
            const uInt& id,
            const string& convertTo);
    int updateWakeTrails(
            const uInt& upperID,
            const uInt& lowerID,
            const setUInt& elemSelection);
    void extrudeLine(
            const double& extrusionLength,
            const uInt& elemNum,
            const setUInt& nodeSelection);
    void separateAtSeam(const setUInt& nodeSelection);
    void updateNeighboringInfo(
            const double& angleThreshold,
            const setUInt& elemSelection=setUInt());
    void updateWakeInfo(
            const double& angleThreshold,
            const double& wakeLengthRatio,
            const uInt& wakeNum,
            const setUInt& elemSelection=setUInt());
    void clearData();
    void clearNeighboringInfo();
    void deleteAllNodes();
    void deleteNodesByID(const vecUInt& nodeIDs);
    void deleteAllElems();
    void deleteElemsByID(const vecUInt& elemIDs);

private:
    mapNode meshNodes;  // map of node structures (actual data storage)
    mapElem meshElems;  // map of element structures (actual data storage)

    // greatest node and element IDs
    uInt maxNodeId;
    uInt maxElemId;

    // number of nodes
    uInt numNode;

    // number of elements
    uInt numBodyQuad;   // number of quadrilateral body elements
    uInt numBodyTria;   // number of triangular body elements
    uInt numWakeQuad;   // number of quadrilateral wake elements
    uInt numWakeTria;   // number of triangular wake elements
    uInt numDummQuad;   // number of quadrilateral dummy elements
    uInt numDummTria;   // number of triangular dummy elements

    // node bounds
    double minNode[3];
    double maxNode[3];
    // body bounds
    double minBody[3];
    double maxBody[3];
    // wake bounds
    double minWake[3];
    double maxWake[3];

    void initializeData();
    void initializeNodeData();
    void initializeElemData();
    Elem* getNeighbor(
            const Elem* element,
            const uInt& firstNodeIndex,
            const uInt& secondNodeIndex);
    Node* insertNodeWithCoordsPrivate(const double coordinates[3]);
    int insertElemWithNodeCoordsPrivate(
            const vecVecDbl& coordinatesAll,
            const uInt& type,
            Elem*& newElemP);
    void forceInsertNodeWithIDCoords(
            const uInt& id,
            const double coordinates[3],
            const int& position = -1);
    void forceInsertElemWithIDNodes(
            const uInt& id,
            const uInt& type,
            const vecPNode& nodes,
            const int& position = -1);
    void deleteNode(mapNodeIt nodeIt);
    void deleteElement(mapElemIt elemIt);
    void updateBounds();
    void updateNodeBounds();
    void updateElemBounds();
    void updateBoundsWithNode(const Node* node);
    void updateBoundsWithElem(const Elem* element);
    void updateMaxNodeID();
    void updateMaxElemID();
    int calcNormal(
            const vecVecDbl& coordinates,
            bool onlyCheck=true,
            double* normal=NULL);
    int calcNormal(
            const vecPNode& nodes,
            bool onlyCheck=true,
            double* normal=NULL);
    vecDbl calcNormal(const Elem* element);
    double angleBetweenElems(
            Elem* elem_1,
            Elem* elem_2);
    void getNodeBranches(
            const setUInt& nodeSelection,
            vecDeqPNode& nodeBranches);
    void growBranch(
            deqPNode& nodeBranch,
            Node* startNode,
            const setPNode& nodes,
            setUInt& usedNodes,
            const string& direction);
};

#endif
