#include "C3DViewer.h"

#include "openGLSymbols.inc"

GLdouble colorFore [3] = {0.65, 0.55, 0.33}; // foreground color
GLdouble colorBack [3] = {0.12, 0.44, 0.60}; // background color
GLdouble colorSel  [3] = {1.00, 0.00, 0.20}; // selection color
GLdouble colorBlack[3] = {0.00, 0.00, 0.00}; // black color

// color for neighbors in order from red (watched element) toward blue (fourth neighbors)
double neighborColors[4][3] = {{1.00, 0.00, 0.00},
                               {0.75, 0.00, 0.25},
                               {0.50, 0.00, 0.50},
                               {0.25, 0.00, 0.75}};

GLfloat light0_color[4] = { 1.0f,  1.0f,  1.0f,  1.0f}; // first light color
GLfloat light0_pos  [4] = {-1.0f, -1.0f, -1.0f,  0.0f}; // first light position
GLfloat light1_color[4] = { 1.0f,  1.0f,  1.0f,  1.0f}; // second light color
GLfloat light1_pos  [4] = { 1.0f,  1.0f,  1.0f,  0.0f}; // second light position

C3DViewer::~C3DViewer()
{
	delete m_context;
}

//-----------------------------------------------------------------------/
// initialize 3D Viewer
//-----------------------------------------------------------------------/
void C3DViewer::initialize()
{
    // set bool to initialized
    m_viewerInitialized = true;

    wxGLCanvas::SetCurrent(*m_context);

    // initialize raster font
    makeRasterFont();

    // remove back faces
    glDisable(GL_CULL_FACE);
    glEnable(GL_DEPTH_TEST);

    // speed-ups
    glEnable(GL_DITHER);
    glShadeModel(GL_SMOOTH);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
    glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);

    // material
    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);
}

//-----------------------------------------------------------------------/
// clear 3D Viewer data
//-----------------------------------------------------------------------/
void C3DViewer::clearData()
{
    m_vrtxMap.clear();
    m_primMap.clear();
    m_resultColors.clear();

    m_selectionMode    = 0;
    m_leftMouseDown    = false;
    m_boxSelectionMode = false;
    m_waitForRotCenter = false;
    m_mousePosition[0] = 0.0;
    m_mousePosition[1] = 0.0;
    for (uInt i=0; i<6; i++) m_corners        [i] = 0.0;
    for (uInt i=0; i<3; i++) m_rotCenter      [i] = 0.0;
    for (uInt i=0; i<4; i++) m_selectionCoords[i] = 0.0;
    m_quaternions[0] = -sqrt(3.0)/4.0;
    m_quaternions[1] = 0.25;
    m_quaternions[2] = sqrt(3.0)/4.0;
    m_quaternions[3] = -0.75;
    m_translation[0] = 0.0;
    m_translation[1] = 0.0;
    m_scale[0] = 2.0;
    m_scale[1] = 2.0;
    m_neighElement = 0;
    for (uInt i=0; i<3; i++){
        m_maxNode[i] = MIN_BOUND;
        m_maxBody[i] = MIN_BOUND;
        m_minNode[i] = MAX_BOUND;
        m_minBody[i] = MAX_BOUND;
    }
}

//-----------------------------------------------------------------------/
// update 3D Viewer data with mesh data
//-----------------------------------------------------------------------/
void C3DViewer::updateMeshData(CMesh* pMesh, CProjectData* pProjectData)
{
    // get node IDs and coordinates
    vecUInt nodeIDs;
    vecVecDbl nodeCoords;
    pMesh->getNodeData(nodeIDs, nodeCoords);

    // store node data into vertex map
    m_vrtxMap.clear();
    for (uInt nodeIndex=0; nodeIndex<nodeIDs.size(); nodeIndex++){
        // create new vertex
        Vrtx newVertex;
        newVertex.id = nodeIDs[nodeIndex];                                              // add ID
        for (uInt i=0; i<3; i++) newVertex.coordinates[i] = nodeCoords[nodeIndex][i];   // add coordinates
        for (uInt i=0; i<3; i++) newVertex.color[i] = colorFore[i];         // add default color
        newVertex.inSelection = false;                                                  // flag not in selection

        // add it to map with position hint
        m_vrtxMap.insert(m_vrtxMap.end(), pair<uInt,Vrtx>(nodeIDs[nodeIndex], newVertex));
    }

    // get element IDs and coordinates
    vecUInt elemIDs;
    vecUInt elemTypes;
    vecVecVecDbl elemCoords;
    pMesh->getElemIDsTypesCoords(elemIDs, elemTypes, elemCoords);

    // store element data into primitive map
    m_primMap.clear();
    for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
        if (!(!pProjectData->showWake && ( elemTypes[elemIndex]==10 || elemTypes[elemIndex]==11) )){
            // create new primitive
            Prim newPrim;
            newPrim.id = elemIDs[elemIndex];                                                // add ID
            newPrim.coordinates.resize(elemCoords[elemIndex].size());                       // add coordinates
            for (uInt nodeIndex=0; nodeIndex<elemCoords[elemIndex].size(); nodeIndex++){
                newPrim.coordinates[nodeIndex].resize(3);
                for (uInt i=0; i<3; i++)
                    newPrim.coordinates[nodeIndex][i] = elemCoords[elemIndex][nodeIndex][i];
            }
            newPrim.color.resize(elemCoords[elemIndex].size());                             // add default color
            for (uInt nodeIndex=0; nodeIndex<elemCoords[elemIndex].size(); nodeIndex++){
                newPrim.color[nodeIndex].resize(3);
                for (uInt i=0; i<3; i++)
                    newPrim.color[nodeIndex][i] = colorFore[i];
            }
            calcNormal(newPrim.coordinates, newPrim.normal);                                // add normal
            newPrim.inSelection = false;                                                    // flag not in selection

            // add it to map with position hint
            m_primMap.insert(m_primMap.end(), pair<uInt,Prim>(elemIDs[elemIndex], newPrim));
        }
    }

    double* maxNode = pMesh->getMaxNode();
    double* minNode = pMesh->getMinNode();
    double* maxBody = pMesh->getMaxBody();
    double* minBody = pMesh->getMinBody();
    m_maxNode[0] = maxNode[0];
    m_maxNode[1] = maxNode[1];
    m_maxNode[2] = maxNode[2];
    m_minNode[0] = minNode[0];
    m_minNode[1] = minNode[1];
    m_minNode[2] = minNode[2];
    m_maxBody[0] = maxBody[0];
    m_maxBody[1] = maxBody[1];
    m_maxBody[2] = maxBody[2];
    m_minBody[0] = minBody[0];
    m_minBody[1] = minBody[1];
    m_minBody[2] = minBody[2];
}

//-----------------------------------------------------------------------/
// update 3D Viewer data with surface results data
//-----------------------------------------------------------------------/
void C3DViewer::updateResultColors(
        CMesh* pMesh,
        CProjectData* pProjectData,
        CResults* pResults)
{
    // clear results colors before further operations
    m_resultColors.clear();

    // exit if surface plot not requested
    if (!pProjectData->surfResultsShown)
        return;

    // get field data
    vecDbl fieldData;
    vecUInt elemIDs;
    if (pResults->getFieldData(pProjectData->selectedData, pProjectData->selectedCase, fieldData, elemIDs))
        return;

    // check field data size
    if (fieldData.empty())
        return;

    // check field vs IDs size
    if (fieldData.size() != elemIDs.size()){
        showErrorMessage("Internal error: Inconsistency in field result size data and element IDs - CONTACT SUPPORT!");
        return;
    }

    // allocate color results vector
    m_resultColors.resize(m_primMap.size());

    // assemble map of indexes and IDs
    map<uInt,uInt> mapIndexID;
    uInt mapIndex=0;
    for (mapPrimIt primIt=m_primMap.begin(); primIt!=m_primMap.end(); primIt++)
        mapIndexID[primIt->second.id] = mapIndex++;

    // check contour type
    if (pProjectData->plotType == "elemental"){
        // fit data to 0.0-1.0 scale (take into account minimum and maximum limits)
        vecDbl fitData = fieldData;
        fitDataToUnit(fitData, pProjectData->minLimit, pProjectData->maxLimit);

        // go through each element
        for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
            // try to find this primitive index
            map<uInt,uInt>::iterator it = mapIndexID.find(elemIDs[elemIndex]);
            if (it != mapIndexID.end()){
                // apply RGB filter
                vecDbl RGBValues = RGBFilter(fitData[elemIndex]);
                // store colors to m_resultColors member
                m_resultColors[it->second].resize(m_primMap[elemIDs[elemIndex]].coordinates.size());
                for (uInt nodeIndex=0; nodeIndex<m_primMap[elemIDs[elemIndex]].coordinates.size(); nodeIndex++){
                    m_resultColors[it->second][nodeIndex].resize(3);
                    for (uInt i=0; i<3; i++)
                        m_resultColors[it->second][nodeIndex][i] = RGBValues[i];
                }
            }
        }
    }
    else{
        // using Inverse Distance Weighing method, find field data values at nodes

        // get data from mesh
        vecUInt nodeIDs;
        vecVecDbl coordinates;
        vecUInt allElemIDs;
        vecVecUInt elemNodeIDs;
        pMesh->getIDWData(nodeIDs, coordinates, allElemIDs, elemNodeIDs);

        // creating node and element index map
        map<uInt,uInt> nodeIndexMap;
        for (uInt nodeIndex=0; nodeIndex<nodeIDs.size(); nodeIndex++)
            nodeIndexMap[nodeIDs[nodeIndex]] = nodeIndex;
        map<uInt,uInt> elemIndexMap;
        for (uInt elemIndex=0; elemIndex<allElemIDs.size(); elemIndex++)
            elemIndexMap[allElemIDs[elemIndex]] = elemIndex;

        // extract element node indexes and element center coordinates
        vecVecUInt elementNodes(elemIDs.size());
        vecVecDbl elemCenterCoords(elemIDs.size());
        // for each element of interest (its ID)
        for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
            // get its index in all elements from mesh
            uInt allElemIndex = elemIndexMap[elemIDs[elemIndex]];
            // resize element nodes to number of nodes
            elementNodes[elemIndex].resize(elemNodeIDs[allElemIndex].size());
            // assign node indexes
            for (uInt nodeIndex=0; nodeIndex<elementNodes[elemIndex].size(); nodeIndex++)
                elementNodes[elemIndex][nodeIndex] = nodeIndexMap[elemNodeIDs[allElemIndex][nodeIndex]];
            // calculate element center coordinates
            vecDbl centerCoord(3, 0.0);
            for (uInt nodeIndex=0; nodeIndex<elementNodes[elemIndex].size(); nodeIndex++){
                for (uInt i=0; i<3; i++)
                    centerCoord[i] += coordinates[nodeIndexMap[elemNodeIDs[allElemIndex][nodeIndex]]][i];
            }
            for (uInt i=0; i<3; i++)
                centerCoord[i] /= elementNodes[elemIndex].size();
            elemCenterCoords[elemIndex] = centerCoord;
        }

        // calculate weights
        vecVecDbl nodeWeights;
        if (weightFactors(
                nodeWeights,
                elementNodes,
                coordinates,
                elemCenterCoords))
            return;

        // calculate node field
        vecDbl nodeField(nodeIDs.size(), 0.0);
        if (weightAverage(
                nodeField,
                nodeWeights,
                elementNodes,
                fieldData))
            return;

        // fit data to 0.0-1.0 scale (take into account minimum and maximum limits)
        fitDataToUnit(nodeField, pProjectData->minLimit, pProjectData->maxLimit);

        // go through each element
        for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
            // try to find this primitive index
            map<uInt,uInt>::iterator it = mapIndexID.find(elemIDs[elemIndex]);
            if (it != mapIndexID.end()){
                // get node ID's of current element
                vecUInt nodes = elemNodeIDs[elemIndexMap[elemIDs[elemIndex]]];

                // store colors to m_resultColors member
                m_resultColors[it->second].resize(nodes.size());

                for (uInt elemNodeIndex=0; elemNodeIndex<nodes.size(); elemNodeIndex++){
                    // get index of current node
                    uInt index = nodeIndexMap[nodes[elemNodeIndex]];

                    // apply RGB filter
                    vecDbl RGBValues = RGBFilter(nodeField[index]);

                    // apply colors
                    m_resultColors[it->second][elemNodeIndex].resize(3);
                    for (uInt i=0; i<3; i++)
                        m_resultColors[it->second][elemNodeIndex][i] = RGBValues[i];
                }
            }
        }
    }
}

//-----------------------------------------------------------------------/
// update color
// should be called each time there is a change in colors
//  - change in mode... solid, neighbors, normal
//  - change in selections
//  - after call to updateMeshData
//  - after call to updateResultColors
//  - in neighboring mode for each refresh
//-----------------------------------------------------------------------/
int C3DViewer::updateColor(
        CMesh* pMesh,
        CProjectData* pProjectData,
        setUInt& elemSelection,
        setUInt& nodeSelection)
{
    if (pProjectData->surfResultsShown){
        // update colors with surface results
        if (m_primMap.size() != m_resultColors.size()){
            showErrorMessage("Internal error: serious... results don't match primitive size");
            return -1;
        }

        // go through each element
        uInt elemIndex = 0;
        for (mapPrimIt it=m_primMap.begin(); it!=m_primMap.end(); it++){
            // extract primitive from map iterator
            Prim* primitive = &it->second;
            // apply color from m_resultColors to primitive
            if (primitive->color.size() == m_resultColors[elemIndex].size()){
                for (uInt nodeIndex=0; nodeIndex<primitive->color.size(); nodeIndex++){
                    for (uInt i=0; i<3; i++)
                        primitive->color[nodeIndex][i] = m_resultColors[elemIndex][nodeIndex][i];
                }
            }
            elemIndex++;
        }
    }
    else{
        // update colors with default foreground value
        for (mapPrimIt it=m_primMap.begin(); it!=m_primMap.end(); it++){
            // extract primitive from map iterator
            Prim* primitive = &it->second;
            for (uInt nodeIndex=0; nodeIndex<primitive->color.size(); nodeIndex++){
                primitive->color[nodeIndex].resize(3);
                for (uInt i=0; i<3; i++)
                    primitive->color[nodeIndex][i] = colorFore[i];
            }
        }
    }

    // update vertex colors with default foreground value
    for (mapVrtxIt it=m_vrtxMap.begin(); it!=m_vrtxMap.end(); it++){
        // extract vertex from map iterator
        Vrtx* vertex = &it->second;
        for (uInt i=0; i<3; i++)
            vertex->color[i] = colorFore[i];
    }

    // update vertex colors if node selection is not empty
    if (!nodeSelection.empty()){
        for (setUIntIt it=nodeSelection.begin(); it!=nodeSelection.end(); it++){
            mapVrtxIt vrtxIt = m_vrtxMap.find(*it);
            if (vrtxIt != m_vrtxMap.end()){
                // extract vertex from map iterator
                Vrtx* vertex = &vrtxIt->second;
                for (uInt i=0; i<3; i++)
                    vertex->color[i] = colorSel[i];
            }
        }
    }

    // update primitive colors if element selection is not empty
    if (!elemSelection.empty()){
        for (setUIntIt it=elemSelection.begin(); it!=elemSelection.end(); it++){
            mapPrimIt primIt = m_primMap.find(*it);
            if (primIt != m_primMap.end()){
                // extract primitive from map iterator
                Prim* primitive = &primIt->second;
                for (uInt nodeIndex=0; nodeIndex<primitive->color.size(); nodeIndex++){
                    primitive->color[nodeIndex].resize(3);
                    for (uInt i=0; i<3; i++)
                        primitive->color[nodeIndex][i] = colorSel[i];
                }
            }
        }
    }

    if (pProjectData->displayMode==3 && m_neighElement){
        // update neighbors' colors
        vecVecUInt neighboring;
        pMesh->getNeighbors(m_neighElement, neighboring);
        for (uInt order=0; order<neighboring.size(); order++){
            for (uInt neighborIndex=0; neighborIndex<neighboring[order].size(); neighborIndex++){
                mapPrimIt it=m_primMap.find(neighboring[order][neighborIndex]);
                if (it != m_primMap.end()){
                    // extract primitive from map iterator
                    Prim* primitive = &it->second;
                    for (uInt nodeIndex=0; nodeIndex<primitive->color.size(); nodeIndex++){
                        primitive->color[nodeIndex].resize(3);
                        for (uInt i=0; i<3; i++)
                            primitive->color[nodeIndex][i] = neighborColors[order][i];
                    }
                }
            }
        }
    }

    return 0;
}

//-----------------------------------------------------------------------/
// update lighting condition
//-----------------------------------------------------------------------/
void C3DViewer::updateLighting(CProjectData* pProjectData)
{
    // update lighting
    if (pProjectData->showLights){
        glLightfv(GL_LIGHT0, GL_POSITION, light0_pos);
        glLightfv(GL_LIGHT0, GL_DIFFUSE,  light0_color);
        glLightfv(GL_LIGHT1, GL_POSITION, light1_pos);
        glLightfv(GL_LIGHT1, GL_DIFFUSE,  light1_color);
        glEnable(GL_LIGHT0);
        glEnable(GL_LIGHT1);
        glEnable(GL_LIGHTING);
    }
    else
        glDisable(GL_LIGHTING);
}

//-----------------------------------------------------------------------/
// refit model to 3D Viewer window
//-----------------------------------------------------------------------/
void C3DViewer::refit(CMesh* pMesh)
{
    // reset center to model center
    double* maxCoords = pMesh->getMaxBody();
    double* minCoords = pMesh->getMinBody();
    for (uInt i=0; i<3; i++)
        m_rotCenter[i] = (maxCoords[i] + minCoords[i])/2;

    // reset translations and scalings
    m_translation[0] = 0;
    m_translation[1] = 0;
    m_scale[0] = 2.0;
    m_scale[1] = 2.0;
}

//-----------------------------------------------------------------------/
// refresh/draw 3D viewer
//-----------------------------------------------------------------------/
void C3DViewer::refresh3DViewer(
        CMesh* pMesh,
        CProjectData* pProjectData,
        setUInt& elemSelection,
        setUInt& nodeSelection)
{
    wxGLCanvas::SetCurrent(*m_context);

    // clear screen before drawing new stuff
    glClearColor(colorBack[0], colorBack[1], colorBack[2], 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // get 3D viewer size
    wxSize sz(GetClientSize());

    // get perspective
    getPerspective(sz);

    // set perspective
    setPerspective();

    // check display mode
    if (pProjectData->displayMode == 0){
        // draw wireframe
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        drawPrimitives(false, colorBlack, true);
    }
    else if (pProjectData->displayMode == 1){
        // draw solid
        if (pProjectData->showEdges){
            // draw edges
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
            drawPrimitives(false, colorBlack, true);
            // draw fills
            glEnable(GL_POLYGON_OFFSET_FILL);
            glPolygonOffset(1.0, 1.0);
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
            drawPrimitives(false);
            glDisable(GL_POLYGON_OFFSET_FILL);
        }
        else{
            // draw only fills
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
            drawPrimitives(false);
        }
    }
    else if (pProjectData->displayMode == 2){
        // draw normal mode
        glEnable(GL_CULL_FACE);
        if (pProjectData->showEdges){
            // draw front
            glCullFace(GL_FRONT);
            // draw edges
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
            drawPrimitives(false, colorBlack, true);
            // draw fills
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
            glEnable(GL_POLYGON_OFFSET_FILL);
            glPolygonOffset(1.0, 1.0);
            drawPrimitives(false, colorBack);
            glDisable(GL_POLYGON_OFFSET_FILL);

            // draw back
            glCullFace(GL_BACK);
            // draw edges
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
            drawPrimitives(false, colorBlack, true);
            // draw fills
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
            glEnable(GL_POLYGON_OFFSET_FILL);
            glPolygonOffset(1.0, 1.0);
            drawPrimitives(false, colorFore);
            glDisable(GL_POLYGON_OFFSET_FILL);
        }
        else{
            // draw front fills
            glCullFace(GL_FRONT);
            glPolygonMode(GL_FRONT, GL_FILL);
            drawPrimitives(false, colorBack);

            // draw back fills
            glCullFace(GL_BACK);
            glPolygonMode(GL_BACK, GL_FILL);
            drawPrimitives(false, colorFore);
        }
        glDisable(GL_CULL_FACE);
    }
    else if (pProjectData->displayMode == 3){
        // update colors (neighboring elements)
        if (m_neighElement)
            updateColor(pMesh, pProjectData, elemSelection, nodeSelection);
        if (pProjectData->showEdges){
            // draw edges
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
            drawPrimitives(false, colorBlack, true);
            // draw fills
            glEnable(GL_POLYGON_OFFSET_FILL);
            glPolygonOffset(1.0, 1.0);
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
            drawPrimitives(false);
            glDisable(GL_POLYGON_OFFSET_FILL);
        }
        else{
            // draw only fills
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
            drawPrimitives(false);
        }
    }

    // draw nodes
    if (pProjectData->showNodes)
        drawVertices(false);

    // draw coordinate system
    if (pProjectData->showCoordSys)
        drawCoordinateSystem();

    // draw node IDs
    if (pProjectData->showNodeIDs)
        drawNodeIDs();

    // draw element IDs
    if (pProjectData->showElemIDs)
        drawElemIDs();

    // draw selection box
    if (m_boxSelectionMode)
        drawSelectionBox();

    // draw legend
    if (pProjectData->showLegend)
        drawLegend(pProjectData);

    glFlush();

    SwapBuffers();
}

//-----------------------------------------------------------------------/
// set orientation of model
//-----------------------------------------------------------------------/
void C3DViewer::setViewToFront() { m_quaternions[0] = -0.5;          m_quaternions[1] = 0.5;          m_quaternions[2] = 0.5;          m_quaternions[3] = -0.5;         }
void C3DViewer::setViewToTop()   { m_quaternions[0] =  0.0;          m_quaternions[1] = 0.0;          m_quaternions[2] = 0.0;          m_quaternions[3] = 1.0;          }
void C3DViewer::setViewToRight() { m_quaternions[0] = 0.0;           m_quaternions[1] = sqrt(2.0)/2.0;m_quaternions[2] = sqrt(2.0)/2.0;m_quaternions[3] = 0.0;          }
void C3DViewer::setViewToBack()  { m_quaternions[0] = 0.5;           m_quaternions[1] = 0.5;          m_quaternions[2] = 0.5;          m_quaternions[3] = 0.5;          }
void C3DViewer::setViewToBottom(){ m_quaternions[0] = 1.0;           m_quaternions[1] = 0.0;          m_quaternions[2] = 0.0;          m_quaternions[3] = 0.0;          }
void C3DViewer::setViewToLeft()  { m_quaternions[0] =  sqrt(2.0)/2.0;m_quaternions[1] = 0.0;          m_quaternions[2] = 0.0;          m_quaternions[3] = sqrt(2.0)/2.0;}
void C3DViewer::setViewToIso()   { m_quaternions[0] = -sqrt(3.0)/4.0;m_quaternions[1] = 0.25;         m_quaternions[2] = sqrt(3.0)/4.0;m_quaternions[3] = -0.75;        }

//-----------------------------------------------------------------------/
// perform left mouse button operations
//-----------------------------------------------------------------------/
void C3DViewer::onLeftMouseButton(
        wxMouseEvent& event,
        setUInt& elemSelection,
        setUInt& nodeSelection)
{
    // get mouse position
    double mouseX = event.GetX();
    double mouseY = event.GetY();

    // exit if control button is not pressed
    if (!event.ControlDown()){
        // disable box selection mode and flag LMB is not down
        m_boxSelectionMode = false;
        m_leftMouseDown = false;

        if (event.LeftDown()){
            // store mouse locations for translation or rotation
            m_mousePosition[0] = mouseX;
            m_mousePosition[1] = mouseY;
        }

        return;
    }

    // if left mouse button is clicked down
    if (event.LeftDown()){
        // store mouse position
        m_selectionCoords[0] = mouseX;
        m_selectionCoords[1] = mouseY;

        // declare that mouse down is clicked
        m_leftMouseDown = true;

        return;
    }
    // if left mouse button is released
    else if (event.LeftUp()){
        // flag LMB is not down
        m_leftMouseDown = false;

        // prepare selection "things"... note 131072 for buffer size... crash is happening if buffer larger
        const uInt selSize = 131072;
        GLuint buffer[selSize];
        glSelectBuffer(selSize, buffer);
        glRenderMode(GL_SELECT);
        glInitNames();
        glPushName(0);

        // get 3D viewer size
        wxSize sz(GetClientSize());

        // get perspective
        getPerspective(sz);

        // get mouse position
        double mouseX = event.GetX();
        double mouseY = event.GetY();

        // check if picking or selection mode
        if (m_boxSelectionMode){
            // disable box selection mode
            m_boxSelectionMode = false;

            // modify perspective for selection
            bool addition = false; // bool to hold info if addition or removal of elements is requested
            modifyPerspectiveForSelection(addition, mouseX, mouseY, sz);

            // set perspective
            setPerspective();

            // render primitives or vertices
            if (m_selectionMode == 0) drawPrimitives(true);
            else                      drawVertices  (true);

            glMatrixMode(GL_PROJECTION);
            glPopMatrix();
            glFlush();

            // process selection hits
            vecUInt hitIDs;
            processSelection(buffer, selSize, hitIDs);
            for (uInt i=0; i<hitIDs.size(); i++){
                if (addition){
                    if (m_selectionMode == 0) elemSelection.insert(hitIDs[i]);
                    else                      nodeSelection.insert(hitIDs[i]);
                }
                else{
                    if (m_selectionMode == 0) elemSelection.erase(hitIDs[i]);
                    else                      nodeSelection.erase(hitIDs[i]);
                }
            }
        }
        else{
            // modify perspective for picking
            modifyPerspectiveForPick(mouseX, mouseY, sz);

            // set perspective
            setPerspective();

            // render primitives or vertices
            if (m_selectionMode == 0) drawPrimitives(true);
            else                      drawVertices  (true);

            glMatrixMode(GL_PROJECTION);
            glPopMatrix();
            glFlush();

            // process picking hits
            int pickID;
            processPick(buffer, selSize, pickID);
            if (pickID > 0){
                if (m_selectionMode == 0){
                    if (elemSelection.find(pickID) == elemSelection.end()) elemSelection.insert(pickID);
                    else                                                   elemSelection.erase (pickID);
                }
                else{
                    if (nodeSelection.find(pickID) == nodeSelection.end()) nodeSelection.insert(pickID);
                    else                                                   nodeSelection.erase (pickID);
                }
            }
        }
    }
}

//-----------------------------------------------------------------------/
// perform middle mouse button up operations
//-----------------------------------------------------------------------/
void C3DViewer::onMouseMiddleUp(
        wxMouseEvent& event,
        CMesh* pMesh)
{
    // cancel if movement occurred
    if (!m_waitForRotCenter)
        return;

    // prepare selection "things"... note 131072 for buffer size... crash is happening if buffer larger
    const uInt selSize = 131072;
    GLuint buffer[selSize];
    glSelectBuffer(selSize, buffer);
    glRenderMode(GL_SELECT);
    glInitNames();
    glPushName(0);

    // get 3D viewer size
    wxSize sz(GetClientSize());

    // get perspective
    getPerspective(sz);

    // get mouse position
    double mouseX = event.GetX();
    double mouseY = event.GetY();

    // modify perspective for picking
    modifyPerspectiveForPick(mouseX, mouseY, sz);

    // set perspective
    setPerspective();

    // render primitives or vertices
    if (m_selectionMode == 0) drawPrimitives(true);
    else                      drawVertices  (true);

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glFlush();

    // get ID of primitive or vertex
    int pickID = 0;
    processPick(buffer, selSize, pickID);

    // exit if no primitives selected
    if (pickID == 0)
        return;

    // center on picked entity
    if (m_selectionMode == 0){
        // center on element
        vecVecDbl coordinates;
        pMesh->getElemCoords(pickID, coordinates);

        // get average of coordinates and store as rotation centerCoord
        for (uInt i=0; i<3; i++)
            m_rotCenter[i] = 0.0;
        for (uInt i=0; i<coordinates.size(); i++){
            for (uInt j=0; j<3; j++)
                m_rotCenter[j] += coordinates[i][j];
        }
        for (uInt i=0; i<3; i++)
            m_rotCenter[i] /= coordinates.size();
    }
    else{
        // center on node
        double coordinates[3];
        pMesh->getNodeCoords(pickID, coordinates);

        // store coordinates as rotation center
        m_rotCenter[0] = coordinates[0];
        m_rotCenter[1] = coordinates[1];
        m_rotCenter[2] = coordinates[2];
    }

    // reset translations
    m_translation[0] = 0.0;
    m_translation[1] = 0.0;

    return;
}

//-----------------------------------------------------------------------/
// perform middle mouse button down operations
//-----------------------------------------------------------------------/
void C3DViewer::onMouseMiddleDown(wxMouseEvent& event)
{
    // flag to wait for MMB up
    m_waitForRotCenter = true;

    // store mouse location for translation
    m_mousePosition[0] = event.GetX();
    m_mousePosition[1] = event.GetY();
}

//-----------------------------------------------------------------------/
// perform mouse wheel operations
//-----------------------------------------------------------------------/
void C3DViewer::onMouseWheel(wxMouseEvent& event)
{
    // calculate ratio for scaling
    double ratio = double(event.GetWheelRotation())/100.0;

    if (ratio < 0){
        ratio *= -1;
        m_scale[0] = m_scale[0]*ratio;
        m_scale[1] = m_scale[1]*ratio;
    }
    else{
        m_scale[0] = m_scale[0]/ratio;
        m_scale[1] = m_scale[1]/ratio;
    }
}

//-----------------------------------------------------------------------/
// perform mouse move operations
//-----------------------------------------------------------------------/
void C3DViewer::onMouseMove(
        wxMouseEvent& event,
        CProjectData* pProjectData)
{
    // cancel waiting for MMB up
    m_waitForRotCenter = false;

    // get size of window
    wxSize sz(GetClientSize());

    // if left mouse button and control button are pressed, enter box selection mode
    if (m_leftMouseDown && event.ControlDown()){
        m_boxSelectionMode = true;
        m_selectionCoords[2] = event.GetX();
        m_selectionCoords[3] = event.GetY();
    }
    else{
        // if only left button is pressed -> rotation
        if (event.LeftIsDown() && !event.MiddleIsDown() && !event.RightIsDown()){
            // drag in progress so simulate trackball, but before check if previous mouse position
            // is on the top left corner
            if (m_mousePosition[0] && m_mousePosition[1]){
                float spin_quat[4];
                trackball(spin_quat,
                         (-sz.x + 2.0*m_mousePosition[0]) / sz.x,
                         ( sz.y - 2.0*m_mousePosition[1]) / sz.y,
                         (-sz.x + 2.0*event.GetX())       / sz.x,
                         ( sz.y - 2.0*event.GetY())       / sz.y);

                add_quats(spin_quat, m_quaternions, m_quaternions);

                // store mouse locations for rotation
                m_mousePosition[0] = event.GetX();
                m_mousePosition[1] = event.GetY();
            }
        }
        // if only middle button is pressed -> translation
        else if (!event.LeftIsDown() && event.MiddleIsDown() && !event.RightIsDown()){
            // calculate difference due to movement
            double diffX = event.GetX()-m_mousePosition[0];
            double diffY = event.GetY()-m_mousePosition[1];

            // calculate translation distance
            m_translation[0] = m_translation[0]+diffX/sz.x*4;
            m_translation[1] = m_translation[1]-diffY/sz.y*4;

            // store mouse locations for translation
            m_mousePosition[0] = event.GetX();
            m_mousePosition[1] = event.GetY();
        }
        // if neighbors checking mode
        else if(pProjectData->displayMode == 3){
            // prepare selection "things"... note 131072 for buffer size... crash is happening if buffer larger
            const uInt selSize = 131072;
            GLuint buffer[selSize];
            glSelectBuffer(selSize, buffer);
            glRenderMode(GL_SELECT);
            glInitNames();
            glPushName(0);

            // get 3D viewer size
            wxSize sz(GetClientSize());

            // get perspective
            getPerspective(sz);

            // get mouse position
            double mouseX = event.GetX();
            double mouseY = event.GetY();

            // modify perspective for picking
            modifyPerspectiveForPick(mouseX, mouseY, sz);

            // set perspective
            setPerspective();

            // render elements
            drawPrimitives(true);

            glMatrixMode(GL_PROJECTION);
            glPopMatrix();
            glFlush();

            // process picking hits for primitive under mouse cursor
            int primitiveID = 0;
            processPick(buffer, selSize, primitiveID);
            if (primitiveID < 0) m_neighElement = 0;
            else                 m_neighElement = primitiveID;
        }

        // disable box selection mode and flag LMB is not down
        m_boxSelectionMode = false;
        m_leftMouseDown = false;
    }
}

//-----------------------------------------------------------------------/
// draw primitives (elements)
//-----------------------------------------------------------------------/
void C3DViewer::drawPrimitives(
        bool drawForSelection,
        GLdouble* color,
        bool drawForLines)
{
    // go through complete primitive map
    for (mapPrimConstIt it=m_primMap.begin(); it!=m_primMap.end(); it++){
        // extract primitive from map iterator
        const Prim* primitive = &it->second;

        // load name of primitive for selection mode
        if (drawForSelection) glLoadName(primitive->id);

        // check if quadrilateral or triangular primitive
        if   (primitive->coordinates.size() == 3) glBegin(GL_TRIANGLES);
        else                                      glBegin(GL_QUADS);

        // give normal information
        glNormal3dv(primitive->normal);

        // draw primitive
        for (uInt nodeIndex=0; nodeIndex<primitive->coordinates.size(); nodeIndex++){
            // give colors
            if (color){
                if (drawForLines)
                    glColor3dv(color);
                else{
                    GLdouble mixedColor[3];
                    for (uInt i=0; i<3; i++)
                        mixedColor[i] = (color[i]+primitive->color[nodeIndex][i])/2;
                    glColor3dv(mixedColor);
                }
            }
            else
                glColor3dv(&(primitive->color[nodeIndex][0]));
            // give coordinates
            glVertex3dv(&(primitive->coordinates[nodeIndex][0]));
        }
        glEnd();
    }
}

//-----------------------------------------------------------------------/
// draw vertices (nodes)
//-----------------------------------------------------------------------/
void C3DViewer::drawVertices(bool drawForSelection)
{
    // go through complete vertex map
    for (mapVrtxConstIt it=m_vrtxMap.begin(); it!=m_vrtxMap.end(); it++){
        // extract vertex from map iterator
        const Vrtx* vertex = &it->second;

        // load name of vertex for selection mode
        if (drawForSelection) glLoadName(vertex->id);

        // draw vertex
        glPointSize(5);
        glBegin(GL_POINTS);
        glColor3dv(vertex->color);
        glVertex3dv(vertex->coordinates);
        glEnd();
    }
}

//-----------------------------------------------------------------------/
// draw coordinate system
//-----------------------------------------------------------------------/
void C3DViewer::drawCoordinateSystem()
{
    // get body extents (10% larger)
    double extendX = m_maxBody[0]*1.1;
    double extendY = m_maxBody[1]*1.1;
    double extendZ = m_maxBody[2]*1.1;

    double maxExtent = max(max(extendX, extendY), extendZ);
    if (maxExtent < -1.0e19)
        maxExtent = 1;

    // correct axes that are too small (zero)
    double maxExtentTenPerc = maxExtent*0.1;
    if (extendX < maxExtentTenPerc) extendX = maxExtent/2;
    if (extendY < maxExtentTenPerc) extendY = maxExtent/2;
    if (extendZ < maxExtentTenPerc) extendZ = maxExtent/2;

    // use black color for both axis and labels
    glColor3d(0, 0, 0);

    // draw axis
    glBegin(GL_LINES);
    glVertex3d(0,0,0); glVertex3d(extendX,0,0); // X axis
    glVertex3d(0,0,0); glVertex3d(0,extendY,0); // Y axis
    glVertex3d(0,0,0); glVertex3d(0,0,extendZ); // Z axis
    glEnd();

    // draw labels
    glPixelStorei (GL_UNPACK_ALIGNMENT, 1);
    maxExtent *= 0.02;
    glRasterPos3d(extendX+maxExtent,0,0); printString("X");
    glRasterPos3d(0,extendY+maxExtent,0); printString("Y");
    glRasterPos3d(0,0,extendZ+maxExtent); printString("Z");
}

//-----------------------------------------------------------------------/
// draw node IDs
//-----------------------------------------------------------------------/
void C3DViewer::drawNodeIDs()
{
    // use black color for labels
    glColor3d(0, 0, 0);

    // draw labels
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    for (mapVrtxIt vrtxIt=m_vrtxMap.begin(); vrtxIt!=m_vrtxMap.end(); vrtxIt++){
        glRasterPos3d(vrtxIt->second.coordinates[0],
                      vrtxIt->second.coordinates[1],
                      vrtxIt->second.coordinates[2]);
        printString(int2Str(vrtxIt->second.id));
    }
}

//-----------------------------------------------------------------------/
// draw element IDs
//-----------------------------------------------------------------------/
void C3DViewer::drawElemIDs()
{
    // use black color for labels
    glColor3d(0, 0, 0);

    // draw labels
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    for (mapPrimIt primIt=m_primMap.begin(); primIt!=m_primMap.end(); primIt++){
        vecDbl coords(3, 0.0);
        for (uInt nodeIndex=0; nodeIndex<primIt->second.coordinates.size(); nodeIndex++){
            for (uInt i=0; i<3; i++)
                coords[i] += primIt->second.coordinates[nodeIndex][i];
        }
        for (uInt i=0; i<3; i++)
            coords[i] /= primIt->second.coordinates.size();

        glRasterPos3d(coords[0], coords[1], coords[2]);
        printString(int2Str(primIt->second.id));
    }
}

//-----------------------------------------------------------------------/
// draw selection box
//-----------------------------------------------------------------------/
void C3DViewer::drawSelectionBox()
{
    // use black color
    glColor3d(0, 0, 0);

    // get viewport size
    int viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);

    glPushMatrix();
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(0.0, viewport[2], 0.0, viewport[3], 0, 1);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // get box corners
    int sx = m_selectionCoords[0];
    int ex = m_selectionCoords[2];
    int sy = viewport[3]-m_selectionCoords[1];
    int ey = viewport[3]-m_selectionCoords[3];

    // draw box
    glBegin(GL_LINES);
    glVertex2i(sx, sy);
    glVertex2i(sx, ey);
    glVertex2i(sx, ey);
    glVertex2i(ex, ey);
    glVertex2i(ex, ey);
    glVertex2i(ex, sy);
    glVertex2i(ex, sy);
    glVertex2i(sx, sy);
    glEnd();

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glPopMatrix();
}

//-----------------------------------------------------------------------/
// draw legend
//-----------------------------------------------------------------------/
void C3DViewer::drawLegend(CProjectData* pProjectData)
{
    // position and size of legend
    int sx = 10;
    int sy = 50;
    int ex = 30;
    int ey = 400;

    // get viewport size
    int viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);

    glPushMatrix();
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(0.0, viewport[2], 0.0, viewport[3], 0, 1);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // write field name
    glColor3d(0, 0, 0);
    glRasterPos2i(sx, viewport[3]-sy+23); printString(pProjectData->selectedData);
    glRasterPos2i(sx, viewport[3]-sy+5 ); printString(pProjectData->selectedCase);

    // draw legend
    uInt numDivision = 20;
    double step = 1.0/double(numDivision-1);
    double dataStep = (pProjectData->maxLimit-pProjectData->minLimit)*step;
    double divisionPlus = ey/(numDivision+1);
    for (uInt i=0; i<numDivision; i++){
        glBegin(GL_QUADS);

        vecDbl colors = RGBFilter(i*step);
        glColor3dv(&colors[0]);

        double iDivPlus = i*divisionPlus;
        double iDivPlusPlus = iDivPlus+divisionPlus;
        glVertex2i(sx, viewport[3]-(sy+iDivPlus));
        glVertex2i(sx, viewport[3]-(sy+iDivPlusPlus));
        glVertex2i(ex, viewport[3]-(sy+iDivPlusPlus));
        glVertex2i(ex, viewport[3]-(sy+iDivPlus));

        glEnd();
        glColor3d(0, 0, 0);

        glRasterPos2i(ex+4, viewport[3]-(sy+divisionPlus*(i+0.5)+6));
        printString(dbl2Str(pProjectData->minLimit+dataStep*i));
    }

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glPopMatrix();
}

//-----------------------------------------------------------------------/
// make raster font
//-----------------------------------------------------------------------/
void C3DViewer::makeRasterFont()
{
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    fontOffset = glGenLists(128);
    GLsizei width = 8;
    GLsizei height = 13;

    // upper case letters
    for (GLuint i=0,j='A'; i<26; i++,j++) {
        glNewList(fontOffset+j, GL_COMPILE);
        glBitmap(width, height, 0.0, 2.0, 10.0, 0.0, letters[i]);
        glEndList();
    }

    // lower case letters
    for (GLuint i=0,j='a'; i<26; i++,j++) {
        glNewList(fontOffset+j, GL_COMPILE);
        glBitmap(width, height, 0.0, 2.0, 10.0, 0.0, lowletters[i]);
        glEndList();
    }

    // numbers and some more
    for (GLuint i=0,j='0'; i<16; i++,j++) {
        glNewList(fontOffset+j, GL_COMPILE);
        glBitmap(width, height, 0.0, 2.0, 10.0, 0.0, numletters[i]);
        glEndList();
    }

    // space and some more
    for (GLuint i=0,j=' '; i<16; i++,j++) {
        glNewList(fontOffset+j, GL_COMPILE);
        glBitmap(width, height, 0.0, 2.0, 10.0, 0.0, specialletters[i]);
        glEndList();
    }
}

//-----------------------------------------------------------------------/
// get perspective
//-----------------------------------------------------------------------/
void C3DViewer::getPerspective(wxSize& sz)
{
    // get model diameter (maximum dimension of model nodes)
    double modelDiam = max(max(max(abs(m_maxNode[0]), abs(m_minNode[0])) ,
                               max(abs(m_maxNode[1]), abs(m_minNode[1])) ),
                               max(abs(m_maxNode[2]), abs(m_minNode[2])) );
    if (modelDiam > 1.0e+19 || // if no nodes
        modelDiam < 1.0e-19 )  // if single node resulting in zero bound
        modelDiam = 1.0;

    // get half diameter of body only
    double bodyDiam = max(max(m_maxBody[0]-m_minBody[0] ,
                              m_maxBody[1]-m_minBody[1] ),
                              m_maxBody[2]-m_minBody[2] )/2;
    if (bodyDiam < 1.0e-19) // if no elements
        bodyDiam = modelDiam;

    // calculate window aspect ratio
    GLdouble aspect = (GLdouble)sz.x/sz.y;

    // calculate model bounds with window aspect ratio, translate values, scale value and model diameter
    if (aspect < 1.0){
        // window taller than wide
        m_corners[0] = (-2-m_translation[0])*bodyDiam       /m_scale[0];
        m_corners[1] = ( 2-m_translation[0])*bodyDiam       /m_scale[0];
        m_corners[2] = (-2-m_translation[1])*bodyDiam/aspect/m_scale[1];
        m_corners[3] = ( 2-m_translation[1])*bodyDiam/aspect/m_scale[1];
    }
    else{
        // window wider than tall
        m_corners[0] = (-2-m_translation[0])*bodyDiam*aspect/m_scale[0];
        m_corners[1] = ( 2-m_translation[0])*bodyDiam*aspect/m_scale[0];
        m_corners[2] = (-2-m_translation[1])*bodyDiam       /m_scale[1];
        m_corners[3] = ( 2-m_translation[1])*bodyDiam       /m_scale[1];
    }

    // 3D Viewer window depth
    m_corners[5] = 2*modelDiam;
    m_corners[4] = -m_corners[5];
}

//-----------------------------------------------------------------------/
// set perspective
//-----------------------------------------------------------------------/
void C3DViewer::setPerspective()
{
    // set 3D viewer window
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(m_corners[0],
            m_corners[1],
            m_corners[2],
            m_corners[3],
            m_corners[4],
            m_corners[5]);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // rotate with trackball
    GLfloat m[4][4];
    build_rotmatrix(m, m_quaternions);
    glMultMatrixf(&m[0][0]);

    // translate to user's center
    glTranslated(-m_rotCenter[0], -m_rotCenter[1], -m_rotCenter[2]);
}

//-----------------------------------------------------------------------/
// modify perspective for pick
//-----------------------------------------------------------------------/
void C3DViewer::modifyPerspectiveForPick(
        double& mouseX,
        double& mouseY,
        wxSize& sz)
{
    // get screen on mouse position
    double varX = (m_corners[1]-m_corners[0])*0.01;
    double varY = (m_corners[3]-m_corners[2])*0.01;
    if (m_selectionMode == 0){
        varX *= 0.1;
        varY *= 0.1;
    }
    double posX = m_corners[0]+(m_corners[1]-m_corners[0])*mouseX/sz.x;
    double posY = m_corners[3]+(m_corners[2]-m_corners[3])*mouseY/sz.y;
    m_corners[0] = posX-varX;
    m_corners[1] = posX+varX;
    m_corners[2] = posY-varY;
    m_corners[3] = posY+varY;
}

//-----------------------------------------------------------------------/
// modify perspective for selection
//-----------------------------------------------------------------------/
void C3DViewer::modifyPerspectiveForSelection(
        bool& addition,
        double& mouseX,
        double& mouseY,
        wxSize& sz)
{
    // reduce horizontal window
    double width = m_corners[1]-m_corners[0];
    if (mouseX > m_selectionCoords[0]){
        addition = true;
        m_corners[1] = m_corners[0]+width*mouseX              /sz.x;
        m_corners[0] = m_corners[0]+width*m_selectionCoords[0]/sz.x;
    }
    else{
        addition = false;
        m_corners[1] = m_corners[0]+width*m_selectionCoords[0]/sz.x;
        m_corners[0] = m_corners[0]+width*mouseX              /sz.x;
    }

    // reduce vertical window
    double height = m_corners[3]-m_corners[2];
    if (mouseY > m_selectionCoords[1]){
        m_corners[2] = m_corners[3]-height*m_selectionCoords[1]/sz.y;
        m_corners[3] = m_corners[3]-height*mouseY              /sz.y;
    }
    else{
        m_corners[2] = m_corners[3]-height*mouseY              /sz.y;
        m_corners[3] = m_corners[3]-height*m_selectionCoords[1]/sz.y;
    }
}

//-----------------------------------------------------------------------/
// process pick
//-----------------------------------------------------------------------/
void C3DViewer::processPick(
        GLuint* buffer,
        const uInt& selectionSize,
        int& pickID)
{
    // get selection hits
    GLuint numberOfHits = 0;
    numberOfHits = glRenderMode(GL_RENDER);

    // exit if no hits
    if (!numberOfHits)
        return;

    // limit too large selection
    else if (numberOfHits > selectionSize)
        numberOfHits = selectionSize;

    // process hits for primitive IDs and z coordinate
    GLuint* hitPointer = (GLuint*)buffer;
    map<float,int> zMap;
    for (GLuint hitIndex=0; hitIndex<numberOfHits; hitIndex++){
        GLuint numOfNamesForHit = *hitPointer++;
        float z1 = (float)*hitPointer++/0x7fffffff;
        float z2 = (float)*hitPointer++/0x7fffffff;
        zMap.insert(pair<float,int>((z1+z2)/2, *hitPointer));
        hitPointer += numOfNamesForHit;
    }

    // extract primitive with smallest z coordinate
    if (!zMap.empty()) pickID = zMap.begin()->second;
    else               pickID = -1;

    return;
}

//-----------------------------------------------------------------------/
// process selection
//-----------------------------------------------------------------------/
void C3DViewer::processSelection(
        GLuint* buffer,
        uInt selSize,
        vecUInt& hitIDs)
{
    // get selection hits
    GLuint numberOfHits = 0;
    numberOfHits = glRenderMode(GL_RENDER);

    if (numberOfHits == 0)
        return;
    else if(numberOfHits > selSize)
        numberOfHits = selSize;

    // exit if zero box
    if (m_corners[0]==m_corners[1] || m_corners[2]==m_corners[3])
        return;

    // process hits for IDs
    GLuint* hitPointer = (GLuint*)buffer;
    hitIDs.resize(numberOfHits);
    for (GLuint i=0; i<numberOfHits; i++){
        GLuint numOfNamesForHit = *hitPointer;
        hitPointer++;
        hitPointer++;
        hitPointer++;
        hitIDs[i] = *hitPointer;
        for (GLuint j=0; j<numOfNamesForHit; j++)
            hitPointer++;
    }
}

//-----------------------------------------------------------------------/
// display text on screen
//-----------------------------------------------------------------------/
void C3DViewer::printString(const string& textToPrint)
{
   glPushAttrib(GL_LIST_BIT);
   glListBase(fontOffset);
   glCallLists(textToPrint.size(), GL_UNSIGNED_BYTE, (GLubyte*)textToPrint.c_str());
   glPopAttrib();
}

//-----------------------------------------------------------------------/
// calculate normal
//-----------------------------------------------------------------------/
int C3DViewer::calcNormal(
        const vecVecDbl& coordinates,
        double normal[3])
{
    // calculating diagonal vectors
    double a[3], b[3];
                                       for (uInt i=0; i<3; i++) a[i] = coordinates[2][i]-coordinates[0][i];
    if      (coordinates.size() == 4){ for (uInt i=0; i<3; i++) b[i] = coordinates[3][i]-coordinates[1][i]; }
    else if (coordinates.size() == 3){ for (uInt i=0; i<3; i++) b[i] = coordinates[2][i]-coordinates[1][i]; }

    // calculating vector normal to diagonals (vector product)
    normal[0] = a[1]*b[2] - a[2]*b[1];
    normal[1] = a[2]*b[0] - a[0]*b[2];
    normal[2] = a[0]*b[1] - a[1]*b[0];

    // calculating modulus of that vector
    double normalMod = sqrt(pow(normal[0], 2) +
                            pow(normal[1], 2) +
                            pow(normal[2], 2) );
    if (normalMod == 0.0)
        return -1;

    // normalize normal
    for (uInt i=0; i<3; i++)
        normal[i] /= normalMod;

    return 0;
}

//-----------------------------------------------------------------------/
// fit data to 0-1 range
//-----------------------------------------------------------------------/
void C3DViewer::fitDataToUnit(
        vecDbl& fieldData,
        const double& min,
        const double& max)
{
    double range = max-min;
    for (uInt i=0; i<fieldData.size(); i++){
        if (fieldData[i] < min) fieldData[i] = min;
        if (fieldData[i] > max) fieldData[i] = max;
        fieldData[i] = (fieldData[i]-min)/range;
    }
}

//-----------------------------------------------------------------------/
// convert value (in 0-1 range) to RGB values
//-----------------------------------------------------------------------/
vecDbl C3DViewer::RGBFilter(const double& value)
{
    // declare RGB vector and fill with zeros
    vecDbl RGBValues(3, 0.0);

    // RGB channels sinusoid shift
    const double Rshift = -pi;
    const double Gshift = -pi/2;
    const double Bshift = 0;

    if      (value < 0.25) { RGBValues[1] = (sin(value*2*pi+Gshift)+1)/2;
                             RGBValues[2] = (sin(value*2*pi+Bshift)+1)/2; }
    else if (value > 0.75) { RGBValues[0] = (sin(value*2*pi+Rshift)+1)/2;
                             RGBValues[1] = (sin(value*2*pi+Gshift)+1)/2; }
    else                   { RGBValues[0] = (sin(value*2*pi+Rshift)+1)/2;
                             RGBValues[1] = (sin(value*2*pi+Gshift)+1)/2;
                             RGBValues[2] = (sin(value*2*pi+Bshift)+1)/2; }

    return RGBValues;
}
