#include "CGUIData.h"

//-----------------------------------------------------------------------/
// initializes data
//-----------------------------------------------------------------------/
CGUIData::CGUIData()
{
    // general data
    apameSolverPath   = "apame";
    useTempDir        = 0;
    logWinRefreshRate = 300;
    width             = 800;
    height            = 600;
    maximized         = 0;
    // neighbors dialog data
    neighThreshold    = 100.0;
    // wake dialog data
    wakeThreshold     = 60.0;
    wakeLength        = 10.0;
    wakeNum           = 10;

    // read GUI data from configuration file
    readGUIData();

    // get current working directory
    currentDir = wxs2Str(wxGetCwd());
}

//-----------------------------------------------------------------------/
// reads GUI data from configuration file
//  - C:\Users\{username}\AppData\Roaming on windows 7
//  - /home/{username} on Linux
//-----------------------------------------------------------------------/
void CGUIData::readGUIData()
{
    // get configuration file location
    string location = wxs2Str(wxStandardPaths::Get().GetUserConfigDir());

    // assemble file name
    string fileName = location+"/"+"apame_"+int2Str(apameGUIVersion)+"_"+int2Str(apameGUISubVersion)+".cfg";

    // check if configuration file exists
    if (!fileExists(fileName))
        return;

    // configuration file exists so read data from it
    ifstream configFile(fileName.c_str());
    if (!configFile.is_open()){
        showErrorMessage("Cannot access ApameGUI configuration file\nCheck permissions of "+fileName+"!");
        return;
    }

    // read lines
    vecStr fileLines;
    string line;
    while (getLine(configFile, line))
        fileLines.push_back(line);
    configFile.close();
    
    // read lines
    for(uInt i=0; i<fileLines.size(); i++){
        if      (fileLines[i] == "apameSolverPath")   readString (apameSolverPath,   i, fileLines);
        else if (fileLines[i] == "useTempDir")        readInteger(useTempDir,        i, fileLines);
        else if (fileLines[i] == "tempDir")           readString (tempDir,           i, fileLines);
        else if (fileLines[i] == "logWinRefreshRate") readInteger(logWinRefreshRate, i, fileLines);
        else if (fileLines[i] == "width")             readInteger(width,             i, fileLines);
        else if (fileLines[i] == "height")            readInteger(height,            i, fileLines);
        else if (fileLines[i] == "maximized")         readInteger(maximized,         i, fileLines);
        else if (fileLines[i] == "neighThreshold")    readDouble (neighThreshold,    i, fileLines);
        else if (fileLines[i] == "wakeThreshold")     readDouble (wakeThreshold,     i, fileLines);
        else if (fileLines[i] == "wakeLength")        readDouble (wakeLength,        i, fileLines);
        else if (fileLines[i] == "wakeNum")           readInteger(wakeNum,           i, fileLines);
    }
}

//-----------------------------------------------------------------------/
// stores GUI data to configuration file
//  - C:\Users\{username}\AppData\Roaming on windows 7
//  - /home/{username} on Linux
//-----------------------------------------------------------------------/
void CGUIData::storeGUIData()
{
    // get configuration file location
    string location = wxs2Str(wxStandardPaths::Get().GetUserConfigDir());

    // assemble file name
    string fileName = location+"/"+"apame_"+int2Str(apameGUIVersion)+"_"+int2Str(apameGUISubVersion)+".cfg";

    // check if configuration file exists and delete it
    if (fileExists(fileName)){
        if (remove(fileName.c_str()) != 0){
            showErrorMessage("Cannot delete existing ApameGUI configuration file\nCheck permissions of "+fileName+"!");
            return;
        }
    }

    // open configuration file for writing
    ofstream cfgFile(fileName.c_str());
    if (!cfgFile.is_open()){
        showErrorMessage("Cannot create ApameGUI configuration file\nCheck permissions of "+fileName+"!");
        return;
    }

    // write GUI data to configuration file
    cfgFile << "ApameGUI configuration file"    << endl;
    cfgFile << "apameGUIVersion"                << endl << apameGUIVersion              << endl;
    cfgFile << "apameGUISubVersion"             << endl << apameGUISubVersion           << endl;
    cfgFile << "apameGUIVersionDate"            << endl << apameGUIVersionDate          << endl;
    cfgFile << "apameSolverPath"                << endl << apameSolverPath              << endl;
    cfgFile << "useTempDir"                     << endl << int2Str(useTempDir)          << endl;
    cfgFile << "tempDir"                        << endl << tempDir                      << endl;
    cfgFile << "logWinRefreshRate"              << endl << int2Str(logWinRefreshRate)   << endl;
    cfgFile << "width"                          << endl << int2Str(width)               << endl;
    cfgFile << "height"                         << endl << int2Str(height)              << endl;
    cfgFile << "maximized"                      << endl << int2Str(maximized)           << endl;
    cfgFile << "neighThreshold"                 << endl << dbl2Str(neighThreshold)      << endl;
    cfgFile << "wakeThreshold"                  << endl << dbl2Str(wakeThreshold)       << endl;
    cfgFile << "wakeLength"                     << endl << dbl2Str(wakeLength)          << endl;
    cfgFile << "wakeNum"                        << endl << int2Str(wakeNum)             << endl;

    // close file
    cfgFile.close();
}

//-----------------------------------------------------------------------/
// checks if end of file will happen with given numbers of lines (numSkip)
//-----------------------------------------------------------------------/
int CGUIData::checkEndOfFile(
        const uInt& numSkip,
        const uInt& lineIndex,
        const uInt& numLines)
{
    if (lineIndex+numSkip >= numLines){
        showLineErrorMessage("Unexpected end of file", lineIndex);
        return -1;
    }

    return 0;
}

//-----------------------------------------------------------------------/
// reads integer data from configuration file
//-----------------------------------------------------------------------/
void CGUIData::readInteger(
        uInt& storage,
        uInt& lineIndex,
        const vecStr& fileLines)
{
    if (checkEndOfFile(1, lineIndex, fileLines.size()))
        return;

    storage = str2Int(fileLines[++lineIndex]);
}

//-----------------------------------------------------------------------/
// reads double data from configuration file
//-----------------------------------------------------------------------/
void CGUIData::readDouble(
        double& storage,
        uInt& lineIndex,
        const vecStr& fileLines)
{
    if (checkEndOfFile(1, lineIndex, fileLines.size()))
        return;

    storage = str2Dbl(fileLines[++lineIndex]);
}

//-----------------------------------------------------------------------/
// reads string data from configuration file
//-----------------------------------------------------------------------/
void CGUIData::readString(
        string& storage,
        uInt& lineIndex,
        const vecStr& fileLines)
{
    if (checkEndOfFile(1, lineIndex, fileLines.size()))
        return;

    storage = fileLines[++lineIndex];
}
