#include "CProjectData.h"

//-----------------------------------------------------------------------/
// initializes data
//-----------------------------------------------------------------------/
CProjectData::CProjectData()
{
    // project name
    projectName = "Untitled";
    // flow parameters
    airspeed = 1.0;
    density  = 1.225;
    pressure = 101325.0;
    mach     = 0.0;
    alfa.resize(1);
    beta.resize(1);
    alfa[0] = 0.0;
    beta[0] = 0.0;
    // reference data
    wingspan = 1.0;
    mac      = 1.0;
    surface  = 1.0;
    origin.resize(3);
    origin[0] = 0.0;
    origin[1] = 0.0;
    origin[2] = 0.0;
    // solver parameters
    method   = 0;
    error    = 1.0e-7;
    coplAng  = 3.1241;
    collDist = 1.0e-7;
    farField = 5.0;
    collCalc = 0;
    velOrder = 1;
    // result requests
    results             = true;
    coefficients        = true;
    forces              = true;
    geometry            = true;
    velocity            = true;
    pressureCoefficient = true;
    centerPoints        = false;
    doubletValues       = true;
    sourceValues        = true;
    velocityComponents  = false;
    meshCharacteristics = false;
    staticPressure      = false;
    dynamicPressure     = false;
    manometerPressure   = true;
    // surface plot dialog data
    surfResultsShown = false;
    plotType         = "elemental";
    limitType        = "automatic";
    minLimit         = 0.0;
    maxLimit         = 0.0;
    // 3DViewer data
    displayMode  = 1;
    showWake     = 0;
    showNodes    = 0;
    showLights   = 0;
    showEdges    = 1;
    showCoordSys = 1;
    showNodeIDs  = 0;
    showElemIDs  = 0;
    showLegend   = 0;
}

//-----------------------------------------------------------------------/
// opens ApameGUI project file
//-----------------------------------------------------------------------/
int CProjectData::openApameGUIFile(
        string& apameGUIFileName,
        CGUIData* pGUIData,
        CMesh* pMesh,
        CResults* pResults)
{
    // open ApameGUI project file for reading
    ifstream apameGUIFile(apameGUIFileName.c_str());
    if (!apameGUIFile.is_open()){
        showErrorMessage("Problem opening file "+apameGUIFileName);
        return -1;
    }

    // get current working directory from project file path
    uInt lastSlashPosition = 0;
    for (uInt i=apameGUIFileName.size()-1; i>0; i--){
        if (apameGUIFileName[i] == '/'){
            lastSlashPosition = i;
            break;
        }
        else if (apameGUIFileName[i] == '\\'){
            lastSlashPosition = i;
            break;
        }
    }
    if (lastSlashPosition)
        pGUIData->currentDir = apameGUIFileName.substr(0, lastSlashPosition);

    // read lines
    vecStr fileLines;
    string line;
    while (getLine(apameGUIFile, line))
        fileLines.push_back(line);
    apameGUIFile.close();

    // check if ApameGUI file
    if (fileLines[0] != "ApameGUI project file")
        return -1;

    // check version
    if (atoi(fileLines[1].c_str()) != apameGUIVersion    ||
        atoi(fileLines[2].c_str()) != apameGUISubVersion )
        showWarningMessage("File is of different version");

    for (uInt lineIndex=3; lineIndex<fileLines.size(); lineIndex++){
        if      (fileLines[lineIndex] == "projectName"){ if (readString(projectName, lineIndex, fileLines)) return -1; }
        // flow parameters
        else if (fileLines[lineIndex] == "airspeed"   ){ if (readFloat (airspeed,    lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "density"    ){ if (readFloat (density,     lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "pressure"   ){ if (readFloat (pressure,    lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "mach"       ){ if (readFloat (mach,        lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "casenum"    ){
            int caseNum = 0;
            if (readInteger(caseNum,  lineIndex, fileLines)) return -1;
            string alfaLine;
            string betaLine;
            if (readString(alfaLine, lineIndex, fileLines)) return -1;
            if (readString(betaLine, lineIndex, fileLines)) return -1;
            vecStr alfaSplit = splitString(alfaLine, ' ');
            vecStr betaSplit = splitString(betaLine, ' ');
            if (alfaSplit.size() != uInt(caseNum)){
                showLineErrorMessage("Number of AOAs does not match number of cases", lineIndex);
                return -1;
            }
            if (betaSplit.size() != uInt(caseNum)){
                showLineErrorMessage("Number of sideslip angles does not match number of cases", lineIndex);
                return -1;
            }
            alfa.resize(caseNum);
            beta.resize(caseNum);
            for (int i=0; i<caseNum; i++){
                alfa[i] = str2Dbl(alfaSplit[i]);
                beta[i] = str2Dbl(betaSplit[i]);
            }
        }
        // reference data
        else if (fileLines[lineIndex] == "wingspan"){ if (readFloat (wingspan, lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "mac"     ){ if (readFloat (mac,      lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "surface" ){ if (readFloat (surface,  lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "origin"){
            string originLine;
            if (readString(originLine, lineIndex, fileLines)) return -1;
            vecStr originSplit = splitString(originLine, ' ');
            if (originSplit.size() != 3){
                showLineErrorMessage("Origin is not defined with three values", lineIndex);
                return -1;
            }
            origin.resize(3);
            for (uInt i=0; i<3; i++)
                origin[i] = str2Dbl(originSplit[i]);
        }
        // solver parameters
        else if (fileLines[lineIndex] == "method"  ){ if (readInteger(method,   lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "error"   ){ if (readFloat  (error,    lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "collDist"){ if (readFloat  (collDist, lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "coplAng" ){ if (readFloat  (coplAng,  lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "farField"){ if (readFloat  (farField, lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "collCalc"){ if (readInteger(collCalc, lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "velOrder"){ if (readInteger(velOrder, lineIndex, fileLines)) return -1; }
        // result requests
        else if (fileLines[lineIndex] == "results"             ){ if (readBool(results,             lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "coefficients"        ){ if (readBool(coefficients,        lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "forces"              ){ if (readBool(forces,              lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "geometry"            ){ if (readBool(geometry,            lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "velocity"            ){ if (readBool(velocity,            lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "pressureCoefficient" ){ if (readBool(pressureCoefficient, lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "centerPoints"        ){ if (readBool(centerPoints,        lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "doubletValues"       ){ if (readBool(doubletValues,       lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "sourceValues"        ){ if (readBool(sourceValues,        lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "velocityComponents"  ){ if (readBool(velocityComponents,  lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "meshCharacteristics" ){ if (readBool(meshCharacteristics, lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "staticPressure"      ){ if (readBool(staticPressure,      lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "dynamicPressure"     ){ if (readBool(dynamicPressure,     lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "manometerPressure"   ){ if (readBool(manometerPressure,   lineIndex, fileLines)) return -1; }
        // surface plot dialog data
        else if (fileLines[lineIndex] == "surfResultsShown" ){ if (readBool  (surfResultsShown, lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "selectedData"     ){ if (readString(selectedData,     lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "selectedCase"     ){ if (readString(selectedCase,     lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "plotType"         ){ if (readString(plotType,         lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "limitType"        ){ if (readString(limitType,        lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "minLimit"         ){ if (readFloat (minLimit,         lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "maxLimit"         ){ if (readFloat (maxLimit,         lineIndex, fileLines)) return -1; }
        // 3DViewer data
        else if (fileLines[lineIndex] == "displayMode"   ){ if (readInteger(displayMode, lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "showWake"    ){ if (readBool(showWake,     lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "showNodes"   ){ if (readBool(showNodes,    lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "showLights"  ){ if (readBool(showLights,   lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "showEdges"   ){ if (readBool(showEdges,    lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "showCoordSys"){ if (readBool(showCoordSys, lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "showNodeIDs" ){ if (readBool(showNodeIDs,  lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "showElemIDs" ){ if (readBool(showElemIDs,  lineIndex, fileLines)) return -1; }
        else if (fileLines[lineIndex] == "showLegend"  ){ if (readBool(showLegend,   lineIndex, fileLines)) return -1; }
        // nodes
        else if (fileLines[lineIndex] == "nodes"){
            int nodeNum = 0;
            if (readInteger(nodeNum,  lineIndex, fileLines)) return -1;
            if (checkEndOfFile(nodeNum, lineIndex, fileLines.size())) return -1;
            for (int nodeIndex=0; nodeIndex<nodeNum; nodeIndex++){
                string nodeLine;
                if (readString(nodeLine, lineIndex, fileLines)) return -1;
                vecStr nodeSplit = splitString(nodeLine, ' ');
                if (nodeSplit.size() != 4){
                    showLineErrorMessage("Node incorrectly defined", lineIndex);
                    return -1;
                }
                double coordinates[3];
                coordinates[0] = str2Dbl(nodeSplit[1]);
                coordinates[1] = str2Dbl(nodeSplit[2]);
                coordinates[2] = str2Dbl(nodeSplit[3]);
                int retCode = pMesh->insertNodeWithIDCoords(str2Int(nodeSplit[0]), coordinates);
                if (retCode == -1){
                    showLineErrorMessage("Node ID already exists", lineIndex);
                    return -1;
                }
                else if (retCode == -2){
                    showLineErrorMessage("Node ID is equal 0", lineIndex);
                    return -1;
                }
            }
        }
        // elements
        else if (fileLines[lineIndex] == "elements"){
            uInt elemStartLine = lineIndex+1;
            // read element data
            int elemNum = 0;
            if (readInteger(elemNum,  lineIndex, fileLines)) return -1;
            if (checkEndOfFile(elemNum, lineIndex, fileLines.size())) return -1;
            vecUInt elemIDs(elemNum);
            vecUInt elemTypes(elemNum);
            vecVecUInt elemNodeIDs(elemNum);
            vecVecUInt neighborIDs(elemNum);
            vecVecUInt trailIDs(elemNum);
            for (int elemIndex=0; elemIndex<elemNum; elemIndex++){
                string elemLine;
                if (readString(elemLine, lineIndex, fileLines)) return -1;
                vecStr elemSplit = splitString(elemLine, ' ');
                if (elemSplit.size() < 5){
                    showLineErrorMessage("Element incorrectly defined", lineIndex);
                    return -1;
                }
                elemTypes[elemIndex] = str2Int(elemSplit[1]);
                if (elemTypes[elemIndex] != 1  &&
                    elemTypes[elemIndex] != 2  &&
                    elemTypes[elemIndex] != 10 &&
                    elemTypes[elemIndex] != 11 &&
                    elemTypes[elemIndex] != 20 &&
                    elemTypes[elemIndex] != 21 ){
                    showLineErrorMessage("Incorrect element type", lineIndex);
                    return -1;
                }
                if (elemTypes[elemIndex]==1 || elemTypes[elemIndex]==20){
                    // non-wake quad element
                    if (elemSplit.size() < 6){
                        showLineErrorMessage("Quad element incorrectly defined", lineIndex);
                        return -1;
                    }
                    elemIDs[elemIndex] = str2Int(elemSplit[0]);
                    elemNodeIDs[elemIndex].resize(4);
                    neighborIDs[elemIndex].resize(elemSplit.size()-6);
                    for (uInt i=0; i<4; i++)
                        elemNodeIDs[elemIndex][i] = str2Int(elemSplit[2+i]);
                    for (uInt i=0; i<neighborIDs[elemIndex].size(); i++)
                        neighborIDs[elemIndex][i] = str2Int(elemSplit[6+i]);
                }
                else if (elemTypes[elemIndex]==2 || elemTypes[elemIndex]==21){
                    // non-wake tria element
                    elemIDs[elemIndex] = str2Int(elemSplit[0]);
                    elemNodeIDs[elemIndex].resize(3);
                    neighborIDs[elemIndex].resize(elemSplit.size()-5);
                    for (uInt i=0; i<3; i++)
                        elemNodeIDs[elemIndex][i] = str2Int(elemSplit[2+i]);
                    for (uInt i=0; i<neighborIDs[elemIndex].size(); i++)
                        neighborIDs[elemIndex][i] = str2Int(elemSplit[5+i]);
                }
                else if (elemTypes[elemIndex] == 10){
                    // quad wake element
                    if (elemSplit.size() < 6){
                        showLineErrorMessage("Quad wake element incorrectly defined", lineIndex);
                        return -1;
                    }
                    elemIDs[elemIndex] = str2Int(elemSplit[0]);
                    elemNodeIDs[elemIndex].resize(4);
                    trailIDs[elemIndex].resize(elemSplit.size()-6);
                    for (uInt i=0; i<4; i++)
                        elemNodeIDs[elemIndex][i] = str2Int(elemSplit[2+i]);
                    for (uInt i=0; i<trailIDs[elemIndex].size(); i++)
                        trailIDs[elemIndex][i] = str2Int(elemSplit[6+i]);
                }
                else if (elemTypes[elemIndex] == 11){
                    // tria wake element
                    elemIDs[elemIndex] = str2Int(elemSplit[0]);
                    elemNodeIDs[elemIndex].resize(3);
                    trailIDs[elemIndex].resize(elemSplit.size()-5);
                    for (uInt i=0; i<3; i++)
                        elemNodeIDs[elemIndex][i] = str2Int(elemSplit[2+i]);
                    for (uInt i=0; i<trailIDs[elemIndex].size(); i++)
                        trailIDs[elemIndex][i] = str2Int(elemSplit[5+i]);
                }
            }
            // store element data to mesh
            for (int elemIndex=0; elemIndex<elemNum; elemIndex++){
                // insert element with ID, type and node IDs
                if (pMesh->insertElemWithIDTypeNodeIDs(elemIDs[elemIndex], elemTypes[elemIndex], elemNodeIDs[elemIndex])){
                    showLineErrorMessage("Problem adding element to mesh", elemStartLine+elemIndex);
                    return -1;
                }
            }
            // store neighboring and wake data to mesh
            for (int elemIndex=0; elemIndex<elemNum; elemIndex++){
                if (elemTypes[elemIndex] == 1  ||
                    elemTypes[elemIndex] == 2  ||
                    elemTypes[elemIndex] == 20 ||
                    elemTypes[elemIndex] == 21 ){
                    if (pMesh->insertNeighborsToElem(elemIDs[elemIndex], neighborIDs[elemIndex])){
                        showLineErrorMessage("Problem adding neighbors to element", elemStartLine+elemIndex);
                        return -1;
                    }
                }
                else{
                    if (pMesh->insertTrailsToElem(elemIDs[elemIndex], trailIDs[elemIndex])){
                        showLineErrorMessage("Problem adding trailing element(s) to wake element", elemStartLine+elemIndex);
                        return -1;
                    }
                }
            }
        }
        // result data
        else if (fileLines[lineIndex] == "solver_run_data"){
            if (!pResults) pResults = new CResults();
            if (readString (pResults->date,          lineIndex, fileLines)) return -1;
            if (readString (pResults->time,          lineIndex, fileLines)) return -1;
            if (readString (pResults->solverVersion, lineIndex, fileLines)) return -1;
            if (readUInt   (pResults->numElem,       lineIndex, fileLines)) return -1;
            if (readUInt   (pResults->numBodyElem,   lineIndex, fileLines)) return -1;
            if (readUInt   (pResults->numNode,       lineIndex, fileLines)) return -1;
            if (readFloat  (pResults->airspeed,      lineIndex, fileLines)) return -1;
            if (readFloat  (pResults->density,       lineIndex, fileLines)) return -1;
            if (readFloat  (pResults->pressure,      lineIndex, fileLines)) return -1;
            if (readFloat  (pResults->mach,          lineIndex, fileLines)) return -1;
            if (readUInt   (pResults->caseNumber,    lineIndex, fileLines)) return -1;
            // read angles of attack
            string alfaLine;
            if (readString(alfaLine, lineIndex, fileLines)) return -1;
            if (alfaLine.empty())
                lineIndex += 2;
            else{
                vecStr alfaSplit = splitString(alfaLine, ' ');
                if (alfaSplit.size() != pResults->caseNumber){
                    showLineErrorMessage("Number of angles of attack does not match number of cases", lineIndex);
                    return -1;
                }
                pResults->alfa.resize(pResults->caseNumber);
                for (uInt i=0; i<alfa.size(); i++)
                    pResults->alfa[i] = str2Dbl(alfaSplit[i]);
                // read sideslip angles
                string betaLine;
                if (readString(betaLine, lineIndex, fileLines)) return -1;
                vecStr betaSplit = splitString(betaLine, ' ');
                if (betaSplit.size() != pResults->caseNumber){
                    showLineErrorMessage("Number of sideslip angles does not match number of cases", lineIndex);
                    return -1;
                }
                pResults->beta.resize(pResults->caseNumber);
                for (uInt i=0; i<beta.size(); i++)
                    pResults->beta[i] = str2Dbl(betaSplit[i]);
            }
            if (readFloat(pResults->wingspan, lineIndex, fileLines)) return -1;
            if (readFloat(pResults->mac,      lineIndex, fileLines)) return -1;
            if (readFloat(pResults->surface,  lineIndex, fileLines)) return -1;
            string originLine;
            if (readString(originLine, lineIndex, fileLines)) return -1;
            vecStr originSplit = splitString(originLine, ' ');
            if (originSplit.size() == 3){
                pResults->origin.resize(3);
                for (uInt i=0; i<3; i++)
                    pResults->origin[i] = str2Dbl(originSplit[i]);
            }
            if (readUInt   (pResults->method,   lineIndex, fileLines)) return -1;
            if (readFloat  (pResults->error,    lineIndex, fileLines)) return -1;
            if (readFloat  (pResults->coplAng,  lineIndex, fileLines)) return -1;
            if (readFloat  (pResults->collDist, lineIndex, fileLines)) return -1;
            if (readFloat  (pResults->farField, lineIndex, fileLines)) return -1;
            if (readUInt   (pResults->collCalc, lineIndex, fileLines)) return -1;
            if (readUInt   (pResults->velOrder, lineIndex, fileLines)) return -1;
        }
        // coefficients of forces and moments
        else if (fileLines[lineIndex] == "coefficients_of_forces_and_moments"){
            if (!pResults) pResults = new CResults();
            int numCases;
            if (readInteger(numCases, lineIndex, fileLines)) return -1;
            if (checkEndOfFile(numCases, lineIndex, fileLines.size())) return -1;
            pResults->Cx   .resize(numCases);
            pResults->Cy   .resize(numCases);
            pResults->Cz   .resize(numCases);
            pResults->Cl   .resize(numCases);
            pResults->Cm   .resize(numCases);
            pResults->Cn   .resize(numCases);
            pResults->Cdrag.resize(numCases);
            pResults->Cside.resize(numCases);
            pResults->Clift.resize(numCases);
            for (int i=0; i<numCases; i++){
                string coefLine;
                if (readString(coefLine, lineIndex, fileLines)) return -1;
                vecStr coefSplit = splitString(coefLine, ' ');
                if (coefSplit.size() != 9){
                    showLineErrorMessage("Incorrect number of coefficients", lineIndex);
                    return -1;
                }
                pResults->Cx   [i] = str2Dbl(coefSplit[0]);
                pResults->Cy   [i] = str2Dbl(coefSplit[1]);
                pResults->Cz   [i] = str2Dbl(coefSplit[2]);
                pResults->Cl   [i] = str2Dbl(coefSplit[3]);
                pResults->Cm   [i] = str2Dbl(coefSplit[4]);
                pResults->Cn   [i] = str2Dbl(coefSplit[5]);
                pResults->Cdrag[i] = str2Dbl(coefSplit[6]);
                pResults->Cside[i] = str2Dbl(coefSplit[7]);
                pResults->Clift[i] = str2Dbl(coefSplit[8]);
            }
        }
        // forces and moments
        else if (fileLines[lineIndex] == "forces_and_moments"){
            if (!pResults) pResults = new CResults();
            int numCases;
            if (readInteger(numCases, lineIndex, fileLines)) return -1;
            if (checkEndOfFile(numCases, lineIndex, fileLines.size())) return -1;
            pResults->Fx   .resize(numCases);
            pResults->Fy   .resize(numCases);
            pResults->Fz   .resize(numCases);
            pResults->Fl   .resize(numCases);
            pResults->Fm   .resize(numCases);
            pResults->Fn   .resize(numCases);
            pResults->Fdrag.resize(numCases);
            pResults->Fside.resize(numCases);
            pResults->Flift.resize(numCases);
            for (int i=0; i<numCases; i++){
                string forcLine;
                if (readString(forcLine, lineIndex, fileLines)) return -1;
                vecStr forcSplit = splitString(forcLine, ' ');
                if (forcSplit.size() != 9){
                    showLineErrorMessage("Incorrect number of forces", lineIndex);
                    return -1;
                }
                pResults->Fx   [i] = str2Dbl(forcSplit[0]);
                pResults->Fy   [i] = str2Dbl(forcSplit[1]);
                pResults->Fz   [i] = str2Dbl(forcSplit[2]);
                pResults->Fl   [i] = str2Dbl(forcSplit[3]);
                pResults->Fm   [i] = str2Dbl(forcSplit[4]);
                pResults->Fn   [i] = str2Dbl(forcSplit[5]);
                pResults->Fdrag[i] = str2Dbl(forcSplit[6]);
                pResults->Fside[i] = str2Dbl(forcSplit[7]);
                pResults->Flift[i] = str2Dbl(forcSplit[8]);
            }
        }
        // surface data
        else if (fileLines[lineIndex] == "surfDatas"){
            if (!pResults) pResults = new CResults();
            int surfDataSize = 0;
            if (readInteger(surfDataSize, lineIndex, fileLines)) return -1;
            pResults->surfDatas.resize(surfDataSize);
            for (uInt surfIndex=0; surfIndex<pResults->surfDatas.size(); surfIndex++){
                // read data name
                if (readString(pResults->surfDatas[surfIndex].dataName, lineIndex, fileLines)) return -1;
                // read case titles
                int caseTitleSize = 0;
                if (readInteger(caseTitleSize, lineIndex, fileLines)) return -1;
                if (checkEndOfFile(caseTitleSize, lineIndex, fileLines.size())) return -1;
                pResults->surfDatas[surfIndex].caseTitles.resize(caseTitleSize);
                for (uInt caseIndex=0; caseIndex<pResults->surfDatas[surfIndex].caseTitles.size(); caseIndex++){
                    if (readString(pResults->surfDatas[surfIndex].caseTitles[caseIndex], lineIndex, fileLines)) return -1;
                }
                // read element IDs
                int elemIDSize = 0;
                if (readInteger(elemIDSize, lineIndex, fileLines)) return -1;
                if (checkEndOfFile(elemIDSize, lineIndex, fileLines.size())) return -1;
                pResults->surfDatas[surfIndex].elemIDs.resize(elemIDSize);
                for (uInt elemIndex=0; elemIndex<pResults->surfDatas[surfIndex].elemIDs.size(); elemIndex++){
                    if (readUInt(pResults->surfDatas[surfIndex].elemIDs[elemIndex], lineIndex, fileLines)) return -1;
                }
                // read data on element
                int dataSize = 0;
                if (readInteger(dataSize, lineIndex, fileLines)) return -1;
                pResults->surfDatas[surfIndex].data.resize(dataSize);
                for (uInt i=0; i<pResults->surfDatas[surfIndex].data.size(); i++){
                    int jSize = 0;
                    if (readInteger(jSize, lineIndex, fileLines)) return -1;
                    if (checkEndOfFile(jSize, lineIndex, fileLines.size())) return -1;
                    pResults->surfDatas[surfIndex].data[i].resize(jSize);
                    for (uInt j=0; j<pResults->surfDatas[surfIndex].data[i].size(); j++){
                        if (readFloat(pResults->surfDatas[surfIndex].data[i][j], lineIndex, fileLines)) return -1;
                    }
                }
            }
        }
    }

    return 0;
}

int CProjectData::saveApameGUIFile(
        string& apameGUIFileName,
        CMesh* pMesh,
        CResults* pResults)
{
    // delete file if already exists
    if (fileExists(apameGUIFileName) && remove((apameGUIFileName).c_str())){
        showErrorMessage("Existing file "+apameGUIFileName+" could not be deleted\nCheck permissions!");
        return -1;
    }

    // open ApameGUI file for writing
    ofstream apameGUIFile(apameGUIFileName.c_str());
    if (!apameGUIFile.is_open()){
        showErrorMessage("Problem writing to file "+apameGUIFileName);
        return -1;
    }

    // write program title
    apameGUIFile << "ApameGUI project file" << endl;

    // write ApameGUI version, subversion and date
    apameGUIFile << apameGUIVersion << endl;
    apameGUIFile << apameGUISubVersion << endl;
    apameGUIFile << apameGUIVersionDate << endl;

    // write project name
    apameGUIFile << "projectName" << endl;
    apameGUIFile << projectName << endl;

    // write floats in scientific format
    apameGUIFile << scientific;

    // flow parameters
    apameGUIFile << "airspeed" << endl;
    apameGUIFile << airspeed << endl;
    apameGUIFile << "density" << endl;
    apameGUIFile << density << endl;
    apameGUIFile << "pressure" << endl;
    apameGUIFile << pressure << endl;
    apameGUIFile << "mach" << endl;
    apameGUIFile << mach << endl;
    apameGUIFile << "casenum" << endl;
    apameGUIFile << alfa.size() << endl;
    for (uInt i=0; i<alfa.size(); i++) apameGUIFile << setw(16) << alfa[i]; apameGUIFile << endl;
    for (uInt i=0; i<beta.size(); i++) apameGUIFile << setw(16) << beta[i]; apameGUIFile << endl;

    // reference data
    apameGUIFile << "wingspan" << endl;
    apameGUIFile << wingspan << endl;
    apameGUIFile << "mac" << endl;
    apameGUIFile << mac << endl;
    apameGUIFile << "surface" << endl;
    apameGUIFile << surface << endl;
    apameGUIFile << "origin" << endl;
    for (uInt i=0; i<3; i++) apameGUIFile << setw(16) << origin[i]; apameGUIFile << endl;

    // solver parameters
    apameGUIFile << "method" << endl;
    apameGUIFile << method << endl;
    apameGUIFile << "error" << endl;
    apameGUIFile << error << endl;
    apameGUIFile << "collDist" << endl;
    apameGUIFile << collDist << endl;
    apameGUIFile << "coplAng" << endl;
    apameGUIFile << coplAng << endl;
    apameGUIFile << "farField" << endl;
    apameGUIFile << farField << endl;
    apameGUIFile << "collCalc" << endl;
    apameGUIFile << collCalc << endl;
    apameGUIFile << "velOrder" << endl;
    apameGUIFile << velOrder << endl;

    // result requests
    apameGUIFile << "results" << endl;
    apameGUIFile << results << endl;
    apameGUIFile << "coefficients" << endl;
    apameGUIFile << coefficients << endl;
    apameGUIFile << "forces" << endl;
    apameGUIFile << forces << endl;
    apameGUIFile << "geometry" << endl;
    apameGUIFile << geometry << endl;
    apameGUIFile << "velocity" << endl;
    apameGUIFile << velocity << endl;
    apameGUIFile << "pressureCoefficient" << endl;
    apameGUIFile << pressureCoefficient << endl;
    apameGUIFile << "centerPoints" << endl;
    apameGUIFile << centerPoints << endl;
    apameGUIFile << "doubletValues" << endl;
    apameGUIFile << doubletValues << endl;
    apameGUIFile << "sourceValues" << endl;
    apameGUIFile << sourceValues << endl;
    apameGUIFile << "velocityComponents" << endl;
    apameGUIFile << velocityComponents << endl;
    apameGUIFile << "meshCharacteristics" << endl;
    apameGUIFile << meshCharacteristics << endl;
    apameGUIFile << "staticPressure" << endl;
    apameGUIFile << staticPressure << endl;
    apameGUIFile << "dynamicPressure" << endl;
    apameGUIFile << dynamicPressure << endl;
    apameGUIFile << "manometerPressure" << endl;
    apameGUIFile << manometerPressure << endl;

    // surface plot dialog data
    apameGUIFile << "surfResultsShown" << endl;
    apameGUIFile << surfResultsShown << endl;
    apameGUIFile << "selectedData" << endl;
    apameGUIFile << selectedData << endl;
    apameGUIFile << "selectedCase" << endl;
    apameGUIFile << selectedCase << endl;
    apameGUIFile << "plotType" << endl;
    apameGUIFile << plotType << endl;
    apameGUIFile << "limitType" << endl;
    apameGUIFile << limitType << endl;
    apameGUIFile << "minLimit" << endl;
    apameGUIFile << minLimit << endl;
    apameGUIFile << "maxLimit" << endl;
    apameGUIFile << maxLimit << endl;

    // 3D Viewer data
    apameGUIFile << "displayMode" << endl;
    apameGUIFile << displayMode << endl;
    apameGUIFile << "showWake" << endl;
    apameGUIFile << showWake << endl;
    apameGUIFile << "showNodes" << endl;
    apameGUIFile << showNodes << endl;
    apameGUIFile << "showLights" << endl;
    apameGUIFile << showLights << endl;
    apameGUIFile << "showEdges" << endl;
    apameGUIFile << showEdges << endl;
    apameGUIFile << "showCoordSys" << endl;
    apameGUIFile << showCoordSys << endl;
    apameGUIFile << "showNodeIDs" << endl;
    apameGUIFile << showNodeIDs << endl;
    apameGUIFile << "showElemIDs" << endl;
    apameGUIFile << showElemIDs << endl;
    apameGUIFile << "showLegend" << endl;
    apameGUIFile << showLegend << endl;

    // get node data from mesh
    vecUInt nodeIDs;
    vecVecDbl coordinates;
    pMesh->getNodeData(nodeIDs, coordinates);
    // write nodes
    apameGUIFile << "nodes" << endl;
    apameGUIFile << nodeIDs.size() << endl;
    for (uInt nodeIndex=0; nodeIndex<nodeIDs.size(); nodeIndex++){
        apameGUIFile << setw(16) << nodeIDs[nodeIndex];
        for (uInt i=0; i<coordinates[nodeIndex].size(); i++)
            apameGUIFile << setw(16) << coordinates[nodeIndex][i];
        apameGUIFile << endl;
    }

    // get element data from mesh
    vecUInt elemIDs;
    vecUInt elemTypes;
    vecVecUInt elemNodeIDs;
    vecVecUInt neighborIDs;
    vecVecUInt trailIDs;
    pMesh->getElemSolverData(elemIDs, elemTypes, elemNodeIDs, neighborIDs, trailIDs);

    // write elements
    apameGUIFile << "elements" << endl;
    apameGUIFile << elemIDs.size() << endl;
    for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
        apameGUIFile << setw(16) << elemIDs[elemIndex];
        apameGUIFile << setw(16) << elemTypes[elemIndex];
        for (uInt i=0; i<elemNodeIDs[elemIndex].size(); i++) apameGUIFile << setw(16) << elemNodeIDs[elemIndex][i];
        for (uInt i=0; i<neighborIDs[elemIndex].size(); i++) apameGUIFile << setw(16) << neighborIDs[elemIndex][i];
        for (uInt i=0; i<trailIDs   [elemIndex].size(); i++) apameGUIFile << setw(16) << trailIDs   [elemIndex][i];
        apameGUIFile << endl;
    }

    // write results data
    if (pResults){
        // solver run data
        apameGUIFile << "solver_run_data" << endl;
        apameGUIFile << pResults->date << endl;
        apameGUIFile << pResults->time << endl;
        apameGUIFile << pResults->solverVersion << endl;
        apameGUIFile << pResults->numElem << endl;
        apameGUIFile << pResults->numBodyElem << endl;
        apameGUIFile << pResults->numNode << endl;
        apameGUIFile << pResults->airspeed << endl;
        apameGUIFile << pResults->density << endl;
        apameGUIFile << pResults->pressure << endl;
        apameGUIFile << pResults->mach << endl;
        apameGUIFile << pResults->caseNumber << endl;
        for (uInt i=0; i<pResults->alfa.size(); i++) apameGUIFile << setw(16) << pResults->alfa[i]; apameGUIFile << endl;
        for (uInt i=0; i<pResults->beta.size(); i++) apameGUIFile << setw(16) << pResults->beta[i]; apameGUIFile << endl;
        apameGUIFile << pResults->wingspan << endl;
        apameGUIFile << pResults->mac << endl;
        apameGUIFile << pResults->surface << endl;
        for (uInt i=0; i<pResults->origin.size(); i++) apameGUIFile << setw(16) << pResults->origin[i]; apameGUIFile << endl;
        apameGUIFile << pResults->method << endl;
        apameGUIFile << pResults->error << endl;
        apameGUIFile << pResults->coplAng << endl;
        apameGUIFile << pResults->collDist << endl;
        apameGUIFile << pResults->farField << endl;
        apameGUIFile << pResults->collCalc << endl;
        apameGUIFile << pResults->velOrder << endl;

        // coefficients of forces and moments
        apameGUIFile << "coefficients_of_forces_and_moments" << endl;
        apameGUIFile << pResults->Cx.size() << endl;
        for (uInt i=0; i<pResults->Cx.size(); i++)
            apameGUIFile << pResults->Cx   [i] << " " << pResults->Cy   [i] << " " << pResults->Cz   [i] << " " <<
                            pResults->Cl   [i] << " " << pResults->Cm   [i] << " " << pResults->Cn   [i] << " " <<
                            pResults->Cdrag[i] << " " << pResults->Cside[i] << " " << pResults->Clift[i] << endl;

        // forces and moments
        apameGUIFile << "forces_and_moments" << endl;
        apameGUIFile << pResults->Fx.size() << endl;
        for (uInt i=0; i<pResults->Fx.size(); i++)
            apameGUIFile << pResults->Fx   [i] << " " << pResults->Fy   [i] << " " << pResults->Fz   [i] << " " <<
                            pResults->Fl   [i] << " " << pResults->Fm   [i] << " " << pResults->Fn   [i] << " " <<
                            pResults->Fdrag[i] << " " << pResults->Fside[i] << " " << pResults->Flift[i] << endl;

        // surface data
        apameGUIFile << "surfDatas" << endl;
        apameGUIFile << pResults->surfDatas.size() << endl;
        for (uInt surfIndex=0; surfIndex<pResults->surfDatas.size(); surfIndex++){
            apameGUIFile << pResults->surfDatas[surfIndex].dataName << endl;
            apameGUIFile << pResults->surfDatas[surfIndex].caseTitles.size() << endl;
            for (uInt i=0; i<pResults->surfDatas[surfIndex].caseTitles.size(); i++)
                apameGUIFile << pResults->surfDatas[surfIndex].caseTitles[i] << endl;
            apameGUIFile << pResults->surfDatas[surfIndex].elemIDs.size() << endl;
            for (uInt i=0; i<pResults->surfDatas[surfIndex].elemIDs.size(); i++)
                apameGUIFile << pResults->surfDatas[surfIndex].elemIDs[i] << endl;
            apameGUIFile << pResults->surfDatas[surfIndex].data.size() << endl;
            for (uInt i=0; i<pResults->surfDatas[surfIndex].data.size(); i++){
                apameGUIFile << pResults->surfDatas[surfIndex].data[i].size() << endl;
                for (uInt j=0; j<pResults->surfDatas[surfIndex].data[i].size(); j++)
                    apameGUIFile << pResults->surfDatas[surfIndex].data[i][j] << endl;
            }
        }
    }

    // close ApameGUI file
    apameGUIFile.close();

    return 0;
}

//-----------------------------------------------------------------------/
// checks if end of file will happen with given numbers of lines (numSkip)
//-----------------------------------------------------------------------/
int CProjectData::checkEndOfFile(
        const uInt& numSkip,
        const uInt& lineIndex,
        const uInt& numLines)
{
    if (lineIndex+numSkip >= numLines){
        showLineErrorMessage("Unexpected end of file", lineIndex);
        return -1;
    }

    return 0;
}

//-----------------------------------------------------------------------/
// reads string data
//-----------------------------------------------------------------------/
int CProjectData::readString(
        string& storage,
        uInt& lineIndex,
        const vecStr& fileLines)
{
    if (checkEndOfFile(1, lineIndex, fileLines.size()))
        return -1;

    storage = fileLines[++lineIndex];

    return 0;
}

//-----------------------------------------------------------------------/
// reads bool data
//-----------------------------------------------------------------------/
int CProjectData::readBool(
        bool& storage,
        uInt& lineIndex,
        const vecStr& fileLines)
{
    if (checkEndOfFile(1, lineIndex, fileLines.size()))
        return -1;

    storage = str2Int(fileLines[++lineIndex])==1;

    return 0;
}

//-----------------------------------------------------------------------/
// reads integer data
//-----------------------------------------------------------------------/
int CProjectData::readInteger(
        int& storage,
        uInt& lineIndex,
        const vecStr& fileLines)
{
    if (checkEndOfFile(1, lineIndex, fileLines.size()))
        return -1;

    storage = str2Int(fileLines[++lineIndex]);

    return 0;
}

//-----------------------------------------------------------------------/
// reads usigned integer data
//-----------------------------------------------------------------------/
int CProjectData::readUInt(
        uInt& storage,
        uInt& lineIndex,
        const vecStr& fileLines)
{
    if (checkEndOfFile(1, lineIndex, fileLines.size()))
        return -1;

    storage = uInt(str2Int(fileLines[++lineIndex]));

    return 0;
}

//-----------------------------------------------------------------------/
// reads float data
//-----------------------------------------------------------------------/
int CProjectData::readFloat(
        double& storage,
        uInt& lineIndex,
        const vecStr& fileLines)
{
    if (checkEndOfFile(1, lineIndex, fileLines.size()))
        return -1;

    storage = str2Dbl(fileLines[++lineIndex]);

    return 0;
}
