#include "importExport.h"

//-----------------------------------------------------------------------/
// reads ApameSolver input file into mesh and project data
//-----------------------------------------------------------------------/
void readApameSolverInputFile(
        CMesh* pMesh,
        CProjectData* pProjectData,
        const string& fileName,
        bool readFlowParams,
        bool readRefData,
        bool readSolverParams,
        bool readResReq,
        bool readNodes,
        bool readElements)
{
    // open file
    ifstream apameFile(fileName.c_str());
    if (!apameFile.is_open()){
        showErrorMessage("Cannot open ApameSolver input file");
        return;
    }

    // read lines
    vecStr fileLines;
    string line;
    while (getLine(apameFile, line))
        fileLines.push_back(line);
    apameFile.close();

    // check if ApameSolver input file
    if (fileLines[0].substr(0,16) != "APAME input file"){
        showErrorMessage("Not an ApameSolver input file");
        return;
    }

    // get shift values (maximum IDs of already existing nodes and elements)
    uInt nodeShift = pMesh->getMaxNodeID();
    uInt elemShift = pMesh->getMaxElemID();

    // read content
    vecUInt nodeIDs;
    vecVecDbl nodeCoords;
    uInt nodeStartLine = 0;
    vecUInt elemIDs;
    vecUInt elemTypes;
    vecVecUInt elemNodes;
    vecVecUInt elemNeighbors;
    uInt elemStartLine = 0;
    for (uInt lineIndex=0; lineIndex<fileLines.size(); lineIndex++){
        if      (readFlowParams && fileLines[lineIndex].substr(0,8) == "AIRSPEED") pProjectData->airspeed = str2Dbl(fileLines[lineIndex].substr(8));
        else if (readFlowParams && fileLines[lineIndex].substr(0,7) == "DENSITY")  pProjectData->density  = str2Dbl(fileLines[lineIndex].substr(7));
        else if (readFlowParams && fileLines[lineIndex].substr(0,8) == "PRESSURE") pProjectData->pressure = str2Dbl(fileLines[lineIndex].substr(8));
        else if (readFlowParams && fileLines[lineIndex].substr(0,4) == "MACH")     pProjectData->mach     = str2Dbl(fileLines[lineIndex].substr(4));
        else if (readFlowParams && fileLines[lineIndex].substr(0,8) == "CASE_NUM"){
            int caseSize = str2Int(fileLines[lineIndex].substr(8));
            if (checkEndOfFile(2, lineIndex, fileLines.size()))
                return;
            vecStr vecLineAlfa = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
            vecStr vecLineBeta = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
            if (vecLineAlfa.size()!=uInt(caseSize) || vecLineBeta.size()!=uInt(caseSize)){
                showLineErrorMessage("Case number not consistent in ApameSolver input file", lineIndex+1);
                return;
            }
            pProjectData->alfa.resize(caseSize);
            pProjectData->beta.resize(caseSize);
            for (int caseIndex=0; caseIndex<caseSize; caseIndex++){
                pProjectData->alfa[caseIndex] = str2Dbl(vecLineAlfa[caseIndex]);
                pProjectData->beta[caseIndex] = str2Dbl(vecLineBeta[caseIndex]);
            }
        }
        else if (readRefData && fileLines[lineIndex].substr(0,8) == "WINGSPAN") pProjectData->wingspan = str2Dbl(fileLines[lineIndex].substr(8));
        else if (readRefData && fileLines[lineIndex].substr(0,3) == "MAC")      pProjectData->mac      = str2Dbl(fileLines[lineIndex].substr(3));
        else if (readRefData && fileLines[lineIndex].substr(0,7) == "SURFACE")  pProjectData->surface  = str2Dbl(fileLines[lineIndex].substr(7));
        else if (readRefData && fileLines[lineIndex].substr(0,6) == "ORIGIN"){
            if (checkEndOfFile(1, lineIndex, fileLines.size()))
                return;
            vecStr vecOrigin = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
            if (vecOrigin.size() != 3){
                showLineErrorMessage("Origin incorrectly defined in ApameSolver input file", lineIndex+1);
                return;
            }
            for (int i=0; i<3; i++)
                pProjectData->origin[i] = str2Dbl(vecOrigin[i]);
        }
        else if (readSolverParams && fileLines[lineIndex].substr(0,6) == "METHOD")   pProjectData->method   = str2Int(fileLines[lineIndex].substr(6));
        else if (readSolverParams && fileLines[lineIndex].substr(0,5) == "ERROR")    pProjectData->error    = str2Dbl(fileLines[lineIndex].substr(5));
        else if (readSolverParams && fileLines[lineIndex].substr(0,8) == "COLLDIST") pProjectData->collDist = str2Dbl(fileLines[lineIndex].substr(8));
        else if (readSolverParams && fileLines[lineIndex].substr(0,8) == "FARFIELD") pProjectData->farField = str2Dbl(fileLines[lineIndex].substr(8));
        else if (readSolverParams && fileLines[lineIndex].substr(0,8) == "COLLCALC") pProjectData->collCalc = str2Int(fileLines[lineIndex].substr(8));
        else if (readSolverParams && fileLines[lineIndex].substr(0,8) == "VELORDER") pProjectData->velOrder = str2Int(fileLines[lineIndex].substr(8));
        else if (readResReq && fileLines[lineIndex].substr(0,7) == "RESULTS")  pProjectData->results             = str2Int(fileLines[lineIndex].substr(7))==1;
        else if (readResReq && fileLines[lineIndex].substr(0,8) == "RES_COEF") pProjectData->coefficients        = str2Int(fileLines[lineIndex].substr(8))==1;
        else if (readResReq && fileLines[lineIndex].substr(0,8) == "RES_FORC") pProjectData->forces              = str2Int(fileLines[lineIndex].substr(8))==1;
        else if (readResReq && fileLines[lineIndex].substr(0,8) == "RES_GEOM") pProjectData->geometry            = str2Int(fileLines[lineIndex].substr(8))==1;
        else if (readResReq && fileLines[lineIndex].substr(0,8) == "RES_VELO") pProjectData->velocity            = str2Int(fileLines[lineIndex].substr(8))==1;
        else if (readResReq && fileLines[lineIndex].substr(0,8) == "RES_PRES") pProjectData->pressureCoefficient = str2Int(fileLines[lineIndex].substr(8))==1;
        else if (readResReq && fileLines[lineIndex].substr(0,8) == "RES_CENT") pProjectData->centerPoints        = str2Int(fileLines[lineIndex].substr(8))==1;
        else if (readResReq && fileLines[lineIndex].substr(0,8) == "RES_DOUB") pProjectData->doubletValues       = str2Int(fileLines[lineIndex].substr(8))==1;
        else if (readResReq && fileLines[lineIndex].substr(0,8) == "RES_SORC") pProjectData->sourceValues        = str2Int(fileLines[lineIndex].substr(8))==1;
        else if (readResReq && fileLines[lineIndex].substr(0,8) == "RES_VELC") pProjectData->velocityComponents  = str2Int(fileLines[lineIndex].substr(8))==1;
        else if (readResReq && fileLines[lineIndex].substr(0,8) == "RES_MESH") pProjectData->meshCharacteristics = str2Int(fileLines[lineIndex].substr(8))==1;
        else if (readResReq && fileLines[lineIndex].substr(0,8) == "RES_STAT") pProjectData->staticPressure      = str2Int(fileLines[lineIndex].substr(8))==1;
        else if (readResReq && fileLines[lineIndex].substr(0,8) == "RES_DYNA") pProjectData->dynamicPressure     = str2Int(fileLines[lineIndex].substr(8))==1;
        else if (readResReq && fileLines[lineIndex].substr(0,8) == "RES_MANO") pProjectData->manometerPressure   = str2Int(fileLines[lineIndex].substr(8))==1;
        else if (readNodes && fileLines[lineIndex].substr(0,5) == "NODES"){
            nodeStartLine = lineIndex+1;
            int nodeNum = str2Int(fileLines[lineIndex].substr(5));
            if (checkEndOfFile(nodeNum, lineIndex, fileLines.size()))
                return;
            nodeIDs.resize(nodeNum);
            nodeCoords.resize(nodeNum);
            for (int nodeIndex=0; nodeIndex<nodeNum; nodeIndex++){
                // get node ID
                nodeIDs[nodeIndex] = nodeShift+nodeIndex+1;
                // get coordinates
                vecStr vecCoords = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
                if (vecCoords.size() != 3){
                    showLineErrorMessage("Node incorrectly defined", lineIndex);
                    return;
                }
                nodeCoords[nodeIndex].resize(3);
                for (int i=0; i<3; i++)
                    nodeCoords[nodeIndex][i] = str2Dbl(vecCoords[i]);
            }
        }
        else if (readElements && fileLines[lineIndex].substr(0,6) == "PANELS"){
            elemStartLine = lineIndex+1;
            uInt elemNum = str2Int(fileLines[lineIndex].substr(6));
            if (checkEndOfFile(elemNum, lineIndex, fileLines.size()))
                return;
            elemIDs.resize(elemNum);
            elemTypes.resize(elemNum);
            elemNodes.resize(elemNum);
            elemNeighbors.resize(elemNum);
            for (uInt elemIndex=0; elemIndex<elemNum; elemIndex++){
                // get element ID
                elemIDs[elemIndex] = elemShift+elemIndex+1;
                // split element line
                vecStr splittedLine = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
                if (splittedLine.size() < 4){
                    showLineErrorMessage("Element incorrectly defined", lineIndex);
                    return;
                }
                // get element type
                elemTypes[elemIndex] = str2Int(splittedLine[0]);
                // get node IDs
                elemNodes[elemIndex].resize(3);
                for (uInt nodeIndex=0; nodeIndex<3; nodeIndex++)
                    elemNodes[elemIndex][nodeIndex] = str2Int(splittedLine[nodeIndex+1]);
                // check type
                uInt neighborsStart = 4;
                if (elemTypes[elemIndex] == 1  ||
                    elemTypes[elemIndex] == 10 ||
                    elemTypes[elemIndex] == 20 ){
                    neighborsStart = 5;
                    if (splittedLine.size() < 5){
                        showLineErrorMessage("Element incorrectly defined", lineIndex);
                        return;
                    }
                    // add fourth node
                    elemNodes[elemIndex].resize(4);
                    elemNodes[elemIndex][3] = str2Int(splittedLine[4]);
                }
                // add neighbors
                elemNeighbors[elemIndex].resize(splittedLine.size()-neighborsStart);
                for (uInt i=neighborsStart,j=0; i<splittedLine.size(); i++,j++)
                    elemNeighbors[elemIndex][j] = str2Int(splittedLine[i]);

                // remove zero elements
                for (uInt i=0; i<elemNeighbors[elemIndex].size(); i++){
                    if (elemNeighbors[elemIndex][i] == 0){
                        elemNeighbors[elemIndex].resize(i);
                        break;
                    }
                }
            }
        }
    }

    // insert nodes into mesh
    for (uInt nodeIndex=0; nodeIndex<nodeIDs.size(); nodeIndex++){
        if (pMesh->insertNodeWithIDCoords(nodeIDs[nodeIndex], &nodeCoords[nodeIndex][0])){
            showLineErrorMessage("Problem inserting node into mesh", nodeIndex+nodeStartLine+1);
            return;
        }
    }

    // insert elements into mesh
    for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
        if (pMesh->insertElemWithIDTypeNodeIDs(elemIDs[elemIndex], elemTypes[elemIndex], elemNodes[elemIndex])){
            showLineErrorMessage("Problem inserting element into mesh", elemIndex+elemStartLine+1);
            return;
        }
    }

    // insert neighboring and trailing info into mesh
    for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
        if (elemTypes[elemIndex] == 1 || elemTypes[elemIndex] == 2){
            if (pMesh->insertNeighborsToElem(elemIDs[elemIndex], elemNeighbors[elemIndex])){
                showLineErrorMessage("Problem inserting neighboring info into mesh", elemIndex+elemStartLine+1);
                return;
            }
        }
        else if (elemTypes[elemIndex] == 10 || elemTypes[elemIndex] == 11){
            if (pMesh->insertTrailsToElem(elemIDs[elemIndex], elemNeighbors[elemIndex])){
                showLineErrorMessage("Problem inserting trailing info into mesh", elemIndex+elemStartLine+1);
                return;
            }
        }
    }
}

//-----------------------------------------------------------------------/
// reads ApameSolver results file for mesh and project data
//-----------------------------------------------------------------------/
void readApameSolverResultMeshAndProjectData(
        CMesh* pMesh,
        CProjectData* pProjectData,
        const string& fileName,
        bool readFlowParams,
        bool readRefData,
        bool readSolverParams,
        bool readNodes,
        bool readElements,
        vecUInt& elemIDs)
{
    // read input file content
    ifstream apameFile(fileName.c_str());
    if (!apameFile.is_open()){
        showErrorMessage("Cannot open ApameSolver results file");
        return;
    }

    // read lines
    vecStr fileLines;
    string line;
    while (getLine(apameFile, line))
        fileLines.push_back(line);
    apameFile.close();

    // check if ApameSolver results file
    if (fileLines[0] != " APAME results file"){
        showLineErrorMessage("Not an ApameSolver results file", 1);
        return;
    }

    // get shift values (maximum IDs of already existing nodes and elements)
    uInt nodeShift = pMesh->getMaxNodeID();
    uInt elemShift = pMesh->getMaxElemID();

    // read content
    uInt numElem = 0;
    uInt numBodyElem = 0;
    uInt numNode = 0;
    uInt numLines = fileLines.size();
    vecUInt nodeIDs;
    vecVecDbl nodeCoords;
    uInt nodeStartLine = 0;
    vecUInt elemTypes;
    vecVecUInt elemNodes;
    vecVecUInt elemNeighbors;
    uInt elemStartLine = 0;
    for (uInt lineIndex=0; lineIndex<numLines; lineIndex++){
        if      (fileLines[lineIndex] == " PANEL_NUM")        {if (readUInt(numElem,     lineIndex, fileLines)) return;}
        else if (fileLines[lineIndex] == " PANEL_NUM_NO_WAKE"){if (readUInt(numBodyElem, lineIndex, fileLines)) return;}
        else if (fileLines[lineIndex] == " NODE_NUM")         {if (readUInt(numNode,     lineIndex, fileLines)) return;}
        else if (readFlowParams && fileLines[lineIndex]==" AIRSPEED")         {if (readFloat  (pProjectData->airspeed, lineIndex, fileLines)) return;}
        else if (readFlowParams && fileLines[lineIndex]==" DENSITY")          {if (readFloat  (pProjectData->density,  lineIndex, fileLines)) return;}
        else if (readFlowParams && fileLines[lineIndex]==" PRESSURE")         {if (readFloat  (pProjectData->pressure, lineIndex, fileLines)) return;}
        else if (readFlowParams && fileLines[lineIndex]==" MACH")             {if (readFloat  (pProjectData->mach,     lineIndex, fileLines)) return;}
        else if (readFlowParams && fileLines[lineIndex]==" CASE_NUM"){
            // store case numbers
            if (checkEndOfFile(3, lineIndex, numLines))
                return;
            uInt caseNumber = str2Int(fileLines[++lineIndex]);
            if (caseNumber <= 0){
                showLineErrorMessage("Case size should be greater then zero", lineIndex+1);
                return;
            }
            // store angles of attack and sideslip angles
            pProjectData->alfa.resize(caseNumber);
            pProjectData->beta.resize(caseNumber);
            vecStr vecLineAlfa = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
            vecStr vecLineBeta = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
            if (vecLineAlfa.size()!=caseNumber || vecLineBeta.size()!=caseNumber){
                showLineErrorMessage("Case number not consistent in ApameSolver results file", lineIndex+1);
                return;
            }
            for (uInt caseIndex=0; caseIndex<caseNumber; caseIndex++){
                pProjectData->alfa[caseIndex] = str2Dbl(vecLineAlfa[caseIndex])*180/pi;
                pProjectData->beta[caseIndex] = str2Dbl(vecLineBeta[caseIndex])*180/pi;
            }
        }
        else if (readRefData && fileLines[lineIndex]==" WINGSPAN"){if (readFloat(pProjectData->wingspan, lineIndex, fileLines)) return;}
        else if (readRefData && fileLines[lineIndex]==" MAC")     {if (readFloat(pProjectData->mac,      lineIndex, fileLines)) return;}
        else if (readRefData && fileLines[lineIndex]==" SURFACE") {if (readFloat(pProjectData->surface,  lineIndex, fileLines)) return;}
        else if (readRefData && fileLines[lineIndex]==" ORIGIN"){
            if (checkEndOfFile(1, lineIndex, numLines))
                return;
            vecStr vecOrigin = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
            if (vecOrigin.size() != 3){
                showLineErrorMessage("Origin incorrectly defined in ApameSolver results file", lineIndex+1);
                return;
            }
            pProjectData->origin.resize(3);
            for (int i=0; i<3; i++)
                pProjectData->origin[i] = str2Dbl(vecOrigin[i]);
        }
        else if (readSolverParams && fileLines[lineIndex]==" METHOD")  {if (readInteger(pProjectData->method,   lineIndex, fileLines)) return;}
        else if (readSolverParams && fileLines[lineIndex]==" ERROR")   {if (readFloat  (pProjectData->error,    lineIndex, fileLines)) return;}
        else if (readSolverParams && fileLines[lineIndex]==" COPL_ANG"){if (readFloat  (pProjectData->coplAng,  lineIndex, fileLines)) return;}
        else if (readSolverParams && fileLines[lineIndex]==" COLLDIST"){if (readFloat  (pProjectData->collDist, lineIndex, fileLines)) return;}
        else if (readSolverParams && fileLines[lineIndex]==" FARFIELD"){if (readFloat  (pProjectData->farField, lineIndex, fileLines)) return;}
        else if (readSolverParams && fileLines[lineIndex]==" COLLCALC"){if (readInteger(pProjectData->collCalc, lineIndex, fileLines)) return;}
        else if (readSolverParams && fileLines[lineIndex]==" VELORDER"){if (readInteger(pProjectData->velOrder, lineIndex, fileLines)) return;}
        else if (readNodes && fileLines[lineIndex++]==" NODES" && numNode){
            nodeStartLine = lineIndex+1;
            if (checkEndOfFile(numNode, lineIndex, numLines))
                return;
            nodeIDs.resize(numNode);
            nodeCoords.resize(numNode);
            for (uInt nodeIndex=0; nodeIndex<numNode; nodeIndex++){
                // get node ID
                nodeIDs[nodeIndex] = nodeShift+nodeIndex+1;
                // get coordinates
                vecStr vecCoords = splitString(change2SingleSpace(fileLines[lineIndex++]), ' ');
                if (vecCoords.size() != 3){
                    showLineErrorMessage("Node incorrectly defined", lineIndex);
                    return;
                }
                nodeCoords[nodeIndex].resize(3);
                for (int i=0; i<3; i++)
                    nodeCoords[nodeIndex][i] = str2Dbl(vecCoords[i]);
            }
        }
        else if (readElements && fileLines[lineIndex++]==" PANELS" && numElem){
            elemStartLine = lineIndex+1;
            if (checkEndOfFile(numElem, lineIndex, numLines))
                return;
            elemIDs.resize(numElem);
            elemTypes.resize(numElem);
            elemNodes.resize(numElem);
            elemNeighbors.resize(numElem);
            for (uInt elemIndex=0; elemIndex<numElem; elemIndex++){
                // get element ID
                elemIDs[elemIndex] = elemShift+elemIndex+1;
                // split element line
                vecStr splittedLine = splitString(change2SingleSpace(fileLines[lineIndex++]), ' ');
                if (splittedLine.size() < 4){
                    showLineErrorMessage("Element incorrectly defined", lineIndex);
                    return;
                }
                // get element type
                elemTypes[elemIndex] = str2Int(splittedLine[0]);
                // get node IDs
                elemNodes[elemIndex].resize(3);
                for (uInt nodeIndex=0; nodeIndex<3; nodeIndex++)
                    elemNodes[elemIndex][nodeIndex] = str2Int(splittedLine[nodeIndex+1]);
                // check type
                uInt neighborsStart = 4;
                if (elemTypes[elemIndex] == 1  ||
                    elemTypes[elemIndex] == 10 ||
                    elemTypes[elemIndex] == 20 ){
                    neighborsStart = 5;
                    if (splittedLine.size() < 5){
                        showLineErrorMessage("Element incorrectly defined", lineIndex);
                        return;
                    }
                    // add fourth node
                    elemNodes[elemIndex][3] = str2Int(splittedLine[4]);
                }
                // add neighbors
                for (uInt i=neighborsStart,j=0; i<splittedLine.size(); i++,j++)
                    elemNeighbors[elemIndex][j] = str2Int(splittedLine[i]);
            }
        }
    }

    // insert nodes into mesh
    for (uInt nodeIndex=0; nodeIndex<nodeIDs.size(); nodeIndex++){
        if (pMesh->insertNodeWithIDCoords(nodeIDs[nodeIndex], &nodeCoords[nodeIndex][0])){
            showLineErrorMessage("Problem inserting node into mesh", nodeIndex+nodeStartLine);
            return;
        }
    }

    // insert elements into mesh
    for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
        if (pMesh->insertElemWithIDTypeNodeIDs(elemIDs[elemIndex], elemTypes[elemIndex], elemNodes[elemIndex])){
            showLineErrorMessage("Problem inserting element into mesh", elemIndex+elemStartLine);
            return;
        }
    }

    // insert neighboring and trailing info into mesh
    for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
        if (elemTypes[elemIndex] == 1 || elemTypes[elemIndex] == 2){
            if (pMesh->insertNeighborsToElem(elemIDs[elemIndex], elemNeighbors[elemIndex])){
                showLineErrorMessage("Problem inserting neighboring info into mesh", elemIndex+elemStartLine);
                return;
            }
        }
        else if (elemTypes[elemIndex] == 10 || elemTypes[elemIndex] == 11){
            if (pMesh->insertTrailsToElem(elemIDs[elemIndex], elemNeighbors[elemIndex])){
                showLineErrorMessage("Problem inserting trailing info into mesh", elemIndex+elemStartLine);
                return;
            }
        }
    }
}

//-----------------------------------------------------------------------/
// reads results only from ApameSolver results file
//-----------------------------------------------------------------------/
int readApameSolverResults(
        CResults* pResults,
        const string& resFileName,
        vecUInt& elemIDs)
{
    // open ApameSolver results file
    ifstream apameFile(resFileName.c_str());
    if (!apameFile.is_open()){
        showErrorMessage("Cannot open ApameSolver results file");
        return -1;
    }

    // read input file content
    vecStr fileLines;
    string line;
    while (getLine(apameFile, line))
        fileLines.push_back(line);
    apameFile.close();

    // check if ApameSolver results file
    if (fileLines[0] != " APAME results file"){
        showLineErrorMessage("Not an ApameSolver results file", 1);
        return -1;
    }

    // read content of results file
    uInt numLines = fileLines.size();
    vecStr caseTitles;   // case titles composed of angle of attack and sideslip angle combinations
    for (uInt lineIndex=0; lineIndex<numLines; lineIndex++){
        if (fileLines[lineIndex] == " APAME results file"){
            // store date and time information
            if (checkEndOfFile(1, lineIndex, numLines))
                return -1;
            vecStr dateAndTime;
            dateAndTime = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
            if (dateAndTime.size() != 4){
                showLineErrorMessage("Time and date error in ApameSolver results file", lineIndex+1);
                return -1;
            }
            pResults->date = dateAndTime[1];
            pResults->time = dateAndTime[3];
        }
        else if (fileLines[lineIndex] == " VERSION"){
            if (checkEndOfFile(1, lineIndex, numLines))
                return -1;
            pResults->solverVersion = removeLeadingSpaces(fileLines[++lineIndex]);
        }
        else if (fileLines[lineIndex] == " PANEL_NUM")        {if (readUInt (pResults->numElem,     lineIndex, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " PANEL_NUM_NO_WAKE"){if (readUInt (pResults->numBodyElem, lineIndex, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " NODE_NUM")         {if (readUInt (pResults->numNode,     lineIndex, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " AIRSPEED")         {if (readFloat(pResults->airspeed,    lineIndex, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " DENSITY")          {if (readFloat(pResults->density,     lineIndex, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " PRESSURE")         {if (readFloat(pResults->pressure,    lineIndex, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " MACH")             {if (readFloat(pResults->mach,        lineIndex, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " CASE_NUM"){
            // store case numbers
            if (checkEndOfFile(3, lineIndex, numLines))
                return -1;
            pResults->caseNumber = str2Int(fileLines[++lineIndex]);
            if (pResults->caseNumber <= 0){
                showLineErrorMessage("Case size should be greater then zero", lineIndex+1);
                return -1;
            }
            // store angles of attack and sideslip angles
            caseTitles.resize(pResults->caseNumber);
            pResults->alfa.resize(pResults->caseNumber);
            pResults->beta.resize(pResults->caseNumber);
            vecStr vecLineAlfa = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
            vecStr vecLineBeta = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
            if (vecLineAlfa.size()!=pResults->caseNumber || vecLineBeta.size()!=pResults->caseNumber){
                showLineErrorMessage("Case number not consistent in ApameSolver results file", lineIndex+1);
                return -1;
            }
            for (uInt caseIndex=0; caseIndex<pResults->caseNumber; caseIndex++){
                pResults->alfa[caseIndex] = str2Dbl(vecLineAlfa[caseIndex])*180/pi;
                pResults->beta[caseIndex] = str2Dbl(vecLineBeta[caseIndex])*180/pi;
                // store title for result fields
                caseTitles[caseIndex] = "Alpha="+dbl2Str(pResults->alfa[caseIndex]) +
                                        " Beta="+dbl2Str(pResults->beta[caseIndex]) ;
            }
        }
        else if (fileLines[lineIndex] == " WINGSPAN"){if (readFloat(pResults->wingspan, lineIndex, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " MAC")     {if (readFloat(pResults->mac,      lineIndex, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " SURFACE") {if (readFloat(pResults->surface,  lineIndex, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " ORIGIN"){
            if (checkEndOfFile(1, lineIndex, numLines))
                return -1;
            vecStr vecOrigin = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
            if (vecOrigin.size() != 3){
                showLineErrorMessage("Origin incorrectly defined in ApameSolver results file", lineIndex+1);
                return -1;
            }
            pResults->origin.resize(3);
            for (int i=0; i<3; i++)
                pResults->origin[i] = str2Dbl(vecOrigin[i]);
        }
        else if (fileLines[lineIndex] == " METHOD")  {if (readUInt (pResults->method,   lineIndex, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " ERROR")   {if (readFloat(pResults->error,    lineIndex, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " COPL_ANG"){if (readFloat(pResults->coplAng,  lineIndex, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " COLLDIST"){if (readFloat(pResults->collDist, lineIndex, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " FARFIELD"){if (readFloat(pResults->farField, lineIndex, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " COLLCALC"){if (readUInt (pResults->collCalc, lineIndex, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " VELORDER"){if (readUInt (pResults->velOrder, lineIndex, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " CX")   {if (readCoefForce(pResults->Cx,    lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " CY")   {if (readCoefForce(pResults->Cy,    lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " CZ")   {if (readCoefForce(pResults->Cz,    lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " CL")   {if (readCoefForce(pResults->Cl,    lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " CM")   {if (readCoefForce(pResults->Cm,    lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " CN")   {if (readCoefForce(pResults->Cn,    lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " CDRAG"){if (readCoefForce(pResults->Cdrag, lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " CSIDE"){if (readCoefForce(pResults->Cside, lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " CLIFT"){if (readCoefForce(pResults->Clift, lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " FX")   {if (readCoefForce(pResults->Fx,    lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " FY")   {if (readCoefForce(pResults->Fy,    lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " FZ")   {if (readCoefForce(pResults->Fz,    lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " FL")   {if (readCoefForce(pResults->Fl,    lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " FM")   {if (readCoefForce(pResults->Fm,    lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " FN")   {if (readCoefForce(pResults->Fn,    lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " FDRAG"){if (readCoefForce(pResults->Fdrag, lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " FSIDE"){if (readCoefForce(pResults->Fside, lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " FLIFT"){if (readCoefForce(pResults->Flift, lineIndex, pResults->caseNumber, fileLines)) return -1;}
        else if (fileLines[lineIndex] == " S")          {if (readSurfaceData(pResults, "Panel Surface Area",   elemIDs, lineIndex, fileLines                     )) return -1;}
        else if (fileLines[lineIndex] == " FF")         {if (readSurfaceData(pResults, "Far-Field Value",      elemIDs, lineIndex, fileLines                     )) return -1;}
        else if (fileLines[lineIndex] == " VELOCITY")   {if (readSurfaceData(pResults, "Velocity Magnitude",   elemIDs, lineIndex, fileLines, pResults->caseNumber, caseTitles)) return -1;}
        else if (fileLines[lineIndex] == " CP")         {if (readSurfaceData(pResults, "Pressure Coefficient", elemIDs, lineIndex, fileLines, pResults->caseNumber, caseTitles)) return -1;}
        else if (fileLines[lineIndex] == " DIPOLE")     {if (readSurfaceData(pResults, "Doublet Strengths",    elemIDs, lineIndex, fileLines, pResults->caseNumber, caseTitles)) return -1;}
        else if (fileLines[lineIndex] == " SOURCE")     {if (readSurfaceData(pResults, "Source Strengths",     elemIDs, lineIndex, fileLines, pResults->caseNumber, caseTitles)) return -1;}
        else if (fileLines[lineIndex] == " P_STATIC")   {if (readSurfaceData(pResults, "Static Pressure",      elemIDs, lineIndex, fileLines, pResults->caseNumber, caseTitles)) return -1;}
        else if (fileLines[lineIndex] == " P_DYNAMIC")  {if (readSurfaceData(pResults, "Dynamic Pressure",     elemIDs, lineIndex, fileLines, pResults->caseNumber, caseTitles)) return -1;}
        else if (fileLines[lineIndex] == " P_MANOMETER"){if (readSurfaceData(pResults, "Manometer Pressure",   elemIDs, lineIndex, fileLines, pResults->caseNumber, caseTitles)) return -1;}
        else if (fileLines[lineIndex] == " NORMAL"){
            if (checkEndOfFile(elemIDs.size(), lineIndex, numLines))
                return -1;

            vecVecDbl normal(3);
            for (uInt i=0; i<3; i++)
                normal[i].resize(elemIDs.size());

            for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
                vecStr lineData = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
                if (lineData.size() != 3){
                    showLineErrorMessage("Incorrect number of element normal components", lineIndex+1);
                    return -1;
                }
                for (uInt i=0; i<3; i++) normal[i][elemIndex] = str2Dbl(lineData[i]);
            }

            vecStr dataNames(3);
            dataNames[0] = "Panel Normal X-component";
            dataNames[1] = "Panel Normal Y-component";
            dataNames[2] = "Panel Normal Z-component";

            for (uInt i=0; i<3; i++){
                pResults->surfDatas.resize(pResults->surfDatas.size()+1);
                pResults->surfDatas.back().dataName = dataNames[i];
                pResults->surfDatas.back().caseTitles.push_back("Case Independent");
                pResults->surfDatas.back().data.push_back(normal[i]);
                pResults->surfDatas.back().elemIDs = elemIDs;
            }
        }
        else if (fileLines[lineIndex] == " VELOCITY_COMP"){
            if (checkEndOfFile(elemIDs.size()*pResults->caseNumber, lineIndex, numLines))
                return -1;

            vecVecVecDbl velocityComponents(3);
            for (uInt i=0; i<3; i++){
                velocityComponents[i].resize(pResults->caseNumber);
                for (uInt caseIndex=0; caseIndex<pResults->caseNumber; caseIndex++)
                    velocityComponents[i][caseIndex].resize(elemIDs.size());
            }

            for (uInt caseIndex=0; caseIndex<pResults->caseNumber; caseIndex++){
                for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
                    vecStr lineData = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
                    if (lineData.size() != 3){
                        showLineErrorMessage("Incorrect number of velocity components", lineIndex+1);
                        return -1;
                    }
                    for (uInt i=0; i<3; i++) velocityComponents[i][caseIndex][elemIndex] = str2Dbl(lineData[i]);
                }
            }

            vecStr dataNames(3);
            dataNames[0] = "Velocity X-component";
            dataNames[1] = "Velocity Y-component";
            dataNames[2] = "Velocity Z-component";

            for (uInt i=0; i<3; i++){
                pResults->surfDatas.resize(pResults->surfDatas.size()+1);
                pResults->surfDatas.back().dataName = dataNames[i];
                pResults->surfDatas.back().caseTitles = caseTitles;
                pResults->surfDatas.back().data = velocityComponents[i];
                pResults->surfDatas.back().elemIDs = elemIDs;
            }
        }
    }

    return 0;
}

//-----------------------------------------------------------------------/
// reads integer data from results file
//-----------------------------------------------------------------------/
int readInteger(
        int& storage,
        uInt& lineIndex,
        const vecStr& fileLines)
{
    if (checkEndOfFile(1, lineIndex, fileLines.size()))
        return -1;

    storage = str2Int(fileLines[++lineIndex]);

    return 0;
}

//-----------------------------------------------------------------------/
// reads unsigned integer data from results file
//-----------------------------------------------------------------------/
int readUInt(
        uInt& storage,
        uInt& lineIndex,
        const vecStr& fileLines)
{
    if (checkEndOfFile(1, lineIndex, fileLines.size()))
        return -1;

    storage = str2Int(fileLines[++lineIndex]);

    return 0;
}

//-----------------------------------------------------------------------/
// reads float data from results file
//-----------------------------------------------------------------------/
int readFloat(
        double& storage,
        uInt& lineIndex,
        const vecStr& fileLines)
{
    if (checkEndOfFile(1, lineIndex, fileLines.size()))
        return -1;

    storage = str2Dbl(fileLines[++lineIndex]);

    return 0;
}

//-----------------------------------------------------------------------/
// reads force and coefficient data from results file
//-----------------------------------------------------------------------/
int readCoefForce(
        vecDbl& storage,
        uInt& lineIndex,
        const uInt& caseNum,
        const vecStr& fileLines)
{
    if (checkEndOfFile(1, lineIndex, fileLines.size()))
        return -1;
    vecStr vecData = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
    if (vecData.size() != caseNum){
        showLineErrorMessage("Data not matching number of cases", lineIndex+1);
        return -1;
    }
    storage.resize(caseNum);
    for (uInt caseIndex=0; caseIndex<caseNum; caseIndex++)
        storage[caseIndex] = str2Dbl(vecData[caseIndex]);

    return 0;
}

//-----------------------------------------------------------------------/
// reads surface data from results file
//-----------------------------------------------------------------------/
int readSurfaceData(
        CResults* pResults,
        const string& dataName,
        const vecUInt& elemIDs,
        uInt& lineIndex,
        const vecStr& fileLines,
        const uInt& caseNum,
        const vecStr& caseTitles)
{
    if (checkEndOfFile(elemIDs.size(), lineIndex, fileLines.size()))
        return -1;
    pResults->surfDatas.resize(pResults->surfDatas.size()+1);
    pResults->surfDatas.back().dataName = dataName;
    if (caseTitles.empty()) pResults->surfDatas.back().caseTitles.push_back("Case Independent");
    else                    pResults->surfDatas.back().caseTitles = caseTitles;
    pResults->surfDatas.back().data.resize(caseNum);
    for (uInt caseIndex=0; caseIndex<caseNum; caseIndex++)
        pResults->surfDatas.back().data[caseIndex].resize(elemIDs.size());
    for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
        vecStr lineData = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
        if (lineData.size() != caseNum){
            showLineErrorMessage("Number of element results does not match case number", lineIndex+1);
            return -1;
        }
        for (uInt caseIndex=0; caseIndex<caseNum; caseIndex++)
            pResults->surfDatas.back().data[caseIndex][elemIndex] = str2Dbl(lineData[caseIndex]);
    }
    pResults->surfDatas.back().elemIDs = elemIDs;

    return 0;
}

//-----------------------------------------------------------------------/
// reads Abaqus input file into mesh
//-----------------------------------------------------------------------/
void readAbaqusFile(
        CMesh* pMesh,
        const string& fileName,
        const long& shiftNodeIDs,
        const long& shiftElemIDs)
{
    // open file
    ifstream inputFile(fileName.c_str());
    if (!inputFile.is_open()){
        showErrorMessage("Cannot open Abaqus file");
        return;
    }

    // read lines with stripped comments
    vecStr fileLines;
    string line;
    while (getLine(inputFile, line))
        fileLines.push_back(line.substr(0, line.find("**")));
    inputFile.close();

    // read nodes
    bool nodefound = false;
    for (uInt lineIndex=0; lineIndex<fileLines.size(); lineIndex++){
        // check if node definition card
        vecStr vecNodeLine = splitString(fileLines[lineIndex], ',');
        if (vecNodeLine.empty())
            continue;

        if (convert2UpperCase(removeLeadingSpaces(vecNodeLine[0])) != "*NODE")
            continue;

        // node definition table found so read nodes
        while (++lineIndex < fileLines.size()){
            // quit reading node definition table if new keyword found
            if (removeLeadingSpaces(fileLines[lineIndex]).substr(0,1) == "*"){
                lineIndex--;
                break;
            }

            // skip empty lines
            if (fileLines[lineIndex].empty() || change2SingleSpace(fileLines[lineIndex])==" ")
                continue;

            // read node line components
            vecStr vecLine = splitString(change2SingleSpace(fileLines[lineIndex]), ',');
            if (vecLine.size() != 4){
                showLineErrorMessage("Incorrect node definition", lineIndex);
                return;
            }

            // read node ID
            uInt nodeID = atoi(vecLine[0].c_str())+shiftNodeIDs;
            if (nodeID <= 0){
                showLineErrorMessage("Node ID should be greater than 0", lineIndex);
                return;
            }

            nodefound = true;

            // read node coordinates
            double coordinates[3];
            for (int i=0; i<3; i++)
                coordinates[i] = atof(vecLine[i+1].c_str());

            int retCode = pMesh->insertNodeWithIDCoords(nodeID, coordinates);
            if (retCode == -2){
                showLineErrorMessage("Node ID already exists", lineIndex);
                return;
            }
            else if (retCode == -3){
                showLineErrorMessage("Node ID is equal 0", lineIndex);
                return;
            }
        }
    }

    // check if something was read
    if (!nodefound)
        showWarningMessage("Zero nodes found");

    // read elements
    bool elemfound = false;
    for (uInt lineIndex=0; lineIndex<fileLines.size(); lineIndex++){
        // check if element definition card
        if (convert2UpperCase(removeLeadingSpaces(fileLines[lineIndex])).substr(0,8) != "*ELEMENT")
            continue;

        // determine element type
        size_t found = convert2UpperCase(fileLines[lineIndex]).find("TYPE");
        if (found == string::npos){
            showLineErrorMessage("Missing element type definition", lineIndex);
            return;
        }

        // check type
        uInt elemType = 0;
        if      (convert2UpperCase(fileLines[lineIndex].substr(found+5,2)) == "S4")
            elemType = 1;
        else if (convert2UpperCase(fileLines[lineIndex].substr(found+5,2)) == "S3")
            elemType = 2;
        else
            continue;

        // element definition table found so read elements
        while (++lineIndex < fileLines.size()){
            // quit reading element definition table if new keyword found
            if (removeLeadingSpaces(fileLines[lineIndex]).substr(0,1) == "*"){
                lineIndex--;
                break;
            }

            // skip empty lines
            if (fileLines[lineIndex].empty() || change2SingleSpace(fileLines[lineIndex])==" ")
                continue;

            // read element line components
            vecStr vecLine = splitString(change2SingleSpace(fileLines[lineIndex]), ',');
            if ((elemType==1 && vecLine.size()<5) ||
                (elemType==2 && vecLine.size()<4) ){
                showLineErrorMessage("Incorrect element definition", lineIndex);
                return;
            }

            // read element ID
            uInt elementID = atoi(vecLine[0].c_str())+shiftElemIDs;
            if (elementID <= 0){
                showLineErrorMessage("Element ID should be greater than 0", lineIndex);
                return;
            }

            elemfound = true;

            // read node IDs
            vecUInt nodeIDs(3);
            if (elemType == 1) nodeIDs.resize(4);
            for (uInt i=0; i<nodeIDs.size(); i++)
                nodeIDs[i] = atoi(vecLine[i+1].c_str())+shiftNodeIDs;

            // insert element into mesh
            int retVal = pMesh->insertElemWithIDTypeNodeIDs(elementID, elemType, nodeIDs);

            // process return value
            string message = "Error in element definition";
            if      (retVal == -2) message = "Duplicate element ID detected";
            else if (retVal == -3) message = "Element ID 0 detected";
            else if (retVal == -4) message = "Duplicate node ID(s) given in element definition";
            else if (retVal == -5) message = "Zero surface element detected";
            else if (retVal == -6) message = "Element refers to non-existing node ID";
            if (retVal != 0){
                showLineErrorMessage(message, lineIndex);
                return;
            }
        }
    }

    // check if something was read
    if (!elemfound)
        showWarningMessage("Zero elements found");
}

//-----------------------------------------------------------------------/
// reads Fluent input file into mesh
//-----------------------------------------------------------------------/
void readFluentFile(
        CMesh* pMesh,
        const string& fileName,
        const long& shiftNodeIDs,
        const long& shiftElemIDs)
{
    // open file
    ifstream inputFile(fileName.c_str());
    if (!inputFile.is_open()){
        showErrorMessage("Cannot open Fluent file");
        return;
    }

    // read lines with stripped comments
    vecStr fileLines;
    string line;
    while (getLine(inputFile, line))
        fileLines.push_back(line.substr(0, line.find("**")));
    inputFile.close();

    // read nodes
    bool nodefound = false;
    for (uInt lineIndex=0; lineIndex<fileLines.size(); lineIndex++){
        if (convert2UpperCase(removeLeadingSpaces(fileLines[lineIndex])).substr(0,17) != "NODAL COORDINATES")
            continue;

        // node definition table found so read nodes
        while (++lineIndex < fileLines.size()){
            // quit reading node definition table if ENDOFSECTION keyword found
            if (convert2UpperCase(removeLeadingSpaces(fileLines[lineIndex])).substr(0,12) == "ENDOFSECTION"){
                lineIndex--;
                break;
            }

            // skip empty lines
            if (fileLines[lineIndex].empty() || change2SingleSpace(fileLines[lineIndex])==" ")
                continue;

            // read node line components
            vecStr vecLine = splitString(change2SingleSpace(fileLines[lineIndex]), ',');
            if (vecLine.size() != 4){
                showLineErrorMessage("Incorrect node definition", lineIndex);
                return;
            }

            // read node ID
            uInt nodeID = atoi(vecLine[0].c_str())+shiftNodeIDs;
            if (nodeID <= 0){
                showLineErrorMessage("Node ID should be greater than 0", lineIndex);
                return;
            }

            nodefound = true;

            // read node coordinates
            double coordinates[3];
            for (int i=0; i<3; i++)
                coordinates[i] = atof(vecLine[i+1].c_str());

            int retCode = pMesh->insertNodeWithIDCoords(nodeID, coordinates);
            if (retCode == -2){
                showLineErrorMessage("Node ID already exists", lineIndex);
                return;
            }
            else if (retCode == -3){
                showLineErrorMessage("Node ID is equal 0", lineIndex);
                return;
            }
        }
    }

    // check if something was read
    if (!nodefound)
        showWarningMessage("Zero nodes found");

    // read elements
    bool elemfound = false;
    for (uInt lineIndex=0; lineIndex<fileLines.size(); lineIndex++){
        // check if element definition card
        if (convert2UpperCase(removeLeadingSpaces(fileLines[lineIndex])).substr(0,14) != "ELEMENTS/CELLS")
            continue;

        // element definition table found so read elements
        while (++lineIndex < fileLines.size()){
            // quit reading element definition table if ENDOFSECTION keyword found
            if (convert2UpperCase(removeLeadingSpaces(fileLines[lineIndex])).substr(0,12) == "ENDOFSECTION"){
                lineIndex--;
                break;
            }

            // skip empty lines
            if (fileLines[lineIndex].empty() || change2SingleSpace(fileLines[lineIndex])==" ")
                continue;

            // check element type
            vecStr vecLine = splitString(change2SingleSpace(fileLines[lineIndex]), ' ');
            uInt elementID;
            uInt elemType;
            vecUInt nodeIDs;
            if (vecLine.size()==7 && vecLine[2]=="4"){
                elemType = 1;

                // read element ID
                elementID = atoi(vecLine[0].c_str())+shiftElemIDs;
                if (elementID <= 0){
                    showLineErrorMessage("Element ID should be greater than 0", lineIndex);
                    return;
                }

                elemfound = true;

                // read node IDs
                nodeIDs.resize(4);
                for (uInt i=0; i<4; i++)
                    nodeIDs[i] = atoi(vecLine[i+3].c_str())+shiftNodeIDs;
            }
            else if (vecLine.size()==6 && vecLine[2]=="3"){
                elemType = 2;

                // read element ID
                elementID = atoi(vecLine[0].c_str())+shiftElemIDs;
                if (elementID <= 0){
                    showLineErrorMessage("Element ID should be greater than 0", lineIndex);
                    return;
                }

                elemfound = true;

                // read node IDs
                nodeIDs.resize(3);
                for (uInt i=0; i<3; i++)
                    nodeIDs[i] = atoi(vecLine[i+3].c_str())+shiftNodeIDs;
            }

            // insert element into mesh
            int retVal = pMesh->insertElemWithIDTypeNodeIDs(elementID, elemType, nodeIDs);

            // process return value
            string message = "Error in element definition";
            if      (retVal == -2) message = "Duplicate element ID detected";
            else if (retVal == -3) message = "Element ID 0 detected";
            else if (retVal == -4) message = "Duplicate node ID(s) given in element definition";
            else if (retVal == -5) message = "Zero surface element detected";
            else if (retVal == -6) message = "Element refers to non-existing node ID";
            if (retVal != 0){
                showLineErrorMessage(message, lineIndex);
                return;
            }
        }
    }

    // check if something was read
    if (!elemfound)
        showWarningMessage("Zero elements found");
}

//-----------------------------------------------------------------------/
// reads Ideas input file into mesh
//-----------------------------------------------------------------------/
void readIdeasFile(
        CMesh* pMesh,
        const string& fileName,
        const long& shiftNodeIDs,
        const long& shiftElemIDs)
{
    // open file
    ifstream inputFile(fileName.c_str());
    if (!inputFile.is_open()){
        showErrorMessage("Cannot open Ideas file");
        return;
    }

    // read lines with stripped comments
    vecStr fileLines;
    string line;
    while (getLine(inputFile, line))
        fileLines.push_back(line.substr(0, line.find("**")));
    inputFile.close();

    // read nodes
    bool nodefound = false;
    for (uInt lineIndex=0; lineIndex<fileLines.size(); lineIndex++){
        if (atoi(fileLines[lineIndex].c_str()) == 2411){
            // node definition table encountered so start reading nodes
            while (true){
                if (checkEndOfFile(1, lineIndex, fileLines.size()))
                    break;

                // read node ID line
                int nodeID = atoi(fileLines[++lineIndex].c_str());

                // check if end of node definition table encountered
                if (nodeID == -1)
                    break;

                nodefound = true;

                nodeID += shiftNodeIDs;

                vecStr vecLine = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
                if (vecLine.size() != 3){
                    showLineErrorMessage("Incorrect node definition", lineIndex);
                    return;
                }

                // read node coordinates
                double coordinates[3];
                for (int i=0; i<3; i++)
                    coordinates[i] = atof(vecLine[i].c_str());

                // add node to mesh
                int retCode = pMesh->insertNodeWithIDCoords(uInt(nodeID), coordinates);
                if (retCode == -2){
                    showLineErrorMessage("Node ID already exists", lineIndex);
                    return;
                }
                else if (retCode == -3){
                    showLineErrorMessage("Node ID is equal 0", lineIndex);
                    return;
                }
            }
        }
    }

    // check if something was read
    if (!nodefound)
        showWarningMessage("Zero nodes found");

    // read elements
    bool elemfound = false;
    for (uInt lineIndex=0; lineIndex<fileLines.size(); lineIndex++){
        if (atoi(fileLines[lineIndex].c_str()) == 2412){
            // element definition table encountered so start reading elements
            while (true){
                if (checkEndOfFile(1, lineIndex, fileLines.size()))
                    break;

                vecStr vecLine = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
                if (vecLine.size() < 2){
                    showLineErrorMessage("Incorrect element definition", lineIndex);
                    return;
                }
                int elemID = str2Int(vecLine[0]);
                uInt IdeasElemType = str2Int(vecLine[1]);

                // check if end of element definition table encountered
                if (elemID == -1)
                    break;

                // check element type
                vecUInt nodeIDs;
                uInt elemType;
                if (IdeasElemType == 41){
                    elemType = 2;

                    // read constituing nodes
                    vecLine = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
                    if (vecLine.size() < 3){
                        showLineErrorMessage("Incorrect element definition", lineIndex);
                        return;
                    }

                    elemfound = true;

                    // read node IDs
                    nodeIDs.resize(3);
                    for (uInt i=0; i<3; i++)
                        nodeIDs[i] = atoi(vecLine[i+1].c_str())+1+shiftNodeIDs;
                }
                else if (IdeasElemType == 44){
                    elemType = 1;

                    // read constituing nodes
                    vecLine = splitString(change2SingleSpace(fileLines[++lineIndex]), ' ');
                    if (vecLine.size() < 4){
                        showLineErrorMessage("Incorrect element definition", lineIndex);
                        return;
                    }
                    elemfound = true;

                    // read node IDs
                    nodeIDs.resize(4);
                    for (uInt i=0; i<4; i++)
                        nodeIDs[i] = atoi(vecLine[i+1].c_str())+1+shiftNodeIDs;
                }
                // skip lines if other type of elements
                else if (IdeasElemType == 11 ) lineIndex = lineIndex+2;
                else if (IdeasElemType == 111) lineIndex++;
                else                           lineIndex++; // assume that only one line needs to be skipped

                // insert element into mesh
                int retVal = pMesh->insertElemWithIDTypeNodeIDs(uInt(elemID)+shiftElemIDs, elemType, nodeIDs);

                // process return value
                string message = "Error in element definition";
                if      (retVal == -2) message = "Duplicate element ID detected";
                else if (retVal == -3) message = "Element ID 0 detected";
                else if (retVal == -4) message = "Duplicate node ID(s) given in element definition";
                else if (retVal == -5) message = "Zero surface element detected";
                else if (retVal == -6) message = "Element refers to non-existing node ID";
                if (retVal != 0){
                    showLineErrorMessage(message, lineIndex);
                    return;
                }
            }
        }
    }

    // check if something was read
    if (!elemfound)
        showWarningMessage("Zero elements found");
}

//-----------------------------------------------------------------------/
// reads Nastran input file into mesh
//-----------------------------------------------------------------------/
void readNastranFile(
        CMesh* pMesh,
        const string& fileName,
        const long& shiftNodeIDs,
        const long& shiftElemIDs)
{
    // open file
    ifstream inputFile(fileName.c_str());
    if (!inputFile.is_open()){
        showErrorMessage("Cannot open Nastran file");
        return;
    }

    // read lines with stripped comments
    vecStr fileLines;
    string line;
    while (getLine(inputFile, line))
        fileLines.push_back(line.substr(0, line.find('$')));
    inputFile.close();

    // read nodes first
    bool nodefound = false;
    for (uInt i=0; i<fileLines.size(); i++){
        // check if this is node line
        if (convert2UpperCase(removeLeadingSpaces(fileLines[i])).substr(0,4) == "GRID"){
            nodefound = true;

            // declare node definition strings
            string sNodeID;
            vecStr coords(3);

            // check if free format: GRID, ID, , X, Y, Z, $ this is a comment
            if (fileLines[i].find(',') != string::npos){
                // split free format line
                vecStr lineParts = splitString(fileLines[i], ',');

                // check number of line parts
                if (lineParts.size() < 6){
                    showLineErrorMessage("Improper free format grid definition", i+1);
                    return;
                }
                sNodeID = lineParts[1];
                coords[0] = lineParts[3];
                coords[1] = lineParts[4];
                coords[2] = lineParts[5];
            }
            // check if small format (by searching of "*" in first field)
            else if (fileLines[i].substr(0,8).find('*') == string::npos){
                // check line size
                if (fileLines[i].size() <= 40){
                    showLineErrorMessage("Line too short for node definition", i+1);
                    return;
                }
                sNodeID = fileLines[i].substr(8,8);
                coords[0] = fileLines[i].substr(24,8);
                coords[1] = fileLines[i].substr(32,8);
                coords[2] = fileLines[i].substr(40,8);
            }
            // large format is what remains
            else{
                // check line size
                if (fileLines[i].size() <= 56){
                    showLineErrorMessage("Line too short for node definition", i+1);
                    return;
                }
                sNodeID = fileLines[i].substr(8,16);
                coords[0] = fileLines[i].substr(40,16);
                coords[1] = fileLines[i].substr(56,16);
                // skip to next line for additional Z coordinate
                if (fileLines.size() == ++i){
                    showLineErrorMessage("Premature end of file", i+1);
                    return;
                }
                if (fileLines[i].size() <= 8){
                    showLineErrorMessage("Line too short for node definition", i+1);
                    return;
                }
                coords[2] = fileLines[i].substr(8,16);
            }

            // extract data from strings
            uInt nodeID = atoi(sNodeID.c_str())+shiftNodeIDs;
            if (nodeID <= 0){
                showLineErrorMessage("Node ID should be greater than 0", i+1);
                return;
            }
            double coordinates[3];
            coordinates[0] = readNastranSci(coords[0]);
            coordinates[1] = readNastranSci(coords[1]);
            coordinates[2] = readNastranSci(coords[2]);

            // insert node into mesh
            int retCode = pMesh->insertNodeWithIDCoords(nodeID, coordinates);
            if (retCode == -1){
                showLineErrorMessage("Node ID already exists", i+1);
                return;
            }
            else if (retCode != 0){
                showLineErrorMessage("Unknown error reading node", i+1);
                return;
            }
        }
    }

    // check if something was read
    if (!nodefound)
        showWarningMessage("Zero nodes found");

    // read elements
    bool elemfound = false;
    for (uInt i=0; i<fileLines.size(); i++){
        // check if this is (supported) element line
        if (convert2UpperCase(removeLeadingSpaces(fileLines[i])).substr(0,6) == "CTRIA3" ||
            convert2UpperCase(removeLeadingSpaces(fileLines[i])).substr(0,6) == "CQUAD4" ){
            elemfound = true;

            // declare element definition strings
            string sElemID;
            vecStr sNodeIDs;

            // check if free format
            if (fileLines[i].find(',') != string::npos){
                // split free format line
                vecStr lineParts = splitString(fileLines[i], ',');

                // add data common to both element types
                if (lineParts.size() > 5){
                    sElemID = lineParts[1];
                    sNodeIDs.resize(3);
                    sNodeIDs[0] = lineParts[3];
                    sNodeIDs[1] = lineParts[4];
                    sNodeIDs[2] = lineParts[5];
                }

                // check if quad type and split line size adequate
                if (convert2UpperCase(removeLeadingSpaces(fileLines[i])).substr(0,6) == "CQUAD4"){
                    if (lineParts.size() > 6){
                        // add additional node
                        sNodeIDs.resize(4);
                        sNodeIDs[3] = lineParts[6];
                    }
                    else{
                        showLineErrorMessage("Improper free format element definition", i+1);
                        return;
                    }
                }
            }
            // check if small format (by searching of "*" in first field)
            else if (fileLines[i].substr(0,8).find('*') == string::npos){
                // check line size
                if (fileLines[i].size() <= 40){
                    showLineErrorMessage("Line too short for element definition", i+1);
                    return;
                }
                sElemID = fileLines[i].substr(8,8);
                sNodeIDs.resize(3);
                sNodeIDs[0] = fileLines[i].substr(24,8);
                sNodeIDs[1] = fileLines[i].substr(32,8);
                sNodeIDs[2] = fileLines[i].substr(40,8);

                // check if quad type and split line size adequate
                if (convert2UpperCase(removeLeadingSpaces(fileLines[i])).substr(0,6)=="CQUAD4" && fileLines[i].size()>48){
                    // add additional node
                    sNodeIDs.resize(4);
                    sNodeIDs[3] = fileLines[i].substr(48,8);
                }
            }
            // large format is what remains
            else{
                // check line size
                if (fileLines[i].size() <= 56){
                    showLineErrorMessage("Line too short for element definition", i+1);
                    return;
                }
                sElemID = fileLines[i].substr(8,16);
                sNodeIDs.resize(3);
                sNodeIDs[0] = fileLines[i].substr(40,16);
                sNodeIDs[1] = fileLines[i].substr(56,16);
                // skip to next line for additional node ID
                if (fileLines.size() == ++i){   // LINE INCREMENT HERE
                    showLineErrorMessage("Premature end of file", i+1);
                    return;
                }
                if (fileLines[i].size() <= 8){
                    showLineErrorMessage("Line too short for element definition", i+1);
                    return;
                }
                sNodeIDs[2] = fileLines[i].substr(8,16);
                if (convert2UpperCase(removeLeadingSpaces(fileLines[i])).substr(0,6)=="CQUAD4" && fileLines[i].size()>32){
                    sNodeIDs.resize(4);
                    sNodeIDs[3] = fileLines[i].substr(32,16);
                }
                else{
                    showLineErrorMessage("Line too short for element definition", i+1);
                    return;
                }
            }

            // convert data from strings
            vecUInt nodeIDs(sNodeIDs.size());
            for (uInt j=0; j<sNodeIDs.size(); j++)
                nodeIDs[j] = str2Int(sNodeIDs[j])+shiftNodeIDs;

            // insert element into mesh
            int retVal = pMesh->insertElemWithIDNodeIDs(str2Int(sElemID)+shiftElemIDs, nodeIDs);

            // process return value
            string message;
            if      (retVal == -1) message = "Duplicate element ID detected";
            else if (retVal == -2) message = "Element ID 0 detected";
            else if (retVal == -4) message = "Duplicate node ID(s) given in element definition";
            else if (retVal == -5) message = "Element refers to non-existing node ID(s)";
            else if (retVal == -6) message = "Zero surface element detected";
            if (retVal != 0){
                showLineErrorMessage(message, i+1);
                return;
            }
        }
    }

    // check if something was read
    if (!elemfound)
        showWarningMessage("Zero elements found");
}

//-----------------------------------------------------------------------/
// converts Nastran scientific format (missing "e") into double
// search is based on finding '-' or '+' sign without preoccuring character
//-----------------------------------------------------------------------/
double readNastranSci(const string& nastranNumber)
{
    // determine if scientific format
    bool scientificFormat = false;
    bool signExists = false;
    int exponentLocation = 0;
    for (int i=nastranNumber.size()-1; i>=0; i--){
        // check if sign exists
        if (nastranNumber[i]=='-' || nastranNumber[i]=='+')
            signExists = true;
        else if (signExists && nastranNumber[i]!=' '){
            // scientific format detected
            scientificFormat = true;
            exponentLocation = i+1;
            break;
        }
    }

    // read data into double
    if (scientificFormat){
        double coef = atof(nastranNumber.substr(0, exponentLocation).c_str());
        double expo = atof(nastranNumber.substr(   exponentLocation).c_str());
        return coef*pow(10, expo);
    }
    else
        return atof(nastranNumber.c_str());
}

//-----------------------------------------------------------------------/
// reads VTK file into mesh
//-----------------------------------------------------------------------/
void readVTKFile(
        CMesh* pMesh,
        const string& fileName,
        const long& shiftNodeIDs,
        const long& shiftElemIDs)
{
    // open file
    ifstream inputFile(fileName.c_str());
    if (!inputFile.is_open()){
        showErrorMessage("Cannot open VTK file");
        return;
    }

    // read lines with stripped comments
    vecStr fileLines;
    string line;
    while (getLine(inputFile, line))
        fileLines.push_back(line.substr(0, line.find("**")));
    inputFile.close();

    // read nodes
    uInt nodeCounter = 0;
    bool nodefound = false;
    for (uInt lineIndex=0; lineIndex<fileLines.size(); lineIndex++){
        vecStr vecStartLine = splitString(change2SingleSpace(fileLines[lineIndex]), ' ');
        if (vecStartLine.size()>1 && convert2UpperCase(vecStartLine[0])=="POINTS"){
            // node definition table found so read nodes
            uInt nodeNum = atoi(vecStartLine[1].c_str());
            if (checkEndOfFile(nodeNum, lineIndex, fileLines.size()))
                break;

            for (uInt nodeIndex=0; nodeIndex<nodeNum; nodeIndex++){
                lineIndex++;
                vecStr vecLine = splitString(change2SingleSpace(fileLines[lineIndex]), ' ');
                if (vecLine.size() != 3){
                    showLineErrorMessage("Incorrect node definition", lineIndex);
                    return;
                }

                uInt nodeID = nodeCounter+1+shiftNodeIDs;
                nodeCounter++;

                // read node coordinates
                double coordinates[3];
                for (int i=0; i<3; i++)
                    coordinates[i] = atof(vecLine[i].c_str());

                int retCode = pMesh->insertNodeWithIDCoords(nodeID, coordinates);
                if (retCode == -2){
                    showLineErrorMessage("Node ID already exists", lineIndex);
                    return;
                }
                else if (retCode == -3){
                    showLineErrorMessage("Node ID is equal 0", lineIndex);
                    return;
                }
                nodefound = true;
            }
        }
    }

    // check if something was read
    if (!nodefound)
        showWarningMessage("Zero nodes found");

    // read elements
    uInt elemCounter = 0;
    bool elemfound = false;
    for (uInt lineIndex=0; lineIndex<fileLines.size(); lineIndex++){
        vecStr vecStartLine = splitString(change2SingleSpace(fileLines[lineIndex]), ' ');
        if (vecStartLine.size()>1 && convert2UpperCase(vecStartLine[0])=="CELLS"){
            // element definition table found so read elements
            uInt cellNum = atoi(vecStartLine[1].c_str());
            if (checkEndOfFile(cellNum, lineIndex, fileLines.size()))
                break;

            for (uInt cellIndex=0; cellIndex<cellNum; cellIndex++){
                lineIndex++;
                vecStr vecLine = splitString(change2SingleSpace(fileLines[lineIndex]), ' ');
                // check element type
                uInt elemType;
                uInt elemId;
                vecUInt nodeIDs;
                if (vecLine.size()==5 && vecLine[0]=="4"){
                    // read linear quad plate elements
                    elemType = 1;

                    // read element ID
                    elemId = elemCounter+1+shiftElemIDs;
                    elemCounter++;
                    if (elemId <= 0){
                        showLineErrorMessage("Element ID should be greater than 0", lineIndex);
                        return;
                    }

                    // read node IDs
                    nodeIDs.resize(4);
                    for (uInt i=0; i<4; i++)
                        nodeIDs[i] = atoi(vecLine[i+1].c_str())+1+shiftNodeIDs;
                }
                else if (vecLine.size()==4 && vecLine[0]=="3"){
                    // read linear tria plate elements
                    elemType = 2;

                    // read element ID
                    elemId = elemCounter+1+shiftElemIDs;
                    elemCounter++;
                    if (elemId <= 0){
                        showLineErrorMessage("Element ID should be greater than 0", lineIndex);
                        return;
                    }

                    // read node IDs
                    nodeIDs.resize(3);
                    for (uInt i=0; i<3; i++)
                        nodeIDs[i] = atoi(vecLine[i+1].c_str())+1+shiftNodeIDs;
                }

                // insert element into mesh
                int retVal = pMesh->insertElemWithIDTypeNodeIDs(elemId, elemType, nodeIDs);

                // process return value
                string message = "Error in element definition";
                if      (retVal == -2) message = "Duplicate element ID detected";
                else if (retVal == -3) message = "Element ID 0 detected";
                else if (retVal == -4) message = "Duplicate node ID(s) given in element definition";
                else if (retVal == -5) message = "Zero surface element detected";
                else if (retVal == -6) message = "Element refers to non-existing node ID";
                if (retVal != 0){
                    showLineErrorMessage(message, lineIndex);
                    return;
                }

                elemfound = true;
            }
        }
    }

    // check if something was read
    if (!elemfound)
        showWarningMessage("Zero elements found");
}

//-----------------------------------------------------------------------/
// writes ApameSolver input file
//-----------------------------------------------------------------------/
int writeApameSolverInput(
        const string& apameFileName,
        CMesh* pMesh,
        CProjectData* pProjectData)
{
    // check ApameSolver data
    if (pProjectData->alfa.size() != pProjectData->beta.size()){
        showErrorMessage("Number of angles of attack and sideslip angles does not match");
        return -2;
    }
    if (pProjectData->origin.size() != 3){
        showErrorMessage("Number of Origin coordinates is not 3");
        return -3;
    }

    // get node data from mesh
    vecUInt nodeIDs;
    vecVecDbl coordinates;
    pMesh->getNodeData(nodeIDs, coordinates);

    // get element data from mesh
    vecUInt elemIDs;
    vecUInt elemTypes;
    vecVecUInt elemNodeIDs;
    vecVecUInt neighborIDs;
    vecVecUInt trailIDs;
    int retCodeElem = pMesh->getElemSolverData(elemIDs, elemTypes, elemNodeIDs, neighborIDs, trailIDs);
    if (retCodeElem == -1){
        showErrorMessage("There are wakes defined without trailing elements. Please check mesh!");
        return -4;
    }
    else if (retCodeElem==1 && pProjectData->velOrder!=0)
        showWarningMessage("Missing neighbors for (some) element(s). Consider running neighbor generation tool!");
    else if (retCodeElem != 0){
        showErrorMessage("Unknown problem reading element data from mesh");
        return -6;
    }

    // creating node and element index maps
    map<uInt,uInt> nodeIndexMap;
    for (uInt nodeIndex=0; nodeIndex<nodeIDs.size(); nodeIndex++)
        nodeIndexMap[nodeIDs[nodeIndex]] = nodeIndex;
    map<uInt,uInt> elemIndexMap;
    for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++)
        elemIndexMap[elemIDs[elemIndex]] = elemIndex;

    // check node, neighbor and trailing references
    for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
        // check nodes
        for (uInt nodeIndex=0; nodeIndex<elemNodeIDs[elemIndex].size(); nodeIndex++){
            if (nodeIndexMap.find(elemNodeIDs[elemIndex][nodeIndex]) == nodeIndexMap.end()){
                showErrorMessage("Element points to non-existing node. Please check mesh!");
                return -7;
            }
        }
        // check neighbors
        for (uInt neighborIndex=0; neighborIndex<neighborIDs[elemIndex].size(); neighborIndex++){
            if (elemIndexMap.find(neighborIDs[elemIndex][neighborIndex]) == elemIndexMap.end()){
                showErrorMessage("Element neighbor points to non-existing element. Please check mesh!");
                return -8;
            }
        }
        // check trails
        if (!trailIDs[elemIndex].empty()){
            if (elemIndexMap.find(trailIDs[elemIndex][0])==elemIndexMap.end() ||
                elemIndexMap.find(trailIDs[elemIndex][1])==elemIndexMap.end() ){
                showErrorMessage("Element trail points to non-existing element. Please check mesh!");
                return -9;
            }
        }
    }

    // open ApameSolver file for writing
    ofstream apameFile(apameFileName.c_str());
    if (!apameFile.is_open()){
        showErrorMessage("Cannot open ApameSolver input file for writing");
        return -1;
    }

    // write ApameSolver data
    apameFile << "APAME input file" << endl;
    apameFile << "VERSION 3.1"      << endl;
    apameFile << "AIRSPEED "        << pProjectData->airspeed << endl;
    apameFile << "DENSITY "         << pProjectData->density << endl;
    apameFile << "PRESSURE "        << pProjectData->pressure << endl;
    apameFile << "MACH "            << pProjectData->mach << endl;
    apameFile << "CASE_NUM "        << pProjectData->alfa.size() << endl;
    for (uInt i=0; i<pProjectData->alfa.size(); i++) apameFile << pProjectData->alfa[i] << " "; apameFile << endl;
    for (uInt i=0; i<pProjectData->beta.size(); i++) apameFile << pProjectData->beta[i] << " "; apameFile << endl;
    apameFile << "WINGSPAN "        << pProjectData->wingspan << endl;
    apameFile << "MAC "             << pProjectData->mac << endl;
    apameFile << "SURFACE "         << pProjectData->surface << endl;
    apameFile << "ORIGIN *"         << endl;
    apameFile << pProjectData->origin[0] << " " << pProjectData->origin[1] << " " << pProjectData->origin[2] << endl;
    apameFile << "METHOD "          << pProjectData->method << endl;
    apameFile << "ERROR "           << pProjectData->error << endl;
    apameFile << "COLLDIST "        << pProjectData->collDist << endl;
    apameFile << "FARFIELD "        << pProjectData->farField << endl;
    apameFile << "COLLCALC "        << pProjectData->collCalc << endl;
    apameFile << "VELORDER "        << pProjectData->velOrder << endl;
    if (pProjectData->results)                apameFile << "RESULTS 1" ; else apameFile << "RESULTS 0" ; apameFile << endl;
    if (pProjectData->coefficients)           apameFile << "RES_COEF 1"; else apameFile << "RES_COEF 0"; apameFile << endl;
    if (pProjectData->forces)                 apameFile << "RES_FORC 1"; else apameFile << "RES_FORC 0"; apameFile << endl;
    if (pProjectData->geometry)               apameFile << "RES_GEOM 1"; else apameFile << "RES_GEOM 0"; apameFile << endl;
    if (pProjectData->velocity)               apameFile << "RES_VELO 1"; else apameFile << "RES_VELO 0"; apameFile << endl;
    if (pProjectData->pressureCoefficient)    apameFile << "RES_PRES 1"; else apameFile << "RES_PRES 0"; apameFile << endl;
    if (pProjectData->centerPoints)           apameFile << "RES_CENT 1"; else apameFile << "RES_CENT 0"; apameFile << endl;
    if (pProjectData->doubletValues)          apameFile << "RES_DOUB 1"; else apameFile << "RES_DOUB 0"; apameFile << endl;
    if (pProjectData->sourceValues)           apameFile << "RES_SORC 1"; else apameFile << "RES_SORC 0"; apameFile << endl;
    if (pProjectData->velocityComponents)     apameFile << "RES_VELC 1"; else apameFile << "RES_VELC 0"; apameFile << endl;
    if (pProjectData->meshCharacteristics)    apameFile << "RES_MESH 1"; else apameFile << "RES_MESH 0"; apameFile << endl;
    if (pProjectData->staticPressure)         apameFile << "RES_STAT 1"; else apameFile << "RES_STAT 0"; apameFile << endl;
    if (pProjectData->dynamicPressure)        apameFile << "RES_DYNA 1"; else apameFile << "RES_DYNA 0"; apameFile << endl;
    if (pProjectData->manometerPressure)      apameFile << "RES_MANO 1"; else apameFile << "RES_MANO 0"; apameFile << endl;

    // write nodes
    apameFile << "NODES " << coordinates.size() << endl;
    for (uInt nodeIndex=0; nodeIndex<coordinates.size(); nodeIndex++){
        for (uInt i=0; i<coordinates[nodeIndex].size(); i++){
            apameFile.width(15);
            apameFile << scientific << coordinates[nodeIndex][i];
        }
        apameFile << endl;
    }

    // writing elements
    apameFile << "PANELS " << elemIDs.size() << endl;
    // first write body and dummy panels
    for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
        if (elemTypes[elemIndex] == 1  ||
            elemTypes[elemIndex] == 2  ||
            elemTypes[elemIndex] == 20 ||
            elemTypes[elemIndex] == 21 ){
            // write type
            apameFile << elemTypes[elemIndex];

            // write element constituing nodes
            for (uInt nodeIndex=0; nodeIndex<elemNodeIDs[elemIndex].size(); nodeIndex++){
                // using previously defined map extract index from node ID
                apameFile.width(12);
                apameFile << nodeIndexMap[elemNodeIDs[elemIndex][nodeIndex]]+1;
            }

            // write element neighbors
            uInt neighborSize = min(neighborIDs[elemIndex].size(), elemNodeIDs[elemIndex].size());
            for (uInt neighborIndex=0; neighborIndex<neighborSize; neighborIndex++){
                // using previously defined map extract index from element ID
                apameFile.width(12);
                apameFile << elemIndexMap[neighborIDs[elemIndex][neighborIndex]]+1;
            }
            // fill missing neighbors with zeros
            for (uInt i=neighborIDs[elemIndex].size(); i<elemNodeIDs[elemIndex].size(); i++){
                apameFile.width(12);
                apameFile << 0;
            }

            apameFile << endl;
        }
    }
    // then write wake panels
    for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
        if (elemTypes[elemIndex] == 10 ||
            elemTypes[elemIndex] == 11 ){
            // write type
            apameFile << elemTypes[elemIndex];

            // write element constituing nodes
            for (uInt nodeIndex=0; nodeIndex<elemNodeIDs[elemIndex].size(); nodeIndex++){
                // using previously defined map extract index from node ID
                apameFile.width(12);
                apameFile << nodeIndexMap[elemNodeIDs[elemIndex][nodeIndex]]+1;
            }

            // write trailing edge elements
            for (uInt trailIndex=0; trailIndex<trailIDs[elemIndex].size(); trailIndex++){
                // using previously defined map extract index from element ID
                apameFile.width(12);
                apameFile << elemIndexMap[trailIDs[elemIndex][trailIndex]]+1;
            }

            apameFile << endl;
        }
    }

    apameFile << "# end of input file" << endl;

    apameFile.close();

    return 0;
}

//-----------------------------------------------------------------------/
// exports Abaqus input file
//-----------------------------------------------------------------------/
int exportAbaqusFile(
        const string& fileName,
        CMesh* pMesh,
        CResults* pResults,
        const bool& writeMesh,
        const bool& writeProperty,
        const bool& writeLoads,
        const string& modulus,
        const string& poisson,
        const string& thickness,
        const string& selectedData,
        const string& selectedCase)
{
    // open Abaqus file for writing
    ofstream abaqusFile(fileName.c_str());
    if (!abaqusFile.is_open()){
        showErrorMessage("Cannot open Abaqus file for writing");
        return -1;
    }

    // write title
    abaqusFile << "************************************************************" << endl;
    abaqusFile << "** Abaqus input file generated by ApameGUI" << endl;
    abaqusFile << "************************************************************" << endl;

    // write mesh
    if (writeMesh){
        // get node data
        vecUInt nodeIDs;
        vecVecDbl coordinates;
        pMesh->getNodeData(nodeIDs, coordinates);

        // write nodes
        if (!nodeIDs.empty()){
            abaqusFile << "*NODE, SYSTEM=R" << endl;
            for (uInt nodeIndex=0; nodeIndex<nodeIDs.size(); nodeIndex++)
                abaqusFile << nodeIDs    [nodeIndex]    << ", " <<
                              coordinates[nodeIndex][0] << ", " <<
                              coordinates[nodeIndex][1] << ", " <<
                              coordinates[nodeIndex][2] << endl;
            abaqusFile << "**" << endl;
        }

        // get element data
        vecUInt elemIDs;
        vecUInt elemTypes;
        vecVecUInt elemNodes;
        pMesh->getElemsData(elemIDs, elemTypes, elemNodes);

        // write quad elements
        if (pMesh->getNumBodyQuad()+pMesh->getNumDummQuad() > 0){
            abaqusFile << "*ELEMENT, TYPE=S4";
            if (writeProperty)
                abaqusFile << ", ELSET=PLATE";
            abaqusFile << endl;
            for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
                if (elemTypes[elemIndex]==1 || elemTypes[elemIndex]==20)
                    abaqusFile << elemIDs  [elemIndex]     << ", " <<
                                  elemNodes[elemIndex][0] << ", " <<
                                  elemNodes[elemIndex][1] << ", " <<
                                  elemNodes[elemIndex][2] << ", " <<
                                  elemNodes[elemIndex][3] << endl;
            }
            abaqusFile << "**" << endl;
        }

        // write tria elements
        if (pMesh->getNumBodyTria()+pMesh->getNumDummTria() > 0){
            abaqusFile << "*ELEMENT, TYPE=S3";
            if (writeProperty)
                abaqusFile << ", ELSET=PLATE";
            abaqusFile << endl;
            for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
                if (elemTypes[elemIndex]==2 || elemTypes[elemIndex]==21)
                    abaqusFile << elemIDs  [elemIndex]     << ", " <<
                                  elemNodes[elemIndex][0] << ", " <<
                                  elemNodes[elemIndex][1] << ", " <<
                                  elemNodes[elemIndex][2] << endl;
            }
            abaqusFile << "**" << endl;
        }
    }

    // write property data
    if (writeProperty){
        // write plate property
        abaqusFile << "*SHELL SECTION, ELSET=PLATE, MATERIAL=ALUMINUM" << endl;
        abaqusFile << " " << thickness << "," << endl;
        abaqusFile << "**" << endl;
        // write material data
        abaqusFile << "*MATERIAL, NAME=ALUMINUM" << endl;
        abaqusFile << "*ELASTIC, TYPE=ISOTROPIC" << endl;
        abaqusFile << " " << modulus << ", " << poisson << ", 0." << endl;
        abaqusFile << "**" << endl;
    }

    // write load
    if (writeLoads){
        // write step data
        abaqusFile << "*STEP, INC=100" << endl;
        abaqusFile << "Static Load from Apame 3D Panel Method" << endl;
        abaqusFile << "*STATIC" << endl;
        abaqusFile << "**" << endl;
        abaqusFile << "*NODE PRINT, FREQUENCY=1" << endl;
        abaqusFile << "    U," << endl;
        abaqusFile << "*FILE FORMAT, ASCII" << endl;
        abaqusFile << "*EL PRINT, ELSET=OUT_PLT,FREQUENCY=1, POSITION=CENTROIDAL" << endl;
        abaqusFile << "    S," << endl;
        abaqusFile << " SINV," << endl;
        abaqusFile << "**" << endl;

        // get index of selected data in results data structure
        int dataIndex = -1;
        for (uInt i=0; i<pResults->surfDatas.size(); i++){
            if (pResults->surfDatas[i].dataName == selectedData){
                dataIndex = i;
                break;
            }
        }
        if (dataIndex == -1){
            showErrorMessage("Selected field not found in results, contact support!");
            return -1;
        }

        // get index of case in results data structure
        int caseIndex = -1;
        for (uInt i=0; i<pResults->surfDatas[dataIndex].caseTitles.size(); i++){
            if (pResults->surfDatas[dataIndex].caseTitles[i] == selectedCase){
                caseIndex = i;
                break;
            }
        }
        if (caseIndex == -1){
            showErrorMessage("Selected case not found in results, contact support!");
            return -1;
        }

        // check data sizes
        if (pResults->surfDatas[dataIndex].data[caseIndex].size() != pResults->surfDatas[dataIndex].elemIDs.size()){
            showErrorMessage("Field data size not matching element number, contact support!");
            return -1;
        }

        // write distributed load data
        abaqusFile << "*DLOAD, OP=NEW" << endl;
        for (uInt elemIndex=0; elemIndex<pResults->surfDatas[dataIndex].elemIDs.size(); elemIndex++)
            abaqusFile << pResults->surfDatas[dataIndex].elemIDs[elemIndex] << ", P, " << -pResults->surfDatas[dataIndex].data[caseIndex][elemIndex] << endl;
        abaqusFile << "**" << endl;
    }

    abaqusFile.close();

    return 0;
}

//-----------------------------------------------------------------------/
// exports Nastran file
//-----------------------------------------------------------------------/
int exportNastranFile(
        const string& fileName,
        CMesh* pMesh,
        CResults* pResults,
        const bool& writeMesh,
        const bool& writeProperty,
        const bool& writeLoads,
        const string& modulus,
        const string& poisson,
        const string& thickness,
        const string& selectedData,
        const string& selectedCase)
{
    // open Nastran file for writing
    ofstream nastranFile(fileName.c_str());
    if (!nastranFile.is_open()){
        showErrorMessage("Cannot open Nastran file for writing");
        return -1;
    }

    // write title
    nastranFile << "$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$" << endl;
    nastranFile << "$  Nastran input file generated by ApameGUI" << endl;
    nastranFile << "$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$" << endl;

    // write load
    if (writeLoads){
        // write step data
        nastranFile << "INIT MASTER(S)" << endl;
        nastranFile << "NASTRAN SYSTEM(319)=1" << endl;
        nastranFile << "ID Apame,Apame" << endl;
        nastranFile << "SOL SESTATIC" << endl;
        nastranFile << "TIME 10000" << endl;
        nastranFile << "CEND" << endl;
        nastranFile << "TITLE = Static Load from Apame 3D Panel Method" << endl;
        nastranFile << "ECHO = NONE" << endl;
        nastranFile << "DISPLACEMENT(PLOT) = ALL" << endl;
        nastranFile << "STRESS(PLOT) = ALL" << endl;
        nastranFile << "LOAD = 1" << endl;
        nastranFile << "BEGIN BULK" << endl;
        nastranFile << "$" << endl;

        // get index of selected data in results data structure
        int dataIndex = -1;
        for (uInt i=0; i<pResults->surfDatas.size(); i++){
            if (pResults->surfDatas[i].dataName == selectedData){
                dataIndex = i;
                break;
            }
        }
        if (dataIndex == -1){
            showErrorMessage("Selected field not found in results, contact support!");
            nastranFile.close();
            return -1;
        }

        // get index of case in results data structure
        int caseIndex = -1;
        for (uInt i=0; i<pResults->surfDatas[dataIndex].caseTitles.size(); i++){
            if (pResults->surfDatas[dataIndex].caseTitles[i] == selectedCase){
                caseIndex = i;
                break;
            }
        }
        if (caseIndex == -1){
            showErrorMessage("Selected case not found in results, contact support!");
            nastranFile.close();
            return -1;
        }

        // check data sizes
        if (pResults->surfDatas[dataIndex].data[caseIndex].size() != pResults->surfDatas[dataIndex].elemIDs.size()){
            showErrorMessage("Field data size not matching element number, contact support!");
            nastranFile.close();
            return -1;
        }

        // write distributed load data
        nastranFile << "*DLOAD, OP=NEW" << endl;
        for (uInt elemIndex=0; elemIndex<pResults->surfDatas[dataIndex].elemIDs.size(); elemIndex++)
            nastranFile << "PLOAD4,1," << pResults->surfDatas[dataIndex].elemIDs[elemIndex] << "," << -pResults->surfDatas[dataIndex].data[caseIndex][elemIndex] << endl;
        nastranFile << "$" << endl;
    }

    // write mesh
    if (writeMesh){
        // get node data
        vecUInt nodeIDs;
        vecVecDbl coordinates;
        pMesh->getNodeData(nodeIDs, coordinates);

        // write nodes
        if (!nodeIDs.empty()){
            for (uInt nodeIndex=0; nodeIndex<nodeIDs.size(); nodeIndex++)
                nastranFile << "GRID," << nodeIDs    [nodeIndex]    << ",0," <<
                                          coordinates[nodeIndex][0] << ","  <<
                                          coordinates[nodeIndex][1] << ","  <<
                                          coordinates[nodeIndex][2] << ",0" << endl;
            nastranFile << "$" << endl;
        }

        // get element data
        vecUInt elemIDs;
        vecUInt elemTypes;
        vecVecUInt elemNodes;
        pMesh->getElemsData(elemIDs, elemTypes, elemNodes);

        // write quad elements
        if (pMesh->getNumBodyQuad()+pMesh->getNumDummQuad() > 0){
            for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
                if (elemTypes[elemIndex]==1 || elemTypes[elemIndex]==20)
                    nastranFile << "CQUAD4," << elemIDs  [elemIndex]    << ",1," <<
                                                elemNodes[elemIndex][0] << ","  <<
                                                elemNodes[elemIndex][1] << ","  <<
                                                elemNodes[elemIndex][2] << ","  <<
                                                elemNodes[elemIndex][3] << endl;
            }
            nastranFile << "$" << endl;
        }

        // write tria elements
        if (pMesh->getNumBodyTria()+pMesh->getNumDummTria() > 0){
            for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
                if (elemTypes[elemIndex]==2 || elemTypes[elemIndex]==21){
                    nastranFile << "CTRIA3," << elemIDs  [elemIndex]    << ",1," <<
                                                elemNodes[elemIndex][0] << ","  <<
                                                elemNodes[elemIndex][1] << ","  <<
                                                elemNodes[elemIndex][2] << endl;
                }
            }
            nastranFile << "$" << endl;
        }
    }

    // write property data
    if (writeProperty){
        // write plate property
        nastranFile << "PSHELL,1,1," << thickness << ",1,1,0." << endl;
        nastranFile << "$" << endl;
        // write material data
        nastranFile << "MAT1,1," << modulus << "," << poisson << ",0.,0.,0." << endl;
        nastranFile << "$" << endl;
    }

    nastranFile << "ENDDATA" << endl;

    nastranFile.close();

    return 0;
}

//-----------------------------------------------------------------------/
// exports VTK file
//-----------------------------------------------------------------------/
int exportVTKFile(
        const string& fileName,
        CMesh* pMesh,
        CResults* pResults,
        const bool& plotResults,
        const string& plotType)
{
    // open VTK file for writing
    ofstream VTKFile(fileName.c_str());
    if (!VTKFile.is_open()){
        showErrorMessage("Cannot open VTK file for writing");
        return -1;
    }

    // write header
    VTKFile << "# vtk DataFile Version 3.0" << endl;
    VTKFile << "VTK Sample File" << endl;
    VTKFile << "ASCII" << endl;

    // get node data
    vecUInt nodeIDs;
    vecVecDbl coordinates;
    pMesh->getNodeData(nodeIDs, coordinates);

    // write nodes
    VTKFile << "DATASET UNSTRUCTURED_GRID" << endl;
    VTKFile << "POINTS " << coordinates.size() << " float" << endl;
    for (uInt nodeIndex=0; nodeIndex<coordinates.size(); nodeIndex++)
        VTKFile << coordinates[nodeIndex][0] << " " <<
                   coordinates[nodeIndex][1] << " " <<
                   coordinates[nodeIndex][2] << endl;

    // creating node index map
    map<uInt,uInt> nodeIndexMap;
    for (uInt nodeIndex=0; nodeIndex<coordinates.size(); nodeIndex++)
        nodeIndexMap[nodeIDs[nodeIndex]] = nodeIndex;

    // get element data
    vecUInt elemIDs;
    vecUInt elemTypes;
    vecVecUInt elemNodes;
    pMesh->getElemsData(elemIDs, elemTypes, elemNodes);

    // write cell definition
    uInt bodyDummyNum = pMesh->getNumBodyQuad() +
                        pMesh->getNumDummQuad() +
                        pMesh->getNumBodyTria() +
                        pMesh->getNumDummTria();
    uInt countEntity = (pMesh->getNumBodyQuad()+pMesh->getNumDummQuad())*5 +
                       (pMesh->getNumBodyTria()+pMesh->getNumDummTria())*4;
    VTKFile << "CELLS " << bodyDummyNum << " " << countEntity << endl;
    for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
        if      (elemTypes[elemIndex]==1 || elemTypes[elemIndex] == 20) VTKFile << "4";
        else if (elemTypes[elemIndex]==2 || elemTypes[elemIndex] == 21) VTKFile << "3";

        if (elemTypes[elemIndex]==1  ||
            elemTypes[elemIndex]==2  ||
            elemTypes[elemIndex]==20 ||
            elemTypes[elemIndex]==21 ){
            for (uInt j=0; j<elemNodes[elemIndex].size(); j++){
                map<uInt,uInt>::iterator it=nodeIndexMap.find(elemNodes[elemIndex][j]);
                if (it != nodeIndexMap.end())
                    VTKFile << " " << it->second;
            }
            VTKFile << endl;
        }
    }

    // write cell type declaration
    VTKFile << "CELL_TYPES " << bodyDummyNum << endl;
    for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
        if      (elemTypes[elemIndex]==1 || elemTypes[elemIndex] == 20) VTKFile << "9" << endl;
        else if (elemTypes[elemIndex]==2 || elemTypes[elemIndex] == 21) VTKFile << "5" << endl;
    }

    // write cell data
    if (plotResults                          && // if writing plot data requested
        !pResults->surfDatas.empty()         && // data results exist
        !pResults->surfDatas[0].data.empty() ){ // data on elements exist
        if (plotType == "elemental"){
            VTKFile << "CELL_DATA " << pResults->surfDatas[0].data[0].size() << endl;

            // write data
            for (uInt dataIndex=0; dataIndex<pResults->surfDatas.size(); dataIndex++){
                for (uInt caseIndex=0; caseIndex<pResults->surfDatas[dataIndex].caseTitles.size(); caseIndex++){
                    VTKFile << "FIELD " << replaceStringSegment(pResults->surfDatas[dataIndex].dataName, " ", "_") << " 1" << endl;
                    VTKFile << replaceStringSegment(pResults->surfDatas[dataIndex].dataName, " ", "_") << "_" << replaceStringSegment(pResults->surfDatas[dataIndex].caseTitles[caseIndex], " ", "_") << " 1 " << bodyDummyNum << " float" << endl;
                    for (uInt elemIndex=0; elemIndex<pResults->surfDatas[dataIndex].data[caseIndex].size(); elemIndex++){
                        VTKFile.width(15);
                        VTKFile << scientific << pResults->surfDatas[dataIndex].data[caseIndex][elemIndex] << endl;
                    }
                }
            }
        }
        else{
            // get data from mesh
            vecUInt nodeIDs;
            vecVecDbl coordinates;
            vecUInt allElemIDs;
            vecVecUInt elemNodeIDs;
            pMesh->getIDWData(nodeIDs, coordinates, allElemIDs, elemNodeIDs);

            // creating node and element index map
            map<uInt,uInt> nodeIndexMap;
            for (uInt nodeIndex=0; nodeIndex<nodeIDs.size(); nodeIndex++)
                nodeIndexMap[nodeIDs[nodeIndex]] = nodeIndex;
            map<uInt,uInt> elemIndexMap;
            for (uInt elemIndex=0; elemIndex<allElemIDs.size(); elemIndex++)
                elemIndexMap[allElemIDs[elemIndex]] = elemIndex;

            // extract element node indexes and element center coordinates
            vecVecUInt elementNodes(elemIDs.size());
            vecVecDbl elemCenterCoords(elemIDs.size());
            // for each element of interest (its ID)
            for (uInt elemIndex=0; elemIndex<elemIDs.size(); elemIndex++){
                // get its index in all elements from mesh
                uInt allElemIndex = elemIndexMap[elemIDs[elemIndex]];
                // resize element nodes to number of nodes
                elementNodes[elemIndex].resize(elemNodeIDs[allElemIndex].size());
                // assign node indexes
                for (uInt nodeIndex=0; nodeIndex<elementNodes[elemIndex].size(); nodeIndex++)
                    elementNodes[elemIndex][nodeIndex] = nodeIndexMap[elemNodeIDs[allElemIndex][nodeIndex]];
                // calculate element center coordinates
                vecDbl centerCoord(3, 0.0);
                for (uInt nodeIndex=0; nodeIndex<elementNodes[elemIndex].size(); nodeIndex++){
                    for (uInt i=0; i<3; i++)
                        centerCoord[i] += coordinates[nodeIndexMap[elemNodeIDs[allElemIndex][nodeIndex]]][i];
                }
                for (uInt i=0; i<3; i++)
                    centerCoord[i] /= elementNodes[elemIndex].size();
                elemCenterCoords[elemIndex] = centerCoord;
            }

            // calculate weights
            vecVecDbl nodeWeights;
            if (weightFactors(
                    nodeWeights,
                    elementNodes,
                    coordinates,
                    elemCenterCoords)){
                showErrorMessage("Problem calculating weights for IDW - Contact Support!");
                return -1;
            }

            // write point data
            VTKFile << "POINT_DATA " << nodeIDs.size() << endl;
            for (uInt dataIndex=0; dataIndex<pResults->surfDatas.size(); dataIndex++){
                for (uInt caseIndex=0; caseIndex<pResults->surfDatas[dataIndex].caseTitles.size(); caseIndex++){
                    VTKFile << "FIELD " << replaceStringSegment(pResults->surfDatas[dataIndex].dataName, " ", "_") << " 1" << endl;
                    VTKFile << replaceStringSegment(pResults->surfDatas[dataIndex].dataName, " ", "_") << "_" << replaceStringSegment(pResults->surfDatas[dataIndex].caseTitles[caseIndex], " ", "_") << " 1 " << nodeIDs.size() << " float" << endl;

                    // calculate node field
                    vecDbl nodeField(nodeIDs.size(), 0.0);
                    if (weightAverage(
                            nodeField,
                            nodeWeights,
                            elementNodes,
                            pResults->surfDatas[dataIndex].data[caseIndex])){
                        showErrorMessage("Problem calculating calculating IDW - Contact Support!");
                        return -1;
                    }

                    for (uInt nodeIndex=0; nodeIndex<nodeField.size(); nodeIndex++){
                        VTKFile.width(15);
                        VTKFile << scientific << nodeField[nodeIndex] << endl;
                    }
                }
            }

            // write velocity vector data, check if they exist first
            bool xFound = false;
            bool yFound = false;
            bool zFound = false;
            uInt xLoc = 0;
            uInt yLoc = 0;
            uInt zLoc = 0;
            for (uInt i=0; i<pResults->surfDatas.size(); i++){
                if      (pResults->surfDatas[i].dataName == "Velocity X-component"){ xFound = true; xLoc = i; }
                else if (pResults->surfDatas[i].dataName == "Velocity Y-component"){ yFound = true; yLoc = i; }
                else if (pResults->surfDatas[i].dataName == "Velocity Z-component"){ zFound = true; zLoc = i; }
            }
            if (xFound && yFound && zFound){
                // write title
                VTKFile << "FIELD Velocity_Vector " << pResults->surfDatas[xLoc].caseTitles.size() << endl;

                for (uInt caseIndex=0; caseIndex<pResults->surfDatas[xLoc].caseTitles.size(); caseIndex++){
                    VTKFile << "Velocity_Vector_" << replaceStringSegment(pResults->surfDatas[xLoc].caseTitles[caseIndex], " ", "_") << " 3 " << nodeIDs.size() << " float" << endl;

                    vecDbl fieldX(nodeIDs.size(), 0.0);
                    vecDbl fieldY(nodeIDs.size(), 0.0);
                    vecDbl fieldZ(nodeIDs.size(), 0.0);
                    weightAverage(fieldX, nodeWeights, elementNodes, pResults->surfDatas[xLoc].data[caseIndex]);
                    weightAverage(fieldY, nodeWeights, elementNodes, pResults->surfDatas[yLoc].data[caseIndex]);
                    weightAverage(fieldZ, nodeWeights, elementNodes, pResults->surfDatas[zLoc].data[caseIndex]);

                    for (uInt nodeIndex=0; nodeIndex<nodeIDs.size(); nodeIndex++){
                        VTKFile << " ";
                        VTKFile.width(15);
                        VTKFile << scientific << fieldX[nodeIndex] << " " << scientific << fieldY[nodeIndex] << " " << scientific << fieldZ[nodeIndex] << endl;
                    }
                }
            }
        }
    }

    VTKFile.close();

    return 0;
}
