// Copyright (C) 2011-2012  Daniel Filkovic

// This file is part of APAME.

// APAME is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// APAME is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with APAME.  If not, see <http://www.gnu.org/licenses/>.

// file 05_example_optimization_simple_wing.sci

// This is an example script that shows how to use ApameSolver for
// optimization purposes within Scilab
// At the end of the script is animate_variables function that is used
// to plot frames for optimization animation

// optimization goals:
//    - minimize drag
//    - approach 730N lift force
// optimization variables:
//    - variable(1) - root chord
//    - variable(2) - tip chord

clear
close
clc

getd('simple_wing/')
getd('general/')

funcprot(0)
stacksize('max')
pi=4.0*atan(1.0);

// evaluation function that returns optimization value
function [O, index]=run_case(variables, index)
    pi=4.0*atan(1.0);
    
    // initialize data structures
    [apame_data]=initialize_apame_data();
    
    // enter job_name
    job_name="simple_wing";
    // request geometry data and pressure coefficient distribution results
    apame_data.res_forc=1;
    // provide path to ApameSolver executable
    solver_name="..\ApameSolver\bin\apame_win64";
    
    // define airspeed and AOA
    apame_data.airspeed=27.778;
    apame_data.alfa=4*pi/180;

    //==============================================================================
    // calling function that creates mesh
    //==============================================================================
    // geometry data
    apame_data.wing_span=5;                               // wing span, from left to right wing tip (whole wing)
    wing_data=struct("croot",0.5,    ...                    // root chord length            [m]
                     "ctip",0.5,    ...                   // tip chord length             [m]
                     "tipAL",0*pi/180,    ...            // tip angle of attack          [radians]
                     "lambda14",0*pi/180,    ...         // wing sweep (1/4 wing)        [radians]
                     "dihedral",0*pi/180,    ...          // dihedral                     [radians]
                     "r_foil","airfoils/66415_40.DAT", ...// root airfoil data file
                     "t_foil","airfoils/66415_40.DAT", ...// tip airfoil data file
                     "wake_rows",10,    ...               // number of wake rows
                     "wake_length",10,    ...             // length of wake (in span multiples)
                     "N_max",100,    ...                  // max. airfoil panel number (for matrix initialization)
                     "Mp",15);                            // number of panels on halfspan
    // wake data
    wake_rows=10;    // number of wake rows
    wake_length=10;  // length of wake (in span multiples)
    //==============================================================================

    // update variables with optimization variables
    wing_data.croot=variables(1);
    wing_data.ctip=variables(2);
    
    // calculate mesh
    [nodes,elements,apame_data]=simple_wing(apame_data,wing_data,wake_rows,wake_length);
    
    // writing apame file... remove this later on
    write_apame_input(job_name+".inp",nodes,elements,apame_data);
    
    // run ApameSolver
    unix_s(solver_name+" "+job_name);
    
    // read ApameSolver results
    apame_results_data=read_apame_results(job_name+'.res');

    // calculate total drag (induced + viscous)
    Re=apame_data.ro*apame_data.airspeed*apame_data.mac/1.7894e-05;
    Cf=0.0583/Re^0.2;
    t=0.15;
    xt=0.5;
    FF=1 + 0.6 * t/xt+100 * t^4;
    FDRAG=apame_results_data.FDRAG+FF*Cf*apame_data.S_wet*apame_data.ro*apame_data.airspeed^2/2;
    
    // calculate optimization value
    O=FDRAG^2 + (730-apame_results_data.FLIFT)^2;
    mprintf("lift: %f  drag: %f  opt_value: %f\n",apame_results_data.FLIFT,FDRAG,O)
    mprintf("variables:")
    for i=1:length(variables)
        mprintf(" %f,",variables(i))
    end
    mprintf("\n")
    
    // save optimization variables
    save_file=file('open','saved_variables.txt','unknown');
    file("last",save_file)
    for i = 1:length(variables)
        write(save_file,variables(i),'(1x,E14.8,$)');
    end
    write(save_file,'');
    file('close',save_file);
endfunction

// initialize optimization variables
variables=[0.5, 0.5];

// save initial optimization variables
if getos() == 'Windows' then
    unix_s("del saved_variables.txt");
else
    unix_s("rm saved_variables.txt");
end
save_file=file('open','saved_variables.txt','new');
for i = 1:length(variables)
    write(save_file,variables(i),'(1x,E14.8,$)');
end
write(save_file,'');
file('close',save_file);

// run Scilab optimization
nm = neldermead_new ();
nm = neldermead_configure(nm,"-numberofvariables",2);
nm = neldermead_configure(nm,"-function",run_case);
nm = neldermead_configure(nm,"-x0",variables');
nm = neldermead_search(nm);
nm
xopt = neldermead_get(nm,"-xopt");

// save resulting optimization variables
save_file=file('open','saved_variables.txt','unknown');
file("last",save_file)
for i = 1:length(xopt)
    write(save_file,xopt(i),'(1x,E14.8,$)');
end
write(save_file,'');
file('close',save_file);

if getos() == 'Windows' then
    unix_s("del fort.2");
else
    unix_s("rm fort.2");
end

// function that plots geometry frames from saved variables
function animate_variables()
    // initialize apame data structure
    [apame_data]=initialize_apame_data();
    
    //==============================================================================
    // geometry data
    apame_data.wing_span=5;                               // wing span, from left to right wing tip (whole wing)
    wing_data=struct("croot",0.5,    ...                    // root chord length            [m]
                     "ctip",0.5,    ...                   // tip chord length             [m]
                     "tipAL",0*pi/180,    ...            // tip angle of attack          [radians]
                     "lambda14",0*pi/180,    ...         // wing sweep (1/4 wing)        [radians]
                     "dihedral",0*pi/180,    ...          // dihedral                     [radians]
                     "r_foil","airfoils/66415_40.DAT", ...// root airfoil data file
                     "t_foil","airfoils/66415_40.DAT", ...// tip airfoil data file
                     "wake_rows",10,    ...               // number of wake rows
                     "wake_length",10,    ...             // length of wake (in span multiples)
                     "N_max",100,    ...                  // max. airfoil panel number (for matrix initialization)
                     "Mp",15);                            // number of panels on halfspan
    // wake data
    wake_rows=10;    // number of wake rows
    wake_length=10;  // length of wake (in span multiples)
    //==============================================================================
    
    // open variable file
    var_file=file('open','saved_variables.txt','unknown');
    variables=read(var_file,-1,2);
    file ('close',var_file);
    
    for num_var=1:size(variables,"r")
        wing_data.croot=variables(num_var,1);
        wing_data.ctip=variables(num_var,2);
        
        // calculate mesh
        [nodes,elements,apame_data]=simple_wing(apame_data,wing_data,wake_rows,wake_length);

        // plot mesh
        j_1=0;
        for i=1:apame_data.elem_num
            if elements(i,1) == 1
                j_1=j_1 + 1;
                X_1(:,j_1)=[nodes(elements(i,2),1);nodes(elements(i,5),1);nodes(elements(i,4),1);nodes(elements(i,3),1)];
                Y_1(:,j_1)=[nodes(elements(i,2),2);nodes(elements(i,5),2);nodes(elements(i,4),2);nodes(elements(i,3),2)];
                Z_1(:,j_1)=[nodes(elements(i,2),3);nodes(elements(i,5),3);nodes(elements(i,4),3);nodes(elements(i,3),3)];        
            end
        end
        plot3d(X_1,Y_1,Z_1)
        h=gce();
        h.color_mode=29;
        a = gca();
        a.box = "on";
        a.isoview = "on";
        a.cube_scaling = "off";
        
        // save frame
        sCount=string(num_var);
        num_count=length(sCount);
        figure_name="figure_";
        for i=1:5-num_count
            figure_name=figure_name+"0";
        end
        figure_name=figure_name+sCount+".png";
        xs2png(gcf(),figure_name);
        clf;
    end
    
endfunction
