// Copyright (C) 2011  Daniel Filkovic

// This file is part of APAME.

// APAME is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// APAME is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with APAME.  If not, see <http://www.gnu.org/licenses/>.

// file lofted_wing.sci

// This is the main calling function that creates wing mesh according to
// given arguments. Wing is defined with 5 airfoil segments that form
// 4-th order polynom

function [nodes,elements,apame_data] = lofted_wing( ...
    apame_data, ...
    airfoils, ...
    wake_rows, ...
    wake_length, ...
    x_af, ...
    y_af, ...
    z_af, ...
    crd_af, ...
    aoa_af, ...
    min_el, ...
    max_el, ...
    N_max, ...
    M_max)

//==========================================================================
// AIRFOIL SEGMENTS CALCULATION
// extracting number of segments
M_seg = length(x_af);
// reading airfoil coordinates from airfoil files
[x_foil,z_foil,N] = read_airfoils(airfoils,N_max,M_seg);
// calculating polynoms and derivatives
[p_dih     ,p_swp     ,p_crd     ,p_aoa     , ...
 p_dih_der ,p_swp_der ,p_crd_der ,p_aoa_der , ...
 p_dih_der2,p_swp_der2,p_crd_der2,p_aoa_der2] = derivating(x_af,y_af,z_af,crd_af,aoa_af,M_seg);
// scaling airfoil coordinates (x and z) according to chord lengths
[x_foil,z_foil]=scale_seg(x_foil,z_foil,crd_af,N,M_seg);
// rotating airfoil coordinates according to dihedral and AOA
[x_foil,y_foil,z_foil]=rotat_seg(x_foil,z_foil,aoa_af,p_dih_der,N,M_seg,y_af,apame_data.wing_span);
// translating airfoil coordinates (x, y and z) according to their locations
[x_foil,y_foil,z_foil]=trans_seg(x_foil,y_foil,z_foil,N,M_seg,x_af,y_af,z_af,apame_data.wing_span);
//__________________________________________________________________________
// MESHING WING
// distributing airfoil ribs position based on global second derivation (curvature)
[y_ribs,M] = ribs_dist(p_dih_der,p_dih_der2,p_swp_der2,p_crd_der2,p_aoa_der2,y_af,min_el,max_el,M_max,M_seg);
// calculating x and z polynoms for each airfoil coordinate
[p_x,p_y,p_z] = poly_det(x_foil,y_foil,z_foil,N,M_seg);
// assembling final geometry based on a ribs position (y_ribs) and polynoms
[x,y,z] = assemble(y_ribs,p_x,p_y,p_z,apame_data.wing_span,N,M);
//__________________________________________________________________________
// ADDITIONAL ACTIONS ON MESH
// convert structured mesh to unstructured (add mirror)
[apame_data,nodes,elements] = convert_to_unstructured(apame_data,N,M,x,y,z);
// calculate reference and wetted surface area and mean aerodynamic chord
[apame_data] = calculate_reference_data(apame_data,nodes,elements,N,M);
// create wake elements
[nodes,elements,apame_data] = create_wake(nodes,elements,N,M,x,y,z,wake_rows,wake_length,apame_data);
//__________________________________________________________________________

endfunction
